@@ -15,24 +15,23 @@
 from __future__ import annotations
 
 from collections import defaultdict
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import enum
 import functools
 import math
 import operator as op
-from typing import Any, Callable, TYPE_CHECKING, cast
+from typing import Any, TYPE_CHECKING, cast
 
-from jax._src import abstract_arrays
 from jax._src import api
-from jax._src import api_util
 from jax._src import basearray
 from jax._src import config
 from jax._src import core
+from jax._src import deprecations
 from jax._src import dispatch
 from jax._src import dtypes
 from jax._src import errors
 from jax._src import profiler
-from jax._src import tree_util
+from jax._src import util
 from jax._src import xla_bridge
 from jax._src.interpreters import mlir
 from jax._src.interpreters import pxla
@@ -42,10 +41,10 @@ from jax._src.lib import xla_client as xc
 from jax._src.lib import xla_extension as xe
 from jax._src.sharding import Sharding
 from jax._src.sharding_impls import (
-    PmapSharding, SingleDeviceSharding,
+    PmapSharding, SingleDeviceSharding, NamedSharding,
     device_replica_id_map, hashed_index, num_addressable_indices, local_to_global_shape)  # pyformat: disable
 from jax._src.typing import ArrayLike, DLDeviceType
-from jax._src.util import safe_zip, unzip3, use_cpp_class, use_cpp_method
+from jax._src.util import safe_zip, unzip3, use_cpp_class, use_cpp_method, cache
 import numpy as np
 
 
@@ -116,11 +115,21 @@ def _reconstruct_array(fun, args, arr_state, aval_state):
   np_value = fun(*args)
   np_value.__setstate__(arr_state)
   jnp_value = api.device_put(np_value)
+  # TODO(slebedev): Remove this branch after December 10th 2024.
+  if "named_shape" in aval_state:
+    deprecations.warn(
+        "jax-aval-named-shape",
+        "Pickled array contains an aval with a named_shape attribute. This is"
+        " deprecated and the code path supporting such avals will be removed."
+        " Please re-pickle the array.",
+        stacklevel=2,
+    )
+    del aval_state["named_shape"]
   jnp_value.aval = jnp_value.aval.update(**aval_state)
   return jnp_value
 
 
-@functools.lru_cache(maxsize=4096)
+@cache(max_size=4096, trace_context_in_key=False)
 def _cached_index_calc(s, shape):
   map_ = s.addressable_devices_indices_map(shape)
   seen_h_indices = set()
@@ -133,7 +142,7 @@ def _cached_index_calc(s, shape):
   return l
 
 
-@functools.lru_cache(maxsize=4096)
+@cache(max_size=4096, trace_context_in_key=False)
 def _process_has_full_value_in_mcjax(s, shape):
   # Return False for single host as a fast path.
   if xla_bridge.process_count() == 1:
@@ -146,6 +155,28 @@ def _process_has_full_value_in_mcjax(s, shape):
   return num_unique_indices == num_addressable_unique_indices
 
 
+def _validate_shape_and_dtype_for_per_device_arrays(
+    arrays: Sequence[ArrayImpl | np.ndarray],
+    sharding: Sharding,
+    aval: core.ShapedArray,
+    expected_shape: Shape,
+):
+  """Validates that per-device arrays are valid and consistent."""
+  expected_dtype = aval.dtype
+  for db in arrays:
+    if db.dtype != expected_dtype:
+      raise ValueError(
+          "Input buffers to `Array` must have matching dtypes. "
+          f"Got {db.dtype}, expected {expected_dtype} for buffer: {db}"
+      )
+    if db.shape != expected_shape:
+      raise ValueError(
+          f"Expected shard shape {expected_shape} doesn't match the single "
+          f"device array shape {db.shape}. Shape of Array is "
+          f"{aval.str_short()} with sharding {sharding}"
+      )
+
+
 class ArrayImpl(basearray.Array):
   # TODO(yashkatariya): Add __slots__ here.
 
@@ -171,17 +202,11 @@ class ArrayImpl(basearray.Array):
     # Don't rearrange if skip_checks is enabled because this assumes that the
     # input buffers are already arranged properly. This usually happens when
     # Array's are created as output of a JAX transformation
-    # (like pjit, xmap, etc).
+    # (like pjit, etc).
     if not _skip_checks or config.enable_checks.value:
       self._check_and_rearrange()
 
   def _check_and_rearrange(self):
-    for db in self._arrays:
-      if db.dtype != self.dtype:
-        raise ValueError(
-            "Input buffers to `Array` must have matching dtypes. "
-            f"Got {db.dtype}, expected {self.dtype} for buffer: {db}")
-
     device_id_to_buffer = {_get_device(db).id: db for db in self._arrays}
 
     addressable_dev = self.sharding.addressable_devices
@@ -209,14 +234,12 @@ class ArrayImpl(basearray.Array):
                     "that are not present in the sharding.")
       raise ValueError(err_msg)
 
-    ss = self.sharding.shard_shape(self.shape)
-    for db in self._arrays:
-      if db.shape != ss:
-        raise ValueError(
-            f"Expected shard shape {ss} doesn't match the single device array "
-            f"shape {db.shape}. Shape of Array is "
-            f"{self.aval.str_short()} with sharding {self.sharding}")
-
+    _validate_shape_and_dtype_for_per_device_arrays(
+        self._arrays,
+        sharding=self.sharding,
+        aval=self.aval,
+        expected_shape=self.sharding.shard_shape(self.shape),
+    )
     # Rearrange arrays based on the device assignment.
     addressable_da = self.sharding._addressable_device_assignment
     self._arrays = [device_id_to_buffer[device.id] for device in addressable_da]
@@ -241,10 +264,21 @@ class ArrayImpl(basearray.Array):
   def sharding(self):
     return self._sharding
 
+  @property
+  def device(self):
+    self._check_if_deleted()
+    if isinstance(self.sharding, SingleDeviceSharding):
+      return list(self.sharding.device_set)[0]
+    return self.sharding
+
   @property
   def weak_type(self):
     return self.aval.weak_type
 
+  @property
+  def committed(self) -> bool:
+    return self._committed
+
   def __str__(self):
     return str(self._value)
 
@@ -319,18 +353,18 @@ class ArrayImpl(basearray.Array):
       except ValueError:
         arr_idx = None
       if arr_idx is not None:
-        a = self._arrays[arr_idx]
-        out = ArrayImpl(
-            a.aval, SingleDeviceSharding(_get_device(a)), [a], committed=False,
-            _skip_checks=True)
+        out = self._arrays[arr_idx]
+        sharding = SingleDeviceSharding(_get_device(out))
 
         if config.pmap_no_rank_reduction.value:
           # If cidx was the index of a single shard, then it corresponds to one
           # shard of the chunked dimension.
           dims = tuple(i for i, x in enumerate(cidx) if isinstance(x, int))
-          return lax.squeeze(out, dimensions=dims)
-        else:
-          return out
+          # Squeeze on committed arrays to avoid data movement to shard 0.
+          out = lax.squeeze(out, dimensions=dims)
+
+        return ArrayImpl(
+            out.aval, sharding, [out], committed=False, _skip_checks=True)
 
     return lax_numpy._rewriting_take(self, idx)
 
@@ -446,8 +480,7 @@ class ArrayImpl(basearray.Array):
 
   def __reduce__(self):
     fun, args, arr_state = self._value.__reduce__()
-    aval_state = {'weak_type': self.aval.weak_type,
-                  'named_shape': self.aval.named_shape}
+    aval_state = {'weak_type': self.aval.weak_type}
     return (_reconstruct_array, (fun, args, arr_state, aval_state))
 
   @use_cpp_method()
@@ -470,7 +503,7 @@ class ArrayImpl(basearray.Array):
     """Returns the total global on-device size of the array in bytes."""
     arr = self._arrays[0]
     per_shard_size = arr.on_device_size_in_bytes()
-    return per_shard_size * len(self.sharding.device_set)
+    return per_shard_size * self.sharding.num_devices
 
   def devices(self) -> set[Device]:
     self._check_if_deleted()
@@ -506,7 +539,8 @@ class ArrayImpl(basearray.Array):
     if self.is_deleted():
       return Layout(None, self.sharding)
     try:
-      return Layout(DeviceLocalLayout(self._pjrt_layout), self.sharding)
+      return Layout(DeviceLocalLayout.from_pjrt_layout(self._pjrt_layout),
+                    self.sharding)
     except xe.XlaRuntimeError as e:
       msg, *_ = e.args
       if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
@@ -614,8 +648,7 @@ class ArrayImpl(basearray.Array):
         npy_value[ind] = self._arrays[i]._single_device_array_to_np_array()
       self._npy_value = npy_value
       self._npy_value.flags.writeable = False
-    # https://docs.python.org/3/library/typing.html#typing.cast
-    return cast(np.ndarray, self._npy_value)
+    return self._npy_value
 
 
 # TODO(b/273265390): ideally we would write this as a decorator on the ArrayImpl
@@ -625,6 +658,14 @@ if not TYPE_CHECKING:
   ArrayImpl = use_cpp_class(xc.ArrayImpl)(ArrayImpl)
 
 
+def _get_shape_from_index(slc: Index, shape: Shape) -> Shape:
+  return tuple(
+      (s.stop or dim) - (s.start or 0)
+      for s, dim in safe_zip(slc, shape)
+      if isinstance(s, slice)  # If element is int, this dimension is reduced
+  )
+
+
 # explicitly set to be unhashable.
 setattr(ArrayImpl, "__hash__", None)
 setattr(ArrayImpl, "__array_priority__", 100)
@@ -653,7 +694,7 @@ def make_array_from_callback(
   Returns:
     A ``jax.Array`` via data fetched from ``data_callback``.
 
-  Example:
+  Examples:
 
     >>> import math
     >>> from jax.sharding import Mesh
@@ -684,51 +725,68 @@ def make_array_from_callback(
         f"sharding should be an instance of `jax.sharding`. Got {sharding} of"
         f" type {type(sharding)}")
 
+  def get_data(index: Index | None) -> ArrayImpl | np.ndarray:
+    # Perhaps cache on index here, then we can unify fully_replicated
+    # and non-fully_replicated cases below and become faster for
+    # partially replicated cases.
+    assert index is not None
+    r = data_callback(index)
+    if isinstance(r, core.Tracer):
+      raise errors.UnexpectedTracerError(
+          "jax.make_array_from_callback cannot be called within a traced"
+          " context."
+      )
+    # Value can be python scalar, resolve it into something with dtype.
+    return xla.canonicalize_dtype(r)
+
   if sharding.is_fully_replicated:
     devices = list(sharding._internal_device_list.addressable_device_list)  # type: ignore
-    per_device_values = [data_callback((slice(None),) * len(shape))] * len(devices)
+    # Only compute data once.
+    per_device_values = [get_data((slice(None),) * len(shape))] * len(devices)
   else:
     device_to_index_map = sharding.addressable_devices_indices_map(shape)
     devices = list(device_to_index_map.keys())
-    per_device_values = [data_callback(device_to_index_map[device])
-                         for device in devices]
-
-  if isinstance(per_device_values[0], core.Tracer):
-    raise errors.UnexpectedTracerError(
-        "jax.make_array_from_callback cannot be called within a traced context.")
-
-  first_value = xla.canonicalize_dtype(per_device_values[0])
-  aval = core.ShapedArray(shape, first_value.dtype, weak_type=False)
-
-  # first value can be numpy array, python scalar, etc.
-  if (sharding.is_fully_replicated and not isinstance(first_value, ArrayImpl)
-      and not dtypes.issubdtype(aval.dtype, dtypes.extended) and dll is None):
-    # Do this check outside because `batched_device_put` won't do these checks
-    # like ArrayImpl.
-    if shape != first_value.shape:
-      raise ValueError(
-            f"Expected shard shape {shape} doesn't match the single device "
-            f"array shape {first_value.shape}. Shape of Array is "
-            f"{aval.str_short()} with sharding {sharding}")
-    return pxla.batched_device_put(
-        aval, sharding, per_device_values, devices, committed=True)
+    per_device_values = [
+        get_data(device_to_index_map[device]) for device in devices
+    ]
 
+  first_value = per_device_values[0]
+  expected_dtype = first_value.dtype
+  expected_shape = sharding.shard_shape(shape)
+  aval = core.ShapedArray(shape, expected_dtype)
+  _validate_shape_and_dtype_for_per_device_arrays(
+      per_device_values,
+      expected_shape=expected_shape,
+      aval=aval,
+      sharding=sharding,
+  )
   if (isinstance(first_value, ArrayImpl)
       and first_value._committed
       and sharding.is_fully_replicated
       and first_value.is_fully_replicated
       and first_value.sharding._device_assignment == tuple(devices)
-      and (first_value.layout.device_local_layout ==
-           pxla._maybe_get_default_layout(Layout(dll, sharding), None, sharding, aval))):
+      and first_value.layout.device_local_layout == dll):
     return first_value
 
+  if dtypes.issubdtype(aval.dtype, dtypes.extended):
+    # TODO(yashkatariya): Can this also use batched_device_put?
+    arrays = api.device_put(per_device_values, devices)
+    return aval.dtype._rules.make_sharded_array(
+        aval, sharding, arrays, committed=True
+    )
+
   if dll is not None:
     devices = [Layout(dll, SingleDeviceSharding(d)) for d in devices]
-  arrays = api.device_put(per_device_values, devices)
-  if dtypes.issubdtype(aval.dtype, dtypes.extended):
-    return aval.dtype._rules.make_sharded_array(aval, sharding, arrays,
-                                                committed=True)
-  return ArrayImpl(aval, sharding, arrays, committed=True)
+    # pxla.batched_device_put doesn't support Layout... Take the slow route
+    arrays = api.device_put(per_device_values, devices)
+    return ArrayImpl(aval, sharding, arrays, committed=True)
+
+  if isinstance(first_value, ArrayImpl) and len(first_value.devices()) > 1:
+    # The output of the callback is already a sharded array, move it to
+    # to target device.
+    per_device_values = api.device_put(per_device_values, devices)
+
+  return pxla.batched_device_put(aval, sharding, per_device_values, devices)
 
 
 def make_array_from_process_local_data(
@@ -837,17 +895,20 @@ def make_array_from_process_local_data(
   setting it to (4, 4) in this case.
 
   Args:
-    sharding: sharding of the global tensor.
-    host_local_data: data on the host to be placed on local devices. Each
+    sharding: Sharding of the global array.
+    local_data: Data on the host to be placed on local devices. Each
       dimension should either match global_shape, or match
       num_addressable_indices(dim).
-    global_shape: the target shape of the global tensor. If None,
-      will infer from host_local_data and sharding.
+    global_shape: The target shape of the global array. If None,
+      will infer from local_data and sharding.
 
   Returns:
     Tensor that will have sharding=sharding and of shape global_shape.
   """
   # pyformat: enable
+  if xla_bridge.process_count() == 1:
+    return api.device_put(local_data, sharding)
+
   # TODO(sandler): consider supporting partially specified global_shape or
   # making local_to_global_shape available in the api.
   local_shape = local_data.shape
@@ -861,7 +922,6 @@ def make_array_from_process_local_data(
       )
   elif None in global_shape:
     raise ValueError(f"{global_shape=} has Nones. This is not supported.")
-  shard_shape = sharding.shard_shape(global_shape)
   full_dim = []
   for i, (data_dim, global_dim) in enumerate(
       zip(local_data.shape, global_shape)
@@ -877,6 +937,9 @@ def make_array_from_process_local_data(
             f"{global_shape=}."
         )
   addressable_shards = sharding.addressable_devices_indices_map(global_shape)
+  shard = next(iter(addressable_shards.values()))
+  assert shard is not None
+  shard_shape = _get_shape_from_index(shard, global_shape)
   slices_for_each_dim: list[list[int]] = [[] for _ in global_shape]
   for shard_index in addressable_shards.values():
     assert shard_index is not None
@@ -885,19 +948,20 @@ def make_array_from_process_local_data(
   for i in range(len(global_shape)):
     slices_for_each_dim[i] = sorted(set(slices_for_each_dim[i]))
 
-  def local_slice(i, slc):
+  @functools.lru_cache(maxsize=4096)
+  def local_slice(i, start):
     # Looks up the index of this slice in the list of slices for this dimension.
     # This will determine the slice in host_local_data
-    start = slices_for_each_dim[i].index(slc.start or 0) * shard_shape[i]
+    start = slices_for_each_dim[i].index(start or 0) * shard_shape[i]
     end = start + shard_shape[i]
     return slice(start, end)
 
   def cb(index: Index | None) -> ArrayLike:
     assert index is not None
-    data_slice = [
-        slc if full_dim[i] else local_slice(i, slc)
+    data_slice = (
+        slc if full_dim[i] else local_slice(i, slc.start)
         for i, slc in enumerate(index)
-    ]
+    )
     return local_data[tuple(data_slice)]
 
   return make_array_from_callback(global_shape, sharding, cb)
@@ -924,9 +988,6 @@ def make_array_from_single_device_arrays(
 
   Examples:
 
-    In this single-process example, we use ``make_array_from_single_device_arrays`` to create an
-    a global array.
-
     >>> import math
     >>> from jax.sharding import Mesh
     >>> from jax.sharding import PartitionSpec as P
@@ -947,40 +1008,8 @@ def make_array_from_single_device_arrays(
     >>> arr = jax.make_array_from_single_device_arrays(global_shape, sharding, arrays)
     >>> assert arr.shape == (8,8) # arr.shape is (8,8) regardless of jax.device_count()
 
-    When using multiple processes, a common data pipeline is to have data parallelism across devices,
-    with each device receiving at least one example. In this case, the following recipe will use
-    `make_array_from_single_device_arrays` to create a global jax.Array.
-
-    First, we create the per host data as Numpy arrays.
-
-    >>> sharding = jax.sharding.NamedSharding(mesh, P(('x', 'y'),))
-    >>> rows_per_device = 2
-    >>> feature_length = 32
-    >>> per_device_shape = (rows_per_device, feature_length)
-    >>> per_host_shape = (rows_per_device * len(mesh.local_devices), feature_length)
-    >>> per_host_generator = lambda : np.arange(np.prod(per_host_shape)).reshape(per_host_shape)
-    >>> per_host_data = per_host_generator() # replace with your own per-host data pipeline that outputs numpy arrays
-
-    Second, we put the Numpy data onto the local devices as single device Jax Arrays. Then we call
-    make_array_from_single_device_arrays to make the global Array.
-
-    >>> global_shape = (rows_per_device * len(sharding.device_set), ) + per_device_shape[1:]
-    >>> per_device_data = np.split(per_host_data, len(mesh.local_devices), axis = 0) # per device data, but on host
-    >>> per_device_data_on_device = jax.device_put(per_device_data, mesh.local_devices) # per device data, now on device
-    >>> output_global_array = jax.make_array_from_single_device_arrays(global_shape, sharding, per_device_data_on_device)
-    ...
-    >>> assert output_global_array.addressable_data(0).shape == per_device_shape
-    >>> assert output_global_array.shape == global_shape
-
-    When using tensor parallelism (equivalent to sharding across both rows and columns in the
-    above example), the above example doesn't generate the data in the sharding that you plan
-    to consume it with. The most common fix is to simply load the data in this data parallel sharding
-    and have the reshard happen automatically within the downstream jitted function.
-    Depending on your use case, you might prefer to directly load sharded data, something that
-    ``make_array_from_single_device_arrays`` can do but will depend on your data loading pipeline
-    also loading in the matching sharding. Loading in a data parallel format is typically
-    fully satisfactory for data loading for LLM use cases.
-
+  For cases where you have a local array and want to convert it to a global
+  jax.Array, use ``jax.make_array_from_process_local_data``.
   """
   # All input arrays should be committed. Checking it is expensive on
   # single-controller systems.
@@ -996,17 +1025,35 @@ def make_array_from_single_device_arrays(
   return ArrayImpl(aval, sharding, cast(Sequence[ArrayImpl], arrays),
                    committed=True)
 
-
-core.pytype_aval_mappings[ArrayImpl] = abstract_arrays.canonical_concrete_aval
-xla.pytype_aval_mappings[ArrayImpl] = op.attrgetter('aval')
 xla.canonicalize_dtype_handlers[ArrayImpl] = pxla.identity
-api_util._shaped_abstractify_handlers[ArrayImpl] = op.attrgetter('aval')
+
+def _get_aval_array(self):
+  if config.sharding_in_types.value and isinstance(self.sharding, NamedSharding):
+    return self.aval.update(sharding=NamedSharding(
+        self.sharding.mesh.abstract_mesh,
+        self.sharding.spec._normalized_spec(self.ndim)))
+  else:
+    return self.aval
+
+core.pytype_aval_mappings[ArrayImpl] = _get_aval_array
+
 # TODO(jakevdp) replace this with true inheritance at the C++ level.
 basearray.Array.register(ArrayImpl)
 
 
 def _array_mlir_constant_handler(val):
-  return mlir.ir_constants(val._value)
+  try:
+    return mlir.ir_constant(val._value)
+  except RuntimeError as e:
+    # TODO(yashkatariya): Ideally we would catch a custom exception from
+    # `_value` function in ArrayImpl instead of checking the error string.
+    if 'Fetching value for `jax.Array` that spans non-addressable' in str(e):
+      raise RuntimeError(
+          "Closing over jax.Array that spans non-addressable (non process"
+          " local) devices is not allowed. Please pass such arrays as arguments"
+          f" to the function. Got jax.Array: {val.aval.str_short()}") from e
+    raise
+
 mlir.register_constant_handler(ArrayImpl, _array_mlir_constant_handler)
 
 
@@ -1046,29 +1093,24 @@ def shard_device_array(x, devices, indices, sharding):
     shards = [x] * len(devices)
   else:
     shards = x._multi_slice(start_indices, limit_indices, removed_dims)
-  aval = api_util.shaped_abstractify(x)
+  aval = core.shaped_abstractify(x)
   return pxla.batched_device_put(aval, sharding, shards, devices)
 
-def _hashable_index(idx):
-  return tree_util.tree_map(
-      lambda x: (x.start, x.stop) if type(x) == slice else x, idx)
-
 
 def shard_sharded_device_array_slow_path(x, devices, indices, sharding):
   candidates = defaultdict(list)
   bufs = [buf.data for buf in x.addressable_shards]
   arr_indices = tuple(x.sharding.devices_indices_map(x.shape).values())
   for buf, idx in safe_zip(bufs, arr_indices):
-    candidates[_hashable_index(idx)].append(buf)
+    candidates[hashed_index(idx)].append(buf)
 
   bufs = []
   for idx, device in safe_zip(indices, devices):
     # Look up all buffers that contain the correct slice of the logical array.
-    candidates_list = candidates[_hashable_index(idx)]
+    candidates_list = candidates[hashed_index(idx)]
     if not candidates_list:
-      # This array isn't sharded correctly. Reshard it via host roundtrip.
-      # TODO(skye): more efficient reshard?
-      return pxla.shard_arg(x._value, sharding, canonicalize=False)
+      return pxla.shard_args([sharding], [None], [None], [x._value],
+                             canonicalize=False)[0]
     # Try to find a candidate buffer already on the correct device,
     # otherwise copy one of them.
     for buf in candidates_list:
@@ -1076,38 +1118,63 @@ def shard_sharded_device_array_slow_path(x, devices, indices, sharding):
         bufs.append(buf)
         break
     else:
-      bufs.append(buf)
-
+      bufs.append(candidates_list[-1])
   return pxla.batched_device_put(x.aval, sharding, bufs, devices)
 
 
-@functools.lru_cache(maxsize=4096)
+@cache(max_size=4096, trace_context_in_key=False)
 def _sharding_indices_and_eq(src_sharding, shape, dst_sharding):
   src_indices = src_sharding.addressable_devices_indices_map(shape).values()
   dst_indices = dst_sharding.addressable_devices_indices_map(shape).values()
   return dst_indices, tuple(src_indices) == tuple(dst_indices)
 
 
-def _array_shard_arg(x, sharding):
-  x._check_if_deleted()
+def _array_shard_arg(xs, shardings, layouts, copy_semantics):
+  util.test_event("_array_shard_arg")
+  results = []
+  batch_xs, batch_devs, batch_shardings, batch_indices = [], [], [], []
+  batch_cs = []
 
-  indices, same_indices = _sharding_indices_and_eq(x.sharding, x.shape, sharding)
-  if not x.is_fully_addressable:
-    if same_indices:
-      return x
-    else:
-      raise NotImplementedError(
-          "Cannot reshard an input that is not fully addressable")
-  else:
-    devices = sharding._addressable_device_assignment
-    if same_indices:
-      return xc.copy_array_to_devices_with_sharding(x, list(devices), sharding)
-    # Resharding starts here:
-    if dispatch.is_single_device_sharding(x.sharding):
-      return shard_device_array(x, devices, indices, sharding)
-    else:
-      return shard_sharded_device_array_slow_path(x, devices, indices, sharding)
+  for i, (x, sharding, layout, cs) in enumerate(
+      safe_zip(xs, shardings, layouts, copy_semantics)):
+    x._check_if_deleted()
+    indices, same_indices = _sharding_indices_and_eq(x.sharding, x.shape, sharding)
+    same_layout = (True if layout is None else
+                   x.layout.device_local_layout == layout)
 
+    if not x.is_fully_addressable:
+      if same_indices and same_layout:
+        results.append(x)
+      else:
+        raise NotImplementedError(
+            "Cannot reshard an input that is not fully addressable")
+    else:
+      devices = sharding._addressable_device_assignment
+      if same_indices and same_layout:
+        # Add a placeholder result that will be filled in later.
+        results.append(None)
+        # Accumulate arguments to `batched_copy_array_to_devices_with_sharding`.
+        batch_xs.append(x)
+        batch_devs.append(list(devices))
+        batch_shardings.append(sharding)
+        batch_indices.append(i)
+        batch_cs.append(cs)
+      # Resharding starts here:
+      elif not same_layout:
+        results.append(api.device_put(x, Layout(layout, sharding)))
+      elif dispatch.is_single_device_sharding(x.sharding):
+        results.append(shard_device_array(x, devices, indices, sharding))
+      else:
+        results.append(
+            shard_sharded_device_array_slow_path(x, devices, indices, sharding))
+
+  util.test_event("batched_copy_array")
+  copy_outs = xc.batched_copy_array_to_devices_with_sharding(
+      batch_xs, batch_devs, batch_shardings, batch_cs)
+  for i, copy_out in safe_zip(batch_indices, copy_outs):
+    assert results[i] is None
+    results[i] = copy_out
+  return results
 pxla.shard_arg_handlers[ArrayImpl] = _array_shard_arg
 
 
@@ -1121,7 +1188,6 @@ def _array_global_result_handler(global_aval, out_sharding, committed):
       global_aval, out_sharding, committed=committed, _skip_checks=True
   )
 pxla.global_result_handlers[core.ShapedArray] = _array_global_result_handler
-pxla.global_result_handlers[core.ConcreteArray] = _array_global_result_handler
 
 # Only used for Arrays that come out of pmap.
 def _array_local_result_handler(aval, sharding, indices):
@@ -1134,13 +1200,17 @@ def _array_local_result_handler(aval, sharding, indices):
       aval, sharding, committed=True, _skip_checks=True
   )
 pxla.local_result_handlers[core.ShapedArray] = _array_local_result_handler
-pxla.local_result_handlers[core.ConcreteArray] = _array_local_result_handler
 
 
 # Token handlers
 
-def _token_shard_arg(x, sharding):
-  return _array_shard_arg(x._buf, sharding)
+def _token_shard_arg(xs, shardings, layouts, copy_semantics):
+  results = []
+  for x, sharding, layout in safe_zip(xs, shardings, layouts):
+    x.block_until_ready()
+    x = np.array([], dtype=bool)
+    results.append(api.device_put(x, Layout(layout, sharding)))
+  return results
 pxla.shard_arg_handlers[core.Token] = _token_shard_arg
 
 
