@@ -13,6 +13,7 @@
 # limitations under the License.
 
 from dataclasses import dataclass
+import functools
 from types import ModuleType
 import warnings
 
@@ -45,6 +46,7 @@ import warnings
 #   )
 # del typing
 def deprecation_getattr(module, deprecations):
+  @functools.cache
   def getattr(name):
     if name in deprecations:
       message, fn = deprecations[name]
@@ -66,6 +68,13 @@ def accelerate_getattr_deprecation(module: ModuleType, name: str) -> None:
   message, _ = module._deprecations[name]
   module._deprecations[name] = (message, None)
 
+def is_accelerated_attribute(module: ModuleType, name: str) -> bool:
+  """Returns true if given name is accelerated.
+
+  Raises an error if name is not a deprecated attribute in module.
+  """
+  return module._deprecations[name][1] is None
+
 # The following mechanism is a separate one, for registering and
 # accelerating deprecations that are not imports (for example, deprecations
 # of a function argument).
@@ -101,3 +110,29 @@ def is_accelerated(deprecation_id: str) -> bool:
   if deprecation_id not in _registered_deprecations:
     raise ValueError(f"{deprecation_id=!r} not registered.")
   return _registered_deprecations[deprecation_id].accelerated
+
+
+def warn(deprecation_id: str, message: str, stacklevel: int) -> None:
+  """Warns about a deprecation, or errors if the deprecation is accelerated."""
+  if is_accelerated(deprecation_id):
+    raise ValueError(message)
+  else:
+    warnings.warn(message, category=DeprecationWarning,
+                  stacklevel=stacklevel + 1)
+
+
+# Register a number of deprecations: we do this here to ensure they're
+# always registered by the time `accelerate` and `is_acelerated` are called.
+register('jax-aval-named-shape')
+register('jax-dlpack-import-legacy')
+register('jax-nn-one-hot-float-input')
+register("jax-numpy-astype-complex-to-real")
+register("jax-numpy-array-none")
+register('jax-numpy-clip-args')
+register('jax-numpy-linalg-matrix_rank-tol')
+register('jax-numpy-linalg-pinv-rcond')
+register('jax-numpy-quantile-interpolation')
+register('jax-numpy-reduction-non-boolean-where')
+register('jax-numpy-trimzeros-not-1d-array')
+register('pallas-gpu-triton')
+register('jax-scipy-special-sph-harm')
