@@ -13,28 +13,13 @@
 # limitations under the License.
 """Shape polymorphism support.
 
-We introduce a set of dimension variables at the top-level of a `jit` function.
-They are introduced implicitly by way of specifying for each dimension of each
-argument a symbolic dimension expression in terms of some dimension variables.
-All dimension variables are assumed to range over integers greater or equal to 1.
-
-Symbolic dimensions overload some integer operations, such as
-add, multiply, divide, equality, etc. The JAX NumPy layer and the LAX layers have been
-touched up to be sensitive to handling shapes that contain symbolic dimensions.
-This enables many JAX programs to be traced with symbolic dimensions
-in some dimensions. A priority has been to enable the batch
-dimension in neural network examples to be polymorphic.
-
-This was built initially for jax2tf, but it is now
-independent of TF. The best documentation at the moment is in the
-jax2tf.convert docstring, and the
-[README](https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md).
+See documentation at https://jax.readthedocs.io/en/latest/export/shape_poly.html.
 """
 
 from __future__ import annotations
 
 import enum
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import dataclasses
 from enum import Enum
 import functools
@@ -43,14 +28,13 @@ import io
 import copy
 import operator as op
 import tokenize
-from typing import Any, Callable, Union, overload
+from typing import Any, Union, overload
 import warnings
 
 import numpy as np
 import opt_einsum
 
 import jax
-from jax.interpreters import xla
 
 from jax._src import config
 from jax._src import core
@@ -86,7 +70,7 @@ This error arises for comparison operations with shapes that
 are non-constant, and the result of the operation cannot be represented as
 a boolean value for all values of the symbolic dimensions involved.
 
-Please see https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#comparison-of-symbolic-dimensions-is-partially-supported
+Please see https://jax.readthedocs.io/en/latest/export/shape_poly.html#comparison-of-symbolic-dimensions-is-partially-supported
 for more details.
 """
 
@@ -95,6 +79,8 @@ for more details.
     # https://github.com/python/mypy/issues/5887
     super().__init__(error_msg)
 
+class UnexpectedDimVar(Exception):
+  pass
 
 class Comparator(Enum):
   EQ = 1
@@ -102,12 +88,17 @@ class Comparator(Enum):
 
 @dataclasses.dataclass(frozen=True)
 class _SymbolicConstraint:
+  # Either e1 == e2 if cmp == Comparator.EQ else e1 >= e2
   cmp: Comparator
   debug_str: str  # The form in which the user expressed it, for error messages
-  diff: _DimExpr  # For GEQ: diff >= 0, and for EQ: diff == 0
+  # e1, e2, and diff == e1 - e2, are normalized w.r.t. previous constraints only
+  e1: DimSize
+  e2: DimSize
+  # we pre-compute diff to avoid having the normalization rule kick in later.
+  diff: DimSize
 
   def __repr__(self):
-    return f"Constraint({self.debug_str}: {self.diff})"
+    return f"Constraint({self.debug_str})"
 
 
 class _DimFactor:
@@ -137,8 +128,6 @@ class _DimFactor:
   MOD = "mod"
   MAX = "max"
   MIN = "min"
-  NON_NEGATIVE = "non_negative"  # The max of the operand and 0. Replaced with
-                                 # max but kept here for backwards compatibility.
 
   __slots__ = ["var", "operation", "operands", "_hash", "_size"]
 
@@ -224,15 +213,22 @@ class _DimFactor:
     """Lexicographic comparison"""
     return self._syntactic_cmp(other) >= 0
 
-  def evaluate(self, env: DimVarEnv):
+  def evaluate(self, env: DimVarEnv, scope: SymbolicScope):
     if self.var is not None:
       try:
         return env[self.var]
       except KeyError:
+        # Perhaps there is a normalization rule for this variable
+        normalized_var = _DimExpr._from_var(self.var, scope)
+        if core.is_constant_dim(normalized_var):
+          return normalized_var
+        non_trivial_normalization = (v1 := normalized_var._to_var()) is None or v1 != self.var  # type: ignore
+        if non_trivial_normalization:
+          return normalized_var._evaluate(env)  # type: ignore
         err_msg = (
-            f"Encountered dimension variable '{self.var}' that is not appearing in the shapes of the used function arguments.\n"
-            "Please see https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#dimension-variables-must-be-solvable-from-the-input-shapes for more details.")
-        raise KeyError(err_msg)
+            f"Encountered dimension variable '{self.var}' that is not appearing in the shapes of the function arguments.\n"
+            "Please see https://jax.readthedocs.io/en/latest/export/shape_poly.html#dimension-variables-must-be-solvable-from-the-input-shapes for more details.")
+        raise UnexpectedDimVar(err_msg)
     else:
       operand_values = [opnd._evaluate(env) for opnd in self.operands]
       if self.operation == _DimFactor.FLOORDIV:
@@ -385,11 +381,11 @@ class _DimTerm:
         raise InconclusiveDimensionOperation(f"Cannot divide {self} by {divisor}.")
     return _DimTerm(new_factors)
 
-  def evaluate(self, env: DimVarEnv):
+  def evaluate(self, env: DimVarEnv, scope: SymbolicScope):
     prod = lambda xs: functools.reduce(_evaluate_multiply, xs) if xs else core.dim_constant(1)
     def pow_opt(v, p: int):
       return v if p == 1 else prod([v] * p)
-    return prod([pow_opt(f.evaluate(env), exp) for f, exp in self._factors])
+    return prod([pow_opt(f.evaluate(env, scope), exp) for f, exp in self._factors])
 
   def __deepcopy__(self, memo):
     return _DimTerm(copy.deepcopy(self._factors, memo))
@@ -419,7 +415,7 @@ class _DimExpr:
   def __init__(self, sorted_terms: SortedTerms,
                scope: SymbolicScope):
     # Do not construct _DimExpr directly, unless you are sure that `terms` is
-    # normalized; Use _DimExpr.normalize.
+    # normalized; Use _DimExpr._normalize_sorted_terms.
     self._sorted_terms = tuple(sorted_terms) or ((_DimTerm_one, 0),)
     self._scope = scope
     self._hash = None
@@ -441,17 +437,12 @@ class _DimExpr:
     return _DimExpr._normalize_sorted_terms(((t, t_k),), scope)
 
   @staticmethod
-  def _from_var(v: str, scope: SymbolicScope) -> _DimExpr:
-    return _DimExpr(((_DimTerm.from_var(v), 1),), scope)
+  def _from_var(v: str, scope: SymbolicScope) -> DimSize:
+    return _DimExpr._normalize_sorted_terms(((_DimTerm.from_var(v), 1),), scope)
 
   @staticmethod
   def _from_operation(operation: str, *operands: DimSize,
                       scope: SymbolicScope) -> DimSize:
-    if operation == _DimFactor.NON_NEGATIVE:  # For parsing, for backwards compatibility
-      return _DimExpr._from_term(
-          _DimTerm.from_operation(_DimFactor.MAX, *operands, 0,
-                                  scope=scope), 1,
-          scope=scope)
     return _DimExpr._from_term(
         _DimTerm.from_operation(operation, *operands, scope=scope), 1,
         scope=scope)
@@ -490,8 +481,9 @@ class _DimExpr:
   def _normalize_term(t: _DimTerm, t_k: int,
                       scope: SymbolicScope) -> Sequence[tuple[_DimTerm, int]]:
     # If (t, t_k) is among the scope normalization rules, then return
-    # a list of updates to apply to the expression containing (t, t_k).
-    # Returns empty sequence if no normalizations are necessary.
+    # a list of `term * coefficient` to add to the expression containing (t, t_k).
+    # Returns the empty sequence if no normalizations are necessary.
+    if not scope._normalization_rules: return []
     updates = []
     after, t_k_after = scope._normalization_rules.get(t, (None, 0))
     if after is not None and t_k % t_k_after == 0:
@@ -662,7 +654,7 @@ class _DimExpr:
       # Here we really ought to raise InconclusiveDimensionOperation, but __eq__
       # cannot raise exceptions, because it is used indirectly when hashing.
       # So, we say that the expressions are disequal, which is really unsound.
-      # See https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#comparison-of-symbolic-dimensions-is-partially-supported
+      # See https://jax.readthedocs.io/en/latest/export/shape_poly.html#comparison-of-symbolic-dimensions-is-partially-supported
       return False
 
     return diff == 0
@@ -767,13 +759,23 @@ class _DimExpr:
       return _DimExpr._linear_combination(self, other, 0, 0, self.scope)
     return _ensure_poly(other, "mul", self.scope).__mul__(self)
 
-  def __pow__(self, power, modulo=None):
-    assert modulo is None
-    try:
-      power = int(power)
-    except:
-      raise InconclusiveDimensionOperation(f"Symbolic dimension cannot be raised to non-integer power '{self}' ^ '{power}'")
-    return functools.reduce(op.mul, [self] * power)
+  def __pow__(self, power: core.DimSize, modulo=None):
+    if modulo is not None:
+      raise NotImplementedError("__pow__ modulo not implemented")
+    if is_symbolic_dim(power):
+      return power.__rpow__(self)  # type: ignore
+    if power != int(power):
+      raise ValueError(f"Symbolic dimension cannot be raised to non-integer powers: '{self}' ** '{power}'")
+    if power >= 0:
+      return functools.reduce(op.mul, [self] * power, 1)
+    # We don't support negative powers, because JAX does not allow negative
+    # powers for integers
+    raise ValueError(f"Symbolic dimension cannot be raised to negative powers: '{self}' ** '{power}'")
+
+  def __rpow__(self, other, modulo=None):
+    if modulo is not None:
+      raise NotImplementedError("__rpow__ modulo not implemented")
+    return self.__jax_array__().__rpow__(other)
 
   def __floordiv__(self, divisor):
     if isinstance(divisor, core.Tracer) or not _convertible_to_poly(divisor):
@@ -839,7 +841,7 @@ class _DimExpr:
       # Here we really ought to raise InconclusiveDimensionOperation, but __eq__
       # cannot raise exceptions, because it is used indirectly when hashing.
       # So, we say that the expressions are disequal, which is really unsound.
-      # See https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#comparison-of-symbolic-dimensions-is-partially-supported
+      # See https://jax.readthedocs.io/en/latest/export/shape_poly.html#comparison-of-symbolic-dimensions-is-partially-supported
       return False
 
     return diff == 0
@@ -901,8 +903,10 @@ class _DimExpr:
       if config.enable_checks.value:
         v1 = divisor * quotient
         v2 = v1 + remainder
-        assert self == v2, (self, v2, type(self), type(v2))
-        assert self == divisor * quotient + remainder, (self, divisor, quotient, remainder)
+        assert self == _ensure_poly(v2, "check", self.scope), (
+            self, v2, type(self), type(v2))
+        assert self == _ensure_poly(divisor * quotient + remainder, "test", self.scope), (
+            self, divisor, quotient, remainder)
       return quotient, remainder
     except InconclusiveDimensionOperation:
       return (_DimExpr._from_operation(_DimFactor.FLOORDIV, self, divisor,
@@ -912,7 +916,7 @@ class _DimExpr:
 
   def _evaluate(self, env: DimVarEnv):
     # Evaluates as a value of dtype=core.dim_value_dtype()
-    terms = [_evaluate_multiply(t.evaluate(env), core.dim_constant(t_k))
+    terms = [_evaluate_multiply(t.evaluate(env, self.scope), core.dim_constant(t_k))
              for t, t_k in self._sorted_terms]
     return functools.reduce(_evaluate_add, terms) if len(terms) > 1 else terms[0]
 
@@ -982,7 +986,7 @@ class SymbolicScope:
 
   Holds the constraints on symbolic expressions.
 
-  See [the README](https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#user-specified-symbolic-constraints)
+  See [the README](https://jax.readthedocs.io/en/latest/export/shape_poly.html#user-specified-symbolic-constraints)
   for more details.
 
   Args:
@@ -1052,29 +1056,52 @@ class SymbolicScope:
     if cmp == Comparator.GEQ and not is_geq:
       e1, e2 = e2, e1
 
-    diff = e1 - e2
-    if (diff_const := _DimExpr._to_constant(diff)) is not None:
-      if ((cmp == Comparator.EQ and diff_const != 0) or
-          (cmp == Comparator.GEQ and diff_const < 0)):
-        raise ValueError(f"Unsatisfiable explicit constraint: {c_str}")
+    # Compute e1 - e2 before we add to normalization rules
+    constr = _SymbolicConstraint(debug_str=c_str, cmp=cmp, e1=e1, e2=e2,
+                                 diff=e1 - e2)
+    self._process_explicit_constraint(constr)
+
+  def _process_explicit_constraint(self, constr: _SymbolicConstraint):
+    if (diff_const := _DimExpr._to_constant(constr.diff)) is not None:
+      if ((constr.cmp == Comparator.EQ and diff_const != 0) or
+          (constr.cmp == Comparator.GEQ and diff_const < 0)):
+        raise ValueError(f"Unsatisfiable explicit constraint: {constr.debug_str}")
       return
 
-    constr = _SymbolicConstraint(debug_str=c_str, cmp=cmp, diff=diff)  # type: ignore[arg-type]
-    self._explicit_constraints.append(constr)
-    if cmp == Comparator.EQ:
-      if not isinstance(e1, _DimExpr):
+    if constr.cmp == Comparator.EQ:
+      if not isinstance(constr.e1, _DimExpr):
         raise ValueError("Invalid equality constraint: {e1} == {e2}. "
                          "The left-hand-side must be of the form `term * coefficient`.")
-      (before, before_k), *rest = e1._sorted_terms
+      (before, before_k), *rest = constr.e1._sorted_terms
       if rest:
         raise ValueError("Invalid equality constraint: {e1} == {e2}. "
                          "The left-hand-side must be of the form `term * coefficient`.")
 
-      after = _ensure_poly(e2, "parse_constraint", e1.scope)  # type: ignore[name-error,unused-ignore]
+      after = _ensure_poly(constr.e2, "parse_constraint", constr.e1.scope)  # type: ignore[name-error,unused-ignore]
       if before in self._normalization_rules:
         raise NotImplementedError(
             f"Found multiple equality constraints with the same left-hand-side: {before}")
       self._normalization_rules[before] = (after, before_k)
+      # Look for constraints of the form mod(before_e1, before_k2) * 1 == 0
+      if (before_k == 1 and
+          isinstance(constr.e2, int) and constr.e2 == 0 and
+          (before_f := before.to_factor()) and
+          before_f.operation == _DimFactor.MOD and
+          (before_k2 := _DimExpr._to_constant(before_f.operands[1])) is not None):
+        # Add before_k2*floordiv(before_e1, before_k2) == before_e1
+        k_times_floordiv = _DimExpr._from_term(
+            _DimTerm.from_operation(
+                _DimFactor.FLOORDIV, *before_f.operands, scope=constr.e1.scope),
+            before_k2, scope=constr.e1.scope)
+        before_e1 = before_f.operands[0]
+        self._process_explicit_constraint(
+            _SymbolicConstraint(cmp=Comparator.EQ,
+                                e1=k_times_floordiv, e2=before_e1,
+                                diff=k_times_floordiv - before_e1,
+                                debug_str=f"{k_times_floordiv} == {before_e1}")
+        )
+
+    self._explicit_constraints.append(constr)
 
   def _check_same_scope(self, other: _DimExpr,
                         when: str = "",
@@ -1085,7 +1112,7 @@ class SymbolicScope:
           f"Invalid mixing of symbolic scopes {when}.\n"
           f"Expected {self_descr}scope {self}\n"
           f"and found for '{other}' ({other_descr}) scope {other.scope}\n"
-          f"See https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#user-specified-symbolic-constraints.")
+          f"See https://jax.readthedocs.io/en/latest/export/shape_poly.html#user-specified-symbolic-constraints.")
 
   def _clear_caches(self):
     self._bounds_cache.clear()
@@ -1170,7 +1197,6 @@ def _geq_decision(e1: DimSize, e2: DimSize, cmp_str: Callable[[], str]) -> bool:
       f"Symbolic dimension comparison {cmp_str()} is inconclusive.{describe_scope}")
 
 core.pytype_aval_mappings[_DimExpr] = _DimExpr._get_aval
-xla.pytype_aval_mappings[_DimExpr] = _DimExpr._get_aval
 dtypes._weak_types.append(_DimExpr)
 
 def _convertible_to_int(p: DimSize) -> bool:
@@ -1194,14 +1220,10 @@ def _convertible_to_poly(p: DimSize) -> bool:
   return isinstance(p, _DimExpr) or _convertible_to_int(p)
 
 def is_symbolic_dim(p: DimSize) -> bool:
+  """Checks if a dimension is symbolic.
+  """
   return isinstance(p, _DimExpr)
 
-def is_poly_dim(p: DimSize) -> bool:
-  # TODO: deprecated January 2024, remove June 2024.
-  warnings.warn("is_poly_dim is deprecated, use export.is_symbolic_dim",
-                DeprecationWarning, stacklevel=2)
-  return is_symbolic_dim(p)
-
 dtypes.python_scalar_dtypes[_DimExpr] = dtypes.python_scalar_dtypes[int]
 
 def _einsum_contract_path(*operands, **kwargs):
@@ -1311,7 +1333,7 @@ dim_as_value_p.def_abstract_eval(lambda dim: core.dim_value_aval())
 def dim_as_value_impl(dim: DimSize):
   raise NotImplementedError(
       "Evaluation rule for 'dim_as_value' is not implemented. "
-      "It seems that you are using shape polymorphism outside jax2tf.")
+      "It seems that you are using shape polymorphism outside jax.export.")
 
 dim_as_value_p.def_impl(dim_as_value_impl)
 def _dim_as_value(dim: DimSize):
@@ -1360,26 +1382,36 @@ def symbolic_shape(shape_spec: str | None,
                    scope: SymbolicScope | None = None,
                    like: Sequence[int | None] | None = None
                    ) -> Sequence[DimSize]:
-  """Constructs a jax.ShapeDtypeStruct with polymorphic shapes.
+  """Constructs a symbolic shape from a string representation.
+
+  See https://jax.readthedocs.io/en/latest/export/shape_poly.html for examples.
 
   Args:
     shape_spec: a symbolic shape specification. None stands for "...".
+      A shape specification is the string representation of a tuple (the
+      parentheses are optional) with comma-separated dimension expressions.
+      A dimension expression can be either: an integer constant,
+      a dimension variable (alphanumeric
+      starting with a letter), e1 + e2, e1 - e2, e1 * e2, floordiv(e1, e2),
+      mod(e1, e2), max(e1, e2), or min(e1, e2).
+    constraints: a sequence of constraints on symbolic dimension expressions, of
+      the form `e1 >= e2` or `e1 <= e2`, or `e1 == e2`.
+      See [the documentation](https://jax.readthedocs.io/en/latest/export/shape_poly.html#user-specified-symbolic-constraints)
+      for usage.
+    scope: optionally, you can specify that the parsed symbolic expressions
+      be created in the given scope. If this is missing, then a new
+      `SymbolicScope` is created with the given `constraints`.
+      You cannot specify both a `scope` and `constraints`.
+      See [the documentation](https://jax.readthedocs.io/en/latest/export/shape_poly.html#user-specified-symbolic-constraints)
+      for usage.
     like: when `shape_spec` contains placeholders ("_", "..."), use this
       shape to fill in the placeholders.
       The dimensions of `like` that are used for filling
-      must be known (not `None`). If a dimension in `like` is known and
+      must be not `None`. If a dimension in `like` is not `None` and
       the corresponding dimension in `shape_spec` is a constant then they
       must be equal.
-    scope: optionally, you can specify that the parsed symbolic expressions
-      be created in a given scope. You cannot specify `constraints` in this case.
-    constraints: a sequence of constraints on symbolic dimension expressions, of
-      the form `e1 >= e2` or `e1 <= e2`. This is used to create a new SymbolicScope
-      shared by all symbolic expressions created.
-      See [the README](https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#user-specified-symbolic-constraints)
-      for more details.
 
-  Returns: a jax.ShapeDTypeStruct with shapes that may contain symbolic
-      expressions involving dimension variables.
+  Returns: a tuple with integers or symbolic expressions involving dimension variables.
   """
   shape_spec_repr = repr(shape_spec)
   if shape_spec is None:
@@ -1398,43 +1430,34 @@ def symbolic_shape(shape_spec: str | None,
 
 def symbolic_args_specs(
     args,  # pytree of arguments
-    polymorphic_shapes,  # prefix pytree of strings
-    symbolic_scope: SymbolicScope | None = None,
-    symbolic_constraints: Sequence[str] = (),
+    shapes_specs,  # prefix pytree of strings
+    constraints: Sequence[str] = (),
+    scope: SymbolicScope | None = None,
 ):
   """Constructs a pytree of jax.ShapeDtypeSpec arguments specs for `export`.
 
-  Note that this function does not ensure that the provided `args` shapes
-  are compatible with `polymorphic_shapes`. The `.shape` of the `args` are
-  used only to fill-in placeholders from `polymorphic_shapes`.
-
-  See docstring of `symbolic_shape` and
-  [the README](https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#shape-polymorphic-conversion)
-  for more details.
+  See the documentation of :func:`jax.export.symbolic_shape` and
+  the [shape polymorphism documentation](https://jax.readthedocs.io/en/latest/export/shape_poly.html) for details.
 
   Args:
     args: a pytree of arguments. These can be jax.Array, or jax.ShapeDTypeSpec.
-      This is used to learn the pytree structure of the arguments, their dtypes,
-      and to fill-in the actual shapes where the `polymorphic_shapes` contains
+      They are used to learn the pytree structure of the arguments, their dtypes,
+      and to fill-in the actual shapes where the `shapes_specs` contains
       placeholders. Note that only the shape dimensions for which
-      `polymorphic_shapes` is a placeholder are used from `args`.
-      The unused dimensions can be `None`, which jax2tf uses when the TF
-      shapes are not known.
-    polymorphic_shapes: should be `None` (all arguments have static shapes),
-      a single string (applies to all arguments), or a pytree matching a prefix
+      `shapes_specs` is a placeholder are used from `args`.
+    shapes_specs: should be `None` (all arguments have static shapes),
+      a single string (see `shape_spec` for :func:`jax.export.symbolic_shape`;
+      applies to all arguments), or a pytree matching a prefix
       of the `args`.
       See [how optional parameters are matched to
       arguments](https://jax.readthedocs.io/en/latest/pytrees.html#applying-optional-parameters-to-pytrees).
-    symbolic_scope: optionally, you can specify that the parsed symbolic expressions
-      be created in a given scope. You cannot specify `symbolic_constraints` in this case.
-    symbolic_constraints: a sequence of constraints on symbolic dimension expressions, of
-      the form `e1 >= e2` or `e1 <= e2`. This is used to create a new SymbolicScope
-      shared by all symbolic expressions created.
-      See more details at https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#user-specified-symbolic-constraints.
+    constraints: as for :func:`jax.export.symbolic_shape`.
+    scope: as for :func:`jax.export.symbolic_shape`.
 
   Returns: a pytree of jax.ShapeDTypeStruct matching the `args` with the shapes
-    replaced with symbolic dimensions as specified by `polymorphic_shapes`.
+    replaced with symbolic dimensions as specified by `shapes_specs`.
   """
+  polymorphic_shapes = shapes_specs
   args_flat, args_tree = tree_util.tree_flatten(args)
 
   shapes_and_dtypes = tuple(map(shape_and_dtype_jax_array, args_flat))
@@ -1454,15 +1477,15 @@ def symbolic_args_specs(
     e, *_ = tree_util.prefix_errors(
         polymorphic_shapes_, args,
         is_leaf=lambda x: x is None)
-    raise e("jax_export polymorphic_shapes") from None
+    raise e("export.symbolic_args_specs shapes_specs") from None
 
   # Now add in the polymorphic shapes
-  if symbolic_scope is None:
-    symbolic_scope = SymbolicScope(symbolic_constraints)
-  elif symbolic_constraints:
-    raise ValueError("Cannot have both `symbolic_scope` and `symbolic_constraints`")
+  if scope is None:
+    scope = SymbolicScope(constraints)
+  elif constraints:
+    raise ValueError("Cannot use both `scope` and `constraints`")
   args_specs_flat = (
-      jax.ShapeDtypeStruct(symbolic_shape(spec, like=s, scope=symbolic_scope), t)
+      jax.ShapeDtypeStruct(symbolic_shape(spec, like=s, scope=scope), t)
       for s, t, spec in zip(shapes, dtypes, polymorphic_shapes_flat))
 
   return args_tree.unflatten(args_specs_flat)
@@ -1471,7 +1494,7 @@ def shape_and_dtype_jax_array(a) -> tuple[Sequence[int | None], DType]:
   """Returns the shape and dtype of a jax.Array or a j"""
   if isinstance(a, jax.ShapeDtypeStruct):
     return a.shape, a.dtype
-  aval = core.raise_to_shaped(core.get_aval(a))
+  aval = core.get_aval(a)
   return aval.shape, aval.dtype
 
 
@@ -1634,8 +1657,6 @@ class _Parser:
     if tok.exact_type == tokenize.NAME:
       if tok.string in (_DimFactor.MOD, _DimFactor.FLOORDIV, _DimFactor.MAX, _DimFactor.MIN):
         return self.factor_binary_op(tok.string, self.next_tok())
-      if tok.string == _DimFactor.NON_NEGATIVE:  # We still parse this for backwards compatibility
-        return self.factor_unary_op(_DimFactor.NON_NEGATIVE, self.next_tok())
       return _DimExpr._from_var(tok.string, self.scope), self.next_tok()
     number_sign = 1
     if tok.exact_type == tokenize.MINUS:  # -k are negative constants
@@ -1717,20 +1738,19 @@ def _dimension_size_lowering_rule(ctx, arg, *, dimension):
 mlir.register_lowering(dimension_size_p, _dimension_size_lowering_rule)
 
 
-def all_dim_vars(args_avals: Sequence[core.AbstractValue]) -> Sequence[str]:
+def all_dim_vars(args_avals: Sequence[core.ShapedArray]) -> Sequence[str]:
   dim_vars: set[str] = set()
   for a in args_avals:
-    for d in a.shape:  # type: ignore[attribute-error,unused-ignore]
+    for d in a.shape:
       if is_symbolic_dim(d):
         dim_vars = dim_vars.union(d._get_vars())
   return sorted(dim_vars)
 
 
-class CachingShapeEvaluator:
-  def __init__(self, **env):
+class ShapeEvaluator:
+  def __init__(self, env: DimVarEnv):
     self.env = env
 
-  @functools.lru_cache(128)
   def evaluate(self, e: DimSize):
     if core.is_constant_dim(e):
       res = op.index(e)  # type: ignore
@@ -1749,7 +1769,7 @@ class ShapeConstraint:
   # is formed by evaluating the DimSize and concatenating the sequence.
   error_message_pieces: Sequence[str | DimSize]
 
-  def check_statically(self, eval: CachingShapeEvaluator) -> None:
+  def check_statically(self, eval: ShapeEvaluator) -> None:
     """Evaluates a constraint statically."""
     left, right = eval.evaluate(self.left), eval.evaluate(self.right)
     try:
@@ -1765,7 +1785,7 @@ class ShapeConstraint:
     if not ok:
       raise self.make_error(eval)
 
-  def compute(self, eval: CachingShapeEvaluator) -> jax.Array | None:
+  def compute(self, eval: ShapeEvaluator) -> jax.Array | None:
     """Computes if the constraint is satisfied.
 
     If the constraint can be resolved statically returns None
@@ -1800,7 +1820,7 @@ class ShapeConstraint:
 
   def error_message_and_inputs(
     self,
-    eval: CachingShapeEvaluator) -> tuple[str, Sequence[Any]]:
+    eval: ShapeEvaluator) -> tuple[str, Sequence[Any]]:
     """Forms the error_message and error message_inputs.
     See shape_assertion.
     """
@@ -1829,7 +1849,7 @@ class ShapeConstraint:
     return ("".join(error_message_strings),
             error_message_inputs)
 
-  def make_error(self, eval: CachingShapeEvaluator) -> Exception:
+  def make_error(self, eval: ShapeEvaluator) -> Exception:
     error_message, error_message_inputs = self.error_message_and_inputs(eval)
     return ValueError(error_message.format(*error_message_inputs))
 
@@ -1845,7 +1865,7 @@ class ShapeConstraints:
     c = ShapeConstraint(comp, left, right, error_message_pieces)
     self.constraints.append(c)
 
-  def check_statically(self, eval: CachingShapeEvaluator) -> None:
+  def check_statically(self, eval: ShapeEvaluator) -> None:
     """Evaluates all the constraints statically.
 
     If the static checking of any constraint fails, raises ValueError.
@@ -1853,10 +1873,10 @@ class ShapeConstraints:
     for constraint in self.constraints:
       constraint.check_statically(eval)
 
-  def shape_assertions(self, eval: CachingShapeEvaluator) -> None:
+  def shape_assertions(self, eval: ShapeEvaluator) -> None:
     """Computes the shape assertions for the set of constraints.
 
-    See jax_export._wrap_main_func docstring.
+    See jax_export.Exported docstring.
     """
     # We want to report the errors in the same order as `check_statically`.
     # So, we process them in order, in case some fail statically, and we
@@ -1911,7 +1931,7 @@ def pretty_print_dimension_descriptor(
 
 @util.cache()
 def solve_dim_vars(
-    args_avals: Sequence[core.AbstractValue],
+    args_avals: Sequence[core.ShapedArray],
     args_kwargs_tree: tree_util.PyTreeDef,
     ) -> tuple[DimVarEnv, ShapeConstraints, Sequence[tuple[str, int, int]]]:
   """Solves dimension variables in a called function's avals in terms of actual argument shapes.
@@ -1956,12 +1976,12 @@ def solve_dim_vars(
   # tuples with argument name and its polymorphic shape ('args[0]', '(a, a + b'))
   polymorphic_shape_specs: list[tuple[str, str]] = []
   for arg_idx, aval in enumerate(args_avals):
-    if all(not is_symbolic_dim(d) for d in aval.shape):  # type: ignore[attribute-error,unused-ignore]
+    if all(not is_symbolic_dim(d) for d in aval.shape):
       continue
     polymorphic_shape_specs.append(
       (pretty_print_dimension_descriptor(args_kwargs_tree, arg_idx, None),
-       str(aval.shape)))  # type: ignore[attribute-error,unused-ignore]
-    for dim_idx, aval_d in enumerate(aval.shape):  # type: ignore[attribute-error,unused-ignore]
+       str(aval.shape)))
+    for dim_idx, aval_d in enumerate(aval.shape):
       if is_symbolic_dim(aval_d):
         synth_dim_var = pretty_print_dimension_descriptor(args_kwargs_tree,
                                                           arg_idx, dim_idx)
@@ -1976,7 +1996,7 @@ def solve_dim_vars(
 
 
 def compute_dim_vars_from_arg_shapes(
-    args_avals: Sequence[core.AbstractValue],
+    args_avals: Sequence[core.ShapedArray],
     *actual_args: jax.Array,
     args_kwargs_tree: tree_util.PyTreeDef) -> Sequence[jax.Array]:
   """Computes values of dimension variables to unify args_avals with actual arguments.
@@ -1986,7 +2006,8 @@ def compute_dim_vars_from_arg_shapes(
   generate the code for computing the dimension variables. It also generates
   the shape assertions.
 
-  Returns: the values of the dimension variables, in the order determined by
+  Returns:
+    The values of the dimension variables, in the order determined by
     `all_dim_vars(args_avals)`.
   """
   dim_vars = all_dim_vars(args_avals)
@@ -1994,13 +2015,13 @@ def compute_dim_vars_from_arg_shapes(
       tuple(args_avals), args_kwargs_tree=args_kwargs_tree)
 
   # Replace the synthetic vars with the dynamic shape of the actual arg
-  synthetic_env = {vname: dimension_size_p.bind(actual_args[arg_idx],
-                                                dimension=dim_idx)
-                   for (vname, arg_idx, dim_idx) in synth_dim_vars}
-  synthetic_eval = CachingShapeEvaluator(**synthetic_env)
+  synthetic_env: DimVarEnv = {
+      vname: dimension_size_p.bind(actual_args[arg_idx], dimension=dim_idx)
+      for (vname, arg_idx, dim_idx) in synth_dim_vars
+  }
+  synthetic_eval = ShapeEvaluator(synthetic_env)
   shape_constraints.shape_assertions(synthetic_eval)
-  dim_values = [synthetic_eval.evaluate(solution[var]) for var in dim_vars]
-  return tuple(dim_values)
+  return tuple(synthetic_eval.evaluate(solution[var]) for var in dim_vars)
 
 def _solve_dim_equations(
     eqns: list[_DimEquation],
@@ -2009,7 +2030,7 @@ def _solve_dim_equations(
   # Returns a shape environment and the shape constraints if it can solve all
   # dimension variables. Raises an exception if it cannot.
   shape_env: DimVarEnv = {}
-  solution_error_message_pieces: list[str | _DimExpr] = [
+  solution_error_message_pieces: list[str | DimSize] = [
     " Obtained dimension variables: "
   ]  # Error message describing the solution
   # Prepare error message piece describing the polymorphic shape specs
@@ -2017,7 +2038,7 @@ def _solve_dim_equations(
     " Using the following polymorphic shapes specifications: " +
     ",".join(f"{arg_name}.shape = {arg_spec}"
              for arg_name, arg_spec in polymorphic_shape_specs)) + "."
-  solution_err_msg_trailer_errors = ". Please see https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#shape-assertion-errors for more details."
+  solution_err_msg_trailer_errors = ". Please see https://jax.readthedocs.io/en/latest/export/shape_poly.html#shape-assertion-errors for more details."
 
   shape_constraints = ShapeConstraints()  # accumulate shape constraints
   scope: SymbolicScope | None = None
@@ -2043,8 +2064,8 @@ def _solve_dim_equations(
     for term, term_k in eqn.aval_dim_expr._sorted_terms:
       # Perhaps we can already evaluate this term (all vars solved)
       try:
-        term_value = term.evaluate(shape_env)
-      except KeyError:
+        term_value = term.evaluate(shape_env, scope)
+      except UnexpectedDimVar:
         # `mon` still uses some variables not yet solved. We handle only the
         # case when `mon` is a single variable.
         v = term.to_var()
@@ -2111,14 +2132,17 @@ def _solve_dim_equations(
     if not shape_env: return
     assert scope is not None
     for constr in scope._explicit_constraints:
-      c_value = constr.diff._evaluate(shape_env)
+      # We can't just construct constr.e1 - constr.e2 because for an equality
+      # constraint it would be reduced to 0.
+      c_diff = constr.diff._evaluate(shape_env) if not core.is_constant_dim(constr.diff) else constr.diff  # type: ignore
       shape_constraints.add_constraint(
-          constr.cmp, c_value, 0,
+          constr.cmp, c_diff, 0,
           error_message_pieces=[
                 f"Input shapes do not match the symbolic shape constraint {constr.debug_str}. "
                 f"Expected '{constr.diff}' to be "
                 f"{'greater or equal' if constr.cmp == Comparator.GEQ else 'equal'} to 0, "
-                "but found ", c_value,
+                "but found ", c_diff,
+
                 ". " + poly_specs_err_msg
               ] + solution_error_message_pieces + [
               solution_err_msg_trailer_errors])
@@ -2129,7 +2153,8 @@ def _solve_dim_equations(
     eqns = [eqn for eqn in eqns if not process_one_eqn(eqn)]
     if not eqns:
       add_explicit_symbolic_constraints(shape_env)
-      return shape_env, shape_constraints  # SUCCESS
+      # SUCCESS
+      return shape_env, shape_constraints  # pytype: disable=bad-return-type
     elif len(eqns) >= nr_eqns:
       break
 
@@ -2146,6 +2171,6 @@ def _solve_dim_equations(
       " Unprocessed specifications: " +
       ", ".join(f"'{eqn.aval_dim_expr}' for dimension size {eqn.dim_name}"
                 for eqn in eqns) +
-      ". Please see https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#dimension-variables-must-be-solvable-from-the-input-shapes for more details."
+      ". Please see https://jax.readthedocs.io/en/latest/export/shape_poly.html#dimension-variables-must-be-solvable-from-the-input-shapes for more details."
   )
   raise ValueError(err_msg)
