@@ -15,16 +15,18 @@
 from __future__ import annotations
 
 import collections
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import functools
 from functools import partial
 import inspect
 import itertools
 import operator
-from typing import Any, Callable, TypeVar
+from typing import Any, TypeVar
 
 from jax.tree_util import tree_flatten, tree_unflatten
 from jax._src import ad_util
+from jax._src.api_util import (
+    _check_no_aliased_ref_args, _check_no_aliased_closed_over_refs)
 from jax._src import config
 from jax._src import core
 from jax._src import dispatch
@@ -33,9 +35,9 @@ from jax._src import effects
 from jax._src import linear_util as lu
 from jax._src import source_info_util
 from jax._src import util
-from jax._src.state.discharge import register_discharge_rule, discharge_state
+from jax._src.state.discharge import register_partial_discharge_rule, discharge_state
 from jax._src.state.types import AbstractRef, RefEffect
-from jax._src.core import ConcreteArray, raise_to_shaped, replace_jaxpr_effects
+from jax._src.core import replace_jaxpr_effects
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
@@ -49,7 +51,6 @@ from jax._src.lib.mlir.dialects import hlo
 import numpy as np
 
 from jax._src.lax.control_flow.common import (
-    _abstractify,
     _avals_short,
     _check_tree_and_avals,
     _initial_style_jaxprs_with_common_consts,
@@ -87,6 +88,7 @@ def switch(index, branches: Sequence[Callable], *operands,
   Args:
     index: Integer scalar type, indicating which branch function to apply.
     branches: Sequence of functions (A -> B) to be applied based on ``index``.
+      All branches must return the same output structure.
     operands: Operands (A) input to whichever branch is applied.
 
   Returns:
@@ -130,37 +132,36 @@ def switch(index, branches: Sequence[Callable], *operands,
   hi = np.array(len(branches) - 1, np.int32)
   index = lax.clamp(lo, index, hi)
 
-  if (config.disable_jit.value and
-      isinstance(core.get_aval(index), ConcreteArray)):
+  if (config.disable_jit.value and core.is_concrete(index)):
     return branches[int(index)](*operands)
 
   ops, ops_tree = tree_flatten(operands)
-  ops_avals = tuple(map(_abstractify, ops))
+  ops_avals = tuple(map(core.get_aval, ops))
+
+  if config.mutable_array_checks.value:
+    dbg = pe.tracing_debug_info(branches[0], ops_tree, None, False, 'switch')
+    _check_no_aliased_ref_args(dbg, ops_avals, ops)
 
   jaxprs, consts, out_trees = _initial_style_jaxprs_with_common_consts(
       branches, ops_tree, ops_avals, primitive_name='switch')
+  if config.mutable_array_checks.value:
+    _check_no_aliased_closed_over_refs(dbg, (*jaxprs[0].consts, *consts), ops)
   for i, (out_tree, jaxpr) in enumerate(zip(out_trees[1:], jaxprs[1:])):
-    _check_tree_and_avals(f"branch 0 and {i + 1} outputs",
+    _check_tree_and_avals("branch 0 output",
                           out_trees[0], jaxprs[0].out_avals,
+                          f"branch {i + 1} output",
                           out_tree, jaxpr.out_avals)
   joined_effects = core.join_effects(*(jaxpr.effects for jaxpr in jaxprs))
   disallowed_effects = effects.control_flow_allowed_effects.filter_not_in(joined_effects)
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `switch`: {disallowed_effects}')
-  if joined_effects:
-    # Raise index in case of effects to allow data-dependence-based discharging
-    # of those effects (even if they don't have an explicit data dependence).
-    index = core.raise_as_much_as_possible(index)
-
-  linear = (False,) * (len(consts) + len(ops))
-  out = cond_p.bind(
-      index, *consts, *ops, branches=tuple(jaxprs), linear=linear)
+  out = cond_p.bind(index, *consts, *ops, branches=tuple(jaxprs))
   return tree_unflatten(out_trees[0], out)
 
 
 def _cond(pred, true_fun: Callable, false_fun: Callable, *operands,
-          operand=_no_operand_sentinel, linear=None):
+          operand=_no_operand_sentinel):
   """Conditionally apply ``true_fun`` or ``false_fun``.
 
   Wraps XLA's `Conditional
@@ -227,34 +228,41 @@ def _cond(pred, true_fun: Callable, false_fun: Callable, *operands,
       msg = ("Pred type must be either boolean or number, got {}.")
       raise TypeError(msg.format(pred_dtype))
 
-  if config.disable_jit.value and isinstance(core.get_aval(pred), ConcreteArray):
+  if config.disable_jit.value and core.is_concrete(pred):
     if pred:
       return true_fun(*operands)
     else:
       return false_fun(*operands)
 
   ops, ops_tree = tree_flatten(operands)
-  if linear is None:
-    linear_ops = [False] * len(ops)
-  else:
-    linear_ops, ops_tree2 = tree_flatten(linear)
-    if ops_tree != ops_tree2:
-      raise TypeError('linear tree and operand tree mismatch')
-  ops_avals = tuple(map(_abstractify, ops))
+  ops_avals = tuple(map(core.get_aval, ops))
 
+  if config.mutable_array_checks.value:
+    dbg = pe.tracing_debug_info(true_fun, ops_tree, None, False, 'cond')
+    _check_no_aliased_ref_args(dbg, ops_avals, ops)
   jaxprs, consts, out_trees = _initial_style_jaxprs_with_common_consts(
       (true_fun, false_fun), ops_tree, ops_avals, 'cond')
-  if any(isinstance(op_aval, AbstractRef) for op_aval in ops_avals):
-    raise ValueError("Cannot pass `Ref`s into `cond`.")
   true_jaxpr, false_jaxpr = jaxprs
+  if config.mutable_array_checks.value:
+    _check_no_aliased_closed_over_refs(dbg, (*true_jaxpr.consts, *consts), ops)
+
   out_tree, false_out_tree = out_trees
   if any(isinstance(out_aval, AbstractRef) for out_aval in
          true_jaxpr.out_avals + false_jaxpr.out_avals):
     raise ValueError("Cannot return `Ref`s from `cond`.")
 
-  _check_tree_and_avals("true_fun and false_fun output",
+  _check_tree_and_avals("true_fun output",
                         out_tree, true_jaxpr.out_avals,
+                        "false_fun output",
                         false_out_tree, false_jaxpr.out_avals)
+  # prune passhtrough outputs
+  true_fwds = pe._jaxpr_forwarding(true_jaxpr.jaxpr)
+  false_fwds = pe._jaxpr_forwarding(false_jaxpr.jaxpr)
+  in_fwd = [i if i == j else None for i, j in zip(true_fwds, false_fwds)]
+  keep = [f is None for f in in_fwd]
+  true_jaxpr = pe.prune_closed_jaxpr_outputs(true_jaxpr, keep)
+  false_jaxpr = pe.prune_closed_jaxpr_outputs(false_jaxpr, keep)
+
   joined_effects = core.join_effects(true_jaxpr.effects, false_jaxpr.effects)
   disallowed_effects = effects.control_flow_allowed_effects.filter_not_in(joined_effects)
   if disallowed_effects:
@@ -262,17 +270,18 @@ def _cond(pred, true_fun: Callable, false_fun: Callable, *operands,
         f'Effects not supported in `cond`: {disallowed_effects}')
 
   index = lax.convert_element_type(pred, np.int32)
-  if joined_effects:
-    # Raise index in case of effects to allow data-dependence-based discharging
-    # of those effects (even if they don't have an explicit data dependence).
-    index = core.raise_as_much_as_possible(index)
   false_jaxpr = replace_jaxpr_effects(false_jaxpr, joined_effects)
   true_jaxpr = replace_jaxpr_effects(true_jaxpr, joined_effects)
 
-  linear = [False] * len(consts) + linear_ops
-  out = cond_p.bind(
-      index, *consts, *ops,
-      branches=(false_jaxpr, true_jaxpr), linear=tuple(linear))
+  out = cond_p.bind(index, *consts, *ops, branches=(false_jaxpr, true_jaxpr))
+  num_consts = len(consts)
+  out_ = iter(out)
+
+  out = [
+    next(out_) if fwd is None else lax.asarray(ops[fwd - num_consts])
+    for fwd in in_fwd
+  ]
+  assert next(out_, None) is None
   return tree_unflatten(out_tree, out)
 
 @api_boundary
@@ -332,7 +341,7 @@ def _cond_abstract_eval(*avals, branches, **_):
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `cond`: {disallowed_effects}')
-  return map(raise_to_shaped, branches[0].out_avals), joined_effects
+  return branches[0].out_avals, joined_effects
 
 def _bcast_select(pred, on_true, on_false):
   if np.ndim(pred) != np.ndim(on_true):
@@ -346,8 +355,7 @@ def _bcast_select_n(pred, *cases):
     pred = lax.broadcast_in_dim(pred, np.shape(cases[0]), idx)
   return lax.select_n(pred, *cases)
 
-def _cond_batching_rule(spmd_axis_name, axis_size, axis_name, main_type, args,
-                        dims, branches, linear):
+def _cond_batching_rule(axis_data, args, dims, branches):
   index, *ops = args
   index_dim, *op_dims = dims
   # TODO(sharadmv): clean this up by adding a specific blocklist
@@ -369,21 +377,19 @@ def _cond_batching_rule(spmd_axis_name, axis_size, axis_name, main_type, args,
     # optimizations to XLA.
     # TODO(mattjj,frostig): assumes branches are side-effect-free, revise!
     index, *ops = (
-        batching.bdim_at_front(x, d, axis_size) for x, d in zip(args, dims))
+        batching.bdim_at_front(x, d, axis_data.size) for x, d in zip(args, dims))
 
     in_batched  = [True] * len(branches[0].in_avals)
     out_batched = [True] * len(branches[0].out_avals)
 
     branches_batched = [
-        batching.batch_jaxpr(
-            jaxpr, axis_size, in_batched, out_batched, axis_name, spmd_axis_name,
-            main_type)[0]
+        batching.batch_jaxpr(jaxpr, axis_data, in_batched, out_batched)[0]
         for jaxpr in branches]
 
     branch_outs = []
     for i, jaxpr in enumerate(branches_batched):
       # Perform a select on the inputs for safety of reverse-mode autodiff; see
-      # https://github.com/google/jax/issues/1052
+      # https://github.com/jax-ml/jax/issues/1052
       predicate = lax.eq(index, lax._const(index, i))
       ops_ = [_bcast_select(predicate, x, lax.stop_gradient(x)) for x in ops]
       branch_outs.append(core.jaxpr_as_fun(jaxpr)(*ops_))
@@ -395,21 +401,18 @@ def _cond_batching_rule(spmd_axis_name, axis_size, axis_name, main_type, args,
            for b, x, d in zip(ops_bat, ops, op_dims)]
 
     branches_out_bat = [
-        batching.batch_jaxpr(jaxpr, axis_size, ops_bat, False, axis_name,
-                             spmd_axis_name, main_type)[1]
+        batching.batch_jaxpr(jaxpr, axis_data, ops_bat, False)[1]
         for jaxpr in branches]
     out_bat = [any(bat) for bat in zip(*branches_out_bat)]
     branches_batched = tuple(
-        batching.batch_jaxpr(jaxpr, axis_size, ops_bat, out_bat, axis_name,
-                             spmd_axis_name, main_type)[0]
+        batching.batch_jaxpr(jaxpr, axis_data, ops_bat, out_bat)[0]
         for jaxpr in branches)
 
     out_dims = [0 if b else batching.not_mapped for b in out_bat]
-    out = cond_p.bind(
-        index, *ops, branches=branches_batched, linear=linear)
+    out = cond_p.bind(index, *ops, branches=branches_batched)
     return out, out_dims
 
-def _cond_jvp(primals, tangents, branches, linear):
+def _cond_jvp(primals, tangents, branches):
   nonzeros = [type(t) is not ad_util.Zero for t in tangents]
 
   index_nz, *ops_nz = nonzeros
@@ -426,17 +429,14 @@ def _cond_jvp(primals, tangents, branches, linear):
   _, *ops_dot = tangents
   ops_dot = _prune_zeros(ops_dot)
 
-  ops_lin = tuple(linear)
-  linear_jvp = ops_lin + (True,) * len(ops_dot)
-  out = cond_p.bind(
-      index, *ops, *ops_dot, branches=branches_jvp, linear=linear_jvp)
+  out = cond_p.bind(index, *ops, *ops_dot, branches=branches_jvp)
   out_primals, out_tangents = split_list(out, [len(out_nz)])
   out_tangents_iter = iter(out_tangents)
-  out_tangents = [next(out_tangents_iter) if nz else ad_util.Zero.from_value(p)
+  out_tangents = [next(out_tangents_iter) if nz else ad_util.Zero.from_primal_value(p)
                   for p, nz in zip(out_primals, out_nz)]
   return out_primals, out_tangents
 
-def _cond_partial_eval(trace, *tracers, branches, linear):
+def _cond_partial_eval(trace, *tracers, branches):
   in_unknowns = [t.pval[0] is not None for t in tracers]
   index_uk, *ops_uk = in_unknowns
   if any(isinstance(eff, RefEffect) for branch in branches for eff in
@@ -447,7 +447,7 @@ def _cond_partial_eval(trace, *tracers, branches, linear):
   if index_uk:
     # When the branch index is unknown, we stage out the whole cond.
     # TODO(mattjj): remove this path when old remat is removed
-    params = dict(branches=branches, linear=linear)
+    params = dict(branches=branches)
     return trace.default_process_primitive(cond_p, tracers, params)
 
   branches_out_uks = []
@@ -477,9 +477,7 @@ def _cond_partial_eval(trace, *tracers, branches, linear):
              for j in branches_known[1:])
 
   in_consts = [t.pval.get_known() for t in tracers if t.pval.is_known()]
-  linear_known = [l for l, uk in zip(linear, ops_uk) if not uk]
-  out_consts_res = cond_p.bind(*in_consts, branches=branches_known,
-                               linear=tuple(linear_known))
+  out_consts_res = cond_p.bind(*in_consts, branches=branches_known)
   out_consts, res = split_list(out_consts_res, [len(out_consts_res) - num_res])
 
   index_tracer = trace.instantiate_const(tracers[0])
@@ -488,9 +486,7 @@ def _cond_partial_eval(trace, *tracers, branches, linear):
   res_tracers = map(trace.new_instantiated_const, res)
   out_tracers = [pe.JaxprTracer(trace, pe.PartialVal.unknown(aval), None)
                  for aval in branches_unknown[0].out_avals]
-  linear_unknown = ([False] * num_res +
-                    [l for l, uk in zip(linear, in_unknowns[1:]) if uk])
-  params = dict(branches=branches_unknown, linear=tuple(linear_unknown))
+  params = dict(branches=branches_unknown)
   name_stack = source_info_util.current_name_stack()[len(trace.name_stack):]
   source = source_info_util.current().replace(name_stack=name_stack)
   eqn = pe.new_eqn_recipe(
@@ -529,7 +525,7 @@ def _cond_partial_eval_custom(saveable, unks_in, inst_in, eqn):
   # jaxpr for each branch.
   branches_known_ : list[core.ClosedJaxpr] = []
   branches_staged_: list[core.ClosedJaxpr] = []
-  branch_res_avals: list[core.AbstractValue] = []
+  branch_res_avals: list[list[core.AbstractValue]] = []
   for jaxpr in branches:
     jaxpr_known, jaxpr_staged, _, inst_out, num_res = \
         pe.partial_eval_jaxpr_custom(
@@ -559,8 +555,7 @@ def _cond_partial_eval_custom(saveable, unks_in, inst_in, eqn):
   # Build the known eqn.
   ins_known, _ = partition_list(unks_in, eqn.invars)  # includes index invar
   out_binders_known, _ = partition_list(unks_out, eqn.outvars)
-  linear_known = [l for l, uk in zip(eqn.params['linear'], ops_uk) if not uk]
-  params_known = dict(branches=branches_known, linear=tuple(linear_known))
+  params_known = dict(branches=branches_known)
   effects_known = _join_cond_effects(branches_known)
   eqn_known = pe.new_jaxpr_eqn(
       ins_known, [*out_binders_known, *res_binders], cond_p, params_known,
@@ -568,8 +563,7 @@ def _cond_partial_eval_custom(saveable, unks_in, inst_in, eqn):
 
   # Build the staged eqn.
   _, out_binders_staged = partition_list(inst_out, eqn.outvars)
-  linear_staged = [False] * len(res_binders) + list(eqn.params['linear'])
-  params_staged = dict(branches=branches_staged, linear=tuple(linear_staged))
+  params_staged = dict(branches=branches_staged)
   effects_staged = _join_cond_effects(branches_staged)
   eqn_staged = pe.new_jaxpr_eqn(
       [eqn.invars[0], *res_binders, *eqn.invars[1:]], out_binders_staged,
@@ -661,7 +655,11 @@ def _ordered_unique(xs):
   return list(d.keys())
 
 def _cond_dce_rule(used_outputs: list[bool], eqn: core.JaxprEqn,
-                   ) -> tuple[list[bool], core.JaxprEqn]:
+                   ) -> tuple[list[bool], core.JaxprEqn | None]:
+
+  if not any(used_outputs) and not pe.has_effects(eqn):
+    return [False] * len(eqn.invars), None
+
   closed_branches = eqn.params['branches']
   branches = [closed_jaxpr.jaxpr for closed_jaxpr in closed_branches]
 
@@ -678,9 +676,7 @@ def _cond_dce_rule(used_outputs: list[bool], eqn: core.JaxprEqn,
                   for closed_jaxpr, jaxpr in zip(closed_branches, dce_branches_)]
 
   # Finally, update parameters and form the new eqn.
-  dce_linear = [l for l, used in zip(eqn.params['linear'], used_inputs) if used]
-  new_params = dict(eqn.params, branches=tuple(dce_branches),
-                    linear=tuple(dce_linear))
+  new_params = dict(eqn.params, branches=tuple(dce_branches))
   new_effects = core.join_effects(*(b.effects for b in dce_branches))
   new_effects = _join_cond_effects(dce_branches_)
   new_eqn = pe.new_jaxpr_eqn(
@@ -697,7 +693,6 @@ def _cond_dce_rule(used_outputs: list[bool], eqn: core.JaxprEqn,
 
 def _transpose_cond_jaxpr(jaxpr, num_res):
   res_avals, primal_avals = split_list(jaxpr.in_avals, [num_res])
-  primal_avals = map(raise_to_shaped, primal_avals)
 
   @lu.wrap_init
   def transposed(*args):
@@ -710,11 +705,11 @@ def _transpose_cond_jaxpr(jaxpr, num_res):
 
   return _make_closed_jaxpr(transposed, res_avals + jaxpr.out_avals)
 
-def _cond_transpose(cts, *args, branches, linear):
-  del linear  # could use for error checking, but see #14026
+def _cond_transpose(cts, *args, branches):
   index, *ops = args
+  assert type(index) is not ad.UndefinedPrimal
   linear = [type(x) is ad.UndefinedPrimal for x in ops]
-  in_avals = map(raise_to_shaped, branches[0].in_avals)
+  in_avals = branches[0].in_avals
   num_res = len(ops) - sum(linear)
   if any(isinstance(eff, RefEffect) for branch in branches for eff in
       branch.jaxpr.effects):
@@ -722,18 +717,15 @@ def _cond_transpose(cts, *args, branches, linear):
 
   branches_trans = tuple(
       _transpose_cond_jaxpr(jaxpr, num_res) for jaxpr in branches)
-  lin_in_avals = [raise_to_shaped(a, weak_type=False)
-                  for a, l in zip(in_avals, linear) if l]
+  lin_in_avals = [a.strip_weak_type() for a, l in zip(in_avals, linear) if l]
   assert all(core.typematch(out_aval, lin_in_aval)
              for jaxpr in branches_trans
              for out_aval, lin_in_aval in zip(jaxpr.out_avals, lin_in_avals))
 
   res = ops[:num_res]
   cts = map(ad.instantiate_zeros, cts)
-  linear_trans = (False,) * num_res + (True,) * len(cts)
 
-  out = cond_p.bind(
-      index, *res, *cts, branches=branches_trans, linear=linear_trans)
+  out = cond_p.bind(index, *res, *cts, branches=branches_trans)
   assert all(map(core.typecheck, lin_in_avals, out))
 
   out_iter = iter(out)
@@ -741,13 +733,7 @@ def _cond_transpose(cts, *args, branches, linear):
   assert next(out_iter, None) is None
   return [None] + out
 
-def _cond_axis_substitution(params, subst, traverse):
-  if not traverse:
-    return params
-  branches = tuple(core.subst_axis_names_jaxpr(jaxpr, subst) for jaxpr in params['branches'])
-  return dict(params, branches=branches)
-
-def _cond_typecheck(bind_time, *in_atoms, branches, linear):
+def _cond_typecheck(bind_time, *in_atoms, branches):
   if not bind_time:
     _, *in_atoms = in_atoms
   avals = [x.aval for x in in_atoms]
@@ -755,14 +741,9 @@ def _cond_typecheck(bind_time, *in_atoms, branches, linear):
   tc(branches, 'branches', 'tuple of ClosedJaxpr',
      type(branches) is tuple and
      all(type(x) is core.ClosedJaxpr for x in branches))
-  tc(linear, 'linear', 'tuple of bool',
-     type(linear) is tuple and all(type(x) is bool for x in linear))
 
   if len(branches) == 0:
     raise core.JaxprTypeError('cond requires at least one branch function')
-  if len(linear) + 1 != len(avals):
-    raise core.JaxprTypeError(f'cond given {len(linear)} linear flags for '
-                              f'{len(avals) - 1} non-predicate operands')
 
   jaxpr0 = branches[0]
   jaxpr0_in_avals_str = _avals_short(jaxpr0.in_avals)
@@ -806,41 +787,29 @@ def _cond_typecheck(bind_time, *in_atoms, branches, linear):
       f'called with operands of type {_avals_short(op_avals)}')
   return jaxpr0.out_avals, joined_effects
 
-def cond_bind(*args, branches, linear):
-  if config.enable_checks.value:
-    avals = map(core.get_aval, args)
-    in_atoms = [core.Var('', a) for a in avals]  # dummies
-    _cond_typecheck(True, *in_atoms, branches=branches, linear=linear)
-    for jaxpr in branches:
-      core.check_jaxpr(jaxpr.jaxpr)
-  return core.AxisPrimitive.bind(cond_p, *args, branches=branches, linear=linear)
-
-cond_p = core.AxisPrimitive('cond')
+cond_p = core.Primitive('cond')
 cond_p.multiple_results = True
 cond_p.def_impl(partial(dispatch.apply_primitive, cond_p))
 cond_p.def_effectful_abstract_eval(_cond_abstract_eval)
-cond_p.def_custom_bind(cond_bind)
 ad.primitive_jvps[cond_p] = _cond_jvp
-ad.reducing_transposes[cond_p] = _cond_transpose
+ad.primitive_transposes[cond_p] = _cond_transpose
 pe.custom_partial_eval_rules[cond_p] = _cond_partial_eval
-batching.spmd_axis_primitive_batchers[cond_p] = _cond_batching_rule
-batching.axis_primitive_batchers[cond_p] = partial(_cond_batching_rule, None)
+batching.fancy_primitive_batchers[cond_p] = _cond_batching_rule
 xla.register_initial_style_primitive(cond_p)
 core.custom_typechecks[cond_p] = partial(_cond_typecheck, False)
-core.axis_substitution_rules[cond_p] = _cond_axis_substitution
 pe.partial_eval_jaxpr_custom_rules[cond_p] = _cond_partial_eval_custom
 pe.dce_rules[cond_p] = _cond_dce_rule
+batching.ragged_prop_rules[cond_p] = batching.ragged_mask_assert_no_op_rule
 
-def _cond_lowering(ctx, index, *args, branches, linear):
-  del linear  # Unused.
+def _cond_lowering(ctx, index, *args, branches):
   joined_effects = core.join_effects(*(branch.effects for branch in branches))
   ordered_effects = list(effects.ordered_effects.filter_in(joined_effects))
   num_tokens = len(ordered_effects)
   tokens_in = ctx.tokens_in.subset(ordered_effects)
   output_token_types = [mlir.token_type() for _ in ordered_effects]
   output_types = [
-      *output_token_types, *map(mlir.aval_to_ir_types, ctx.avals_out)]
-  flat_output_types = util.flatten(output_types)
+      *output_token_types, *map(mlir.aval_to_ir_type, ctx.avals_out)]
+  flat_output_types = mlir.flatten_ir_types(output_types)
 
   # CaseOp takes a single argument 'index' and the corresponding blocks
   # have no arguments; the computation within the block uses implicit
@@ -851,35 +820,65 @@ def _cond_lowering(ctx, index, *args, branches, linear):
   for i, jaxpr in enumerate(branches):
     branch = case_op.regions[i].blocks.append()
     with ir.InsertionPoint(branch):
-      consts = [mlir.ir_constants(xla.canonicalize_dtype(x)) for x in jaxpr.consts]
+      consts = [mlir.ir_constant(xla.canonicalize_dtype(x)) for x in jaxpr.consts]
       out_vals, tokens_out = mlir.jaxpr_subcomp(
           ctx.module_context, jaxpr.jaxpr, name_stack.extend(f'branch_{i}_fun'),
-          tokens_in, consts, *map(mlir.wrap_singleton_ir_values, args),
+          tokens_in, consts, *args,
           dim_var_values=ctx.dim_var_values)
       out_tokens = [tokens_out.get(eff) for eff in ordered_effects]
       out_vals = [*out_tokens, *out_vals]
-      hlo.return_(util.flatten(out_vals))
+      hlo.return_(mlir.flatten_ir_values(out_vals))
 
-  tokens_and_outputs = util.unflatten(case_op.results, map(len, output_types))
+  tokens_and_outputs = mlir.unflatten_ir_values_like_types(
+    case_op.results, output_types)
   tokens, outputs = util.split_list(tokens_and_outputs, [num_tokens])
   ctx.set_tokens_out(mlir.TokenSet(zip(ordered_effects, tokens)))
   return outputs
 
 mlir.register_lowering(cond_p, _cond_lowering)
 
-@register_discharge_rule(cond_p)
-def _cond_state_discharge_rule(in_avals, out_avals, *args, branches, linear):
+@register_partial_discharge_rule(cond_p)
+def _cond_state_discharge_rule(should_discharge, in_avals, out_avals, index, *args, branches):
+  assert not should_discharge[0], "Can't discharge the index."
   discharged_branches = tuple(
-      core.ClosedJaxpr(discharge_state(branch.jaxpr, ())[0], ())
-      for branch in branches)
-  out_vals = cond_p.bind(*args, branches=discharged_branches, linear=linear)
-  out_vals, out_ref_vals = util.split_list(
-      out_vals, [len(out_avals)])
+      discharge_state(branch.jaxpr, (), should_discharge=should_discharge[1:])[0]
+      for branch in branches
+  )
+  # Don't thread the ref values through the cond if they never change.
+  forwarded_outvars = None
+  for branch in discharged_branches:
+    invar_pos = {v: i for i, v in enumerate(branch.invars)}
+    branch_forwarding = [
+        invar_pos.get(v, None) if isinstance(v, core.Var) else None
+        for v in branch.outvars[len(out_avals) :]
+    ]
+    if forwarded_outvars is None:
+      forwarded_outvars = branch_forwarding
+    else:
+      forwarded_outvars = [
+          i if i == j else None
+          for i, j in zip(forwarded_outvars, branch_forwarding)
+      ]
+  assert forwarded_outvars is not None
+  all_outvars_fwd = [None] * len(out_avals) + forwarded_outvars
+  new_branches = tuple(
+      core.ClosedJaxpr(
+          branch.replace(outvars=[v for v, fwd in zip(branch.outvars, all_outvars_fwd)
+                                  if fwd is None]), ())
+      for branch in discharged_branches
+  )
+  out_vals_no_fwd = cond_p.bind(index, *args, branches=new_branches)
+  out_vals, out_ref_vals_no_fwd = util.split_list(out_vals_no_fwd, [len(out_avals)])
+  # Insert forwarded values into reference outputs
+  ref_val_no_fwd_iter = iter(out_ref_vals_no_fwd)
+  out_ref_vals = [next(ref_val_no_fwd_iter) if fwd is None else args[fwd]
+                  for fwd in forwarded_outvars]
+  # Map reference outputs back to their invars
   ref_val_iter = iter(out_ref_vals)
   new_invals = []
-  for aval in in_avals:
-    new_invals.append(
-        next(ref_val_iter) if isinstance(aval, AbstractRef) else None)
+  for should, aval in zip(should_discharge, in_avals):
+    discharged_inval = isinstance(aval, AbstractRef) and should
+    new_invals.append(next(ref_val_iter) if discharged_inval else None)
   return new_invals, out_vals
 
 
@@ -948,12 +947,27 @@ def platform_dependent(*args: Any,
   platform_index = platform_index_p.bind(
     platforms=tuple(tuple(ps) for ps in platforms_lists),
     has_default=(default is not None))
+
   if default is not None:
     branches = branches + (default,)
   # Use a switch, to get the proper transformation rules for free. Since
   # platform index has no dependence on the input data, it won't be vectorized
   # under vmap.
-  return switch(platform_index, branches, *args)
+  # If the switch and the platform_index_p above are in the same compilation
+  # unit then constant-folding will remove the unnecessary branches. However,
+  # if we run in eager mode the switch below cannot be constant-folded and
+  # the compilation may fail if some of the branches contain custom calls not
+  # recognized on the compilation platform. Detect eager mode and keep only the
+  # needed branch.
+  try:
+    # Note/TODO(mvoz): This actually rarely seems to concretize - we could look into
+    # core.ensure_compile_time_eval to get better single-branch selection.
+    platform_index_concrete = core.concrete_or_error(operator.index, platform_index)
+  except core.ConcretizationTypeError:
+    return switch(platform_index, branches, *args)
+  else:
+    assert 0 <= platform_index_concrete < len(branches)
+    return branches[platform_index_concrete](*args)
 
 # A primitive to compute the index of a platform into a list of platforms.
 # Args:
@@ -978,7 +992,9 @@ def _platform_index_lowering(ctx: mlir.LoweringRuleContext,
   def lower_constant(
       ctx: mlir.LoweringRuleContext, *, i: int
   ) -> Sequence[ir.Value]:
-    return mlir.ir_constants(np.int32(i))
+    v = mlir.ir_constant(np.int32(i))
+    assert isinstance(v, ir.Value), v
+    return [v]
   platform_rules: dict[str, mlir.LoweringRule] = {}
   for i, ps in enumerate(platforms):
     rule = partial(lower_constant, i=i)
