@@ -15,9 +15,14 @@
 # ruff: noqa
 
 try:
-  from jaxlib.mlir._mlir_libs import _mosaic_gpu_ext  # pytype: disable=import-error
+  try:
+    from jaxlib.mosaic.gpu import _mosaic_gpu_ext  # pytype: disable=import-error
+  except ImportError as e:
+    print(e)
+    from jax_cuda12_plugin import _mosaic_gpu_ext  # pytype: disable=import-error
 except ImportError as e:
-  raise ModuleNotFoundError(
-      "Cannot import the Mosaic GPU bindings. You may need to build jaxlib from"
-      " source."
-  ) from e
+  print("="*128)
+  print(e)
+  raise ModuleNotFoundError("Failed to import the Mosaic GPU bindings") from e
+else:
+  _mosaic_gpu_ext.register_passes()
