@@ -19,30 +19,31 @@ from __future__ import annotations
 import enum
 import functools
 import string
-from typing import Any
+from typing import Any, Callable
 
 import jax
 from jax import lax
 from jax import tree_util
 from jax._src import ad_util
+from jax._src import api_util
 from jax._src import core as jax_core
+from jax._src import dtypes
 from jax._src import effects
+from jax._src import linear_util as lu
 from jax._src import pretty_printer as pp
 from jax._src import state
 from jax._src import util
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
+from jax._src.interpreters import partial_eval as pe
 from jax._src.pallas import core as pallas_core
 from jax._src.state import discharge as state_discharge
 from jax._src.state import indexing
+from jax._src.state import types as state_types
 from jax._src.state import primitives as sp
 from jax.interpreters import mlir
 import jax.numpy as jnp
 
-
-# TODO(sharadmv): enable type checking
-# mypy: ignore-errors
-
 partial = functools.partial
 Slice = indexing.Slice
 NDIndexer = indexing.NDIndexer
@@ -51,43 +52,58 @@ map, unsafe_map = util.safe_map, map
 zip, unsafe_zip = util.safe_zip, zip
 
 program_id_p = jax_core.Primitive("program_id")
+batching.ragged_prop_rules[program_id_p] = batching.ragged_mask_no_op_rule
 
 def program_id(axis: int) -> jax.Array:
+  """Returns the kernel execution position along the given axis of the grid.
+
+  For example, with a 2D `grid` in the kernel execution corresponding to the
+  grid coordinates `(1, 2)`,
+  `program_id(axis=0)` returns `1` and `program_id(axis=1)` returns `2`.
+
+  The returned value is an array of shape `()` and dtype `int32`.
+
+  Args:
+    axis: the axis of the grid along which to count the program.
+  """
   return program_id_p.bind(axis=axis)
 
-def program_id_bind(*, axis: int):
+def program_id_bind_with_trace(trace, _, params):
+  axis = params.pop("axis")
   grid_env = pallas_core.current_grid_env()
   if grid_env:
-    return grid_env[axis].axis_index
-  return jax_core.Primitive.bind(program_id_p, axis=axis)
-program_id_p.def_custom_bind(program_id_bind)
-
-def _program_id_impl(*, axis: int):
-  grid_env = pallas_core.current_grid_env()
-  return grid_env[axis].axis_index
-program_id_p.def_impl(_program_id_impl)
+    return grid_env[axis].index
+  frame = pallas_core.axis_frame()
+  # Query the size of the axis to make sure it's a valid axis (and error
+  # otherwise).
+  _ = frame.size(axis)
+  return jax_core.Primitive.bind_with_trace(program_id_p, trace, (), dict(axis=axis))
+# TODO(dougalm): figure out how put the grid_env contest on the relevant trace
+program_id_p.def_bind_with_trace(program_id_bind_with_trace)
 
+@program_id_p.def_abstract_eval
 def _program_id_abstract_eval(**_):
   return jax_core.ShapedArray((), jnp.int32)
-program_id_p.def_abstract_eval(_program_id_abstract_eval)
-
 
 num_programs_p = jax_core.Primitive("num_programs")
 
-def num_programs(axis: int) -> jax.Array:
+def num_programs(axis: int) -> int | jax.Array:
+  """Returns the size of the grid along the given axis."""
   return num_programs_p.bind(axis=axis)
 
-@num_programs_p.def_custom_bind
-def _num_programs_bind(*, axis: int):
+def _num_programs_bind_with_trace(trace, _, params):
+  axis = params.pop("axis")
+  # We might be using a local grid env
   grid_env = pallas_core.current_grid_env()
   if grid_env:
-    return jnp.asarray(grid_env[axis].axis_size, dtype=jnp.int32)
-  return jax_core.Primitive.bind(num_programs_p, axis=axis)
-
-@num_programs_p.def_impl
-def _num_programs_impl(*, axis: int):
-  grid_env = pallas_core.current_grid_env()
-  return jnp.asarray(grid_env[axis].axis_size, dtype=jnp.int32)
+    return grid_env[axis].size
+  # Otherwise, we look up the size of the grid in the axis env
+  frame = pallas_core.axis_frame()
+  size = frame.size(axis)
+  if size is pallas_core.dynamic_grid_dim:
+    return jax_core.Primitive.bind_with_trace(num_programs_p, trace, (), dict(axis=axis))
+  return size
+num_programs_p.def_bind_with_trace(_num_programs_bind_with_trace)
 
 @num_programs_p.def_abstract_eval
 def _num_programs_abstract_eval(**_):
@@ -149,6 +165,7 @@ def _atomic_rmw_discharge_rule(
 state_discharge.register_discharge_rule(atomic_rmw_p)(_atomic_rmw_discharge_rule)
 
 
+@atomic_rmw_p.def_effectful_abstract_eval
 def _atomic_abstract_eval(*avals_flat, args_tree, atomic_type: AtomicOpType):
   ref, _, _, _ = args_tree.unflatten(avals_flat)
   if ref.dtype == jnp.dtype("float16") and atomic_type != AtomicOpType.ADD:
@@ -165,41 +182,154 @@ def _atomic_abstract_eval(*avals_flat, args_tree, atomic_type: AtomicOpType):
   return _swap_abstract_eval(*avals_flat, args_tree=args_tree)
 
 
-atomic_rmw_p.def_effectful_abstract_eval(_atomic_abstract_eval)
-
-def atomic_rmw(x_ref_or_view, idx, val, *, mask: Any | None = None,
-               atomic_type: AtomicOpType):
-  x_ref, indexers = sp.get_ref_and_indexers(x_ref_or_view, idx, "atomic_rmw")
-  args_flat, args_tree = tree_util.tree_flatten((x_ref, indexers, val, mask))
+def _atomic_rmw(x_ref_or_view, idx, val, *, mask: Any | None = None,
+                atomic_type: AtomicOpType):
+  x_ref, transforms = sp.get_ref_and_transforms(
+      x_ref_or_view, idx, "atomic_rmw"
+  )
+  args_flat, args_tree = tree_util.tree_flatten((x_ref, transforms, val, mask))
   return atomic_rmw_p.bind(
       *args_flat, args_tree=args_tree, atomic_type=atomic_type
   )
 
-atomic_xchg = functools.partial(atomic_rmw, atomic_type=AtomicOpType.XCHG)
-atomic_add = functools.partial(atomic_rmw, atomic_type=AtomicOpType.ADD)
-atomic_max = functools.partial(atomic_rmw, atomic_type=AtomicOpType.MAX)
-atomic_min = functools.partial(atomic_rmw, atomic_type=AtomicOpType.MIN)
-atomic_and = functools.partial(atomic_rmw, atomic_type=AtomicOpType.AND)
-atomic_or = functools.partial(atomic_rmw, atomic_type=AtomicOpType.OR)
-atomic_xor = functools.partial(atomic_rmw, atomic_type=AtomicOpType.XOR)
+def atomic_xchg(x_ref_or_view, idx, val, *, mask: Any | None = None):
+  """Atomically exchanges the given value with the value at the given index.
+
+  Args:
+    x_ref_or_view: The ref to operate on.
+    idx: The indexer to use.
+    mask: TO BE DOCUMENTED.
+
+  Returns:
+    The value at the given index prior to the aupdate.
+  """
+  return _atomic_rmw(
+      x_ref_or_view, idx, val, mask=mask, atomic_type=AtomicOpType.XCHG
+  )
+
+
+def atomic_add(x_ref_or_view, idx, val, *, mask: Any | None = None):
+  """Atomically computes ``x_ref_or_view[idx] += val``.
+
+  Args:
+    x_ref_or_view: The ref to operate on.
+    idx: The indexer to use.
+    mask: TO BE DOCUMENTED.
+
+  Returns:
+    The value at the given index prior to the atomic operation.
+  """
+  return _atomic_rmw(
+      x_ref_or_view, idx, val, mask=mask, atomic_type=AtomicOpType.ADD
+  )
+
+
+def atomic_max(x_ref_or_view, idx, val, *, mask: Any | None = None):
+  """Atomically computes ``x_ref_or_view[idx] = max(x_ref_or_view[idx], val)``.
+
+  Args:
+    x_ref_or_view: The ref to operate on.
+    idx: The indexer to use.
+    mask: TO BE DOCUMENTED.
+
+  Returns:
+    The value at the given index prior to the atomic operation.
+  """
+  return _atomic_rmw(
+      x_ref_or_view, idx, val, mask=mask, atomic_type=AtomicOpType.MAX
+  )
+
+
+def atomic_min(x_ref_or_view, idx, val, *, mask: Any | None = None):
+  """Atomically computes ``x_ref_or_view[idx] = min(x_ref_or_view[idx], val)``.
+
+  Args:
+    x_ref_or_view: The ref to operate on.
+    idx: The indexer to use.
+    mask: TO BE DOCUMENTED.
+
+  Returns:
+    The value at the given index prior to the atomic operation.
+  """
+  return _atomic_rmw(
+      x_ref_or_view, idx, val, mask=mask, atomic_type=AtomicOpType.MIN
+  )
+
+
+def atomic_and(x_ref_or_view, idx, val, *, mask: Any | None = None):
+  """Atomically computes ``x_ref_or_view[idx] &= val``.
+
+  Args:
+    x_ref_or_view: The ref to operate on.
+    idx: The indexer to use.
+    mask: TO BE DOCUMENTED.
+
+  Returns:
+    The value at the given index prior to the atomic operation.
+  """
+  return _atomic_rmw(
+      x_ref_or_view, idx, val, mask=mask, atomic_type=AtomicOpType.AND
+  )
+
+
+def atomic_or(x_ref_or_view, idx, val, *, mask: Any | None = None):
+  """Atomically computes ``x_ref_or_view[idx] |= val``.
+
+  Args:
+    x_ref_or_view: The ref to operate on.
+    idx: The indexer to use.
+    mask: TO BE DOCUMENTED.
+
+  Returns:
+    The value at the given index prior to the atomic operation.
+  """
+  return _atomic_rmw(
+      x_ref_or_view, idx, val, mask=mask, atomic_type=AtomicOpType.OR
+  )
+
+
+def atomic_xor(x_ref_or_view, idx, val, *, mask: Any | None = None):
+  """Atomically computes ``x_ref_or_view[idx] ^= val``.
+
+  Args:
+    x_ref_or_view: The ref to operate on.
+    idx: The indexer to use.
+    mask: TO BE DOCUMENTED.
+
+  Returns:
+    The value at the given index prior to the atomic operation.
+  """
+  return _atomic_rmw(
+      x_ref_or_view, idx, val, mask=mask, atomic_type=AtomicOpType.XOR
+  )
 
 atomic_cas_p = jax_core.Primitive("atomic_cas")
 
+@atomic_cas_p.def_effectful_abstract_eval
 def _atomic_cas_abstract_eval(ref_aval, cmp_aval, val_aval):
-  if cmp_aval.dtype != val_aval.dtype:
-    raise ValueError("Dtypes in cmp/val need to match")
+  if cmp_aval.dtype != val_aval.dtype or cmp_aval.shape != val_aval.shape:
+    raise ValueError("cmp and val must have identical dtypes and shapes")
   if ref_aval.shape:
-    raise ValueError("Ref must be scalar.")
+    raise ValueError("ref must be scalar.")
   if cmp_aval.shape:
-    raise ValueError("Cmp must be scalar.")
+    raise ValueError("cmp must be scalar.")
   if val_aval.shape:
-    raise ValueError("Val must be scalar.")
-  if cmp_aval.shape != val_aval.shape:
-    raise ValueError("Dtypes in cmp/val need to match")
+    raise ValueError("val must be scalar.")
   return jax_core.ShapedArray(val_aval.shape, val_aval.dtype), {state.WriteEffect(0)}
-atomic_cas_p.def_effectful_abstract_eval(_atomic_cas_abstract_eval)
+
 
 def atomic_cas(ref, cmp, val):
+  """Performs an atomic compare-and-swap of the value in the ref with the
+  given value.
+
+  Args:
+    ref: The ref to operate on.
+    cmp: The expected value to compare against.
+    val: The value to swap in.
+
+  Returns:
+    The value at the given index prior to the atomic operation.
+  """
   return atomic_cas_p.bind(ref, cmp, val)
 
 @state_discharge.register_discharge_rule(atomic_cas_p)
@@ -218,9 +348,9 @@ def max_contiguous(x, values):
     values = [values]
   return max_contiguous_p.bind(x, values=values)
 
+@max_contiguous_p.def_abstract_eval
 def _max_contiguous_abstract_eval(aval, **_):
   return aval
-max_contiguous_p.def_abstract_eval(_max_contiguous_abstract_eval)
 
 multiple_of_p = jax_core.Primitive("multiple_of")
 
@@ -232,13 +362,14 @@ def multiple_of(x: jax.Array, values: list[int] | int) -> jax.Array:
     values = [values]
   return multiple_of_p.bind(x, values=values)
 
+@multiple_of_p.def_abstract_eval
 def _multiple_of_abstract_eval(aval, **_):
   return aval
-multiple_of_p.def_abstract_eval(_multiple_of_abstract_eval)
 
 load_p = jax_core.Primitive('masked_load')
 
 
+@load_p.def_effectful_abstract_eval
 def _load_abstract_eval(*avals_flat, args_tree, **_):
   ref, indexers, _, _ = args_tree.unflatten(avals_flat)
   return (
@@ -247,8 +378,6 @@ def _load_abstract_eval(*avals_flat, args_tree, **_):
   )
 
 
-load_p.def_effectful_abstract_eval(_load_abstract_eval)
-
 def _load_pp_rule(eqn, context, settings):
   # Pretty prints `a = load x i` as `x[i] <- a`
   y, = eqn.outvars
@@ -259,7 +388,7 @@ def _load_pp_rule(eqn, context, settings):
   result = [
       lhs,
       pp.text(' <- '),
-      sp.pp_ref_indexers(context, x, indexers)
+      sp.pp_ref_transforms(context, x, indexers)
   ]
   if mask is not None:
     result += [
@@ -301,10 +430,17 @@ ad.primitive_jvps[load_p] = _load_jvp
 def uninitialized_value(shape, dtype):
   if jnp.issubdtype(dtype, jnp.floating):
     return jnp.full(shape, jnp.nan, dtype)
+  # Note: Currently semaphore is i16[], meaning this case needs to be
+  # handled before the general case for integers.
+  # TODO(justinfu): Handle semaphores with a custom extended dtype.
+  elif jnp.issubdtype(dtype, pallas_core.SEMAPHORE_INTERPRET_DTYPE):
+    return jnp.full(shape, 0, dtype)
   elif jnp.issubdtype(dtype, jnp.integer):
     return jnp.full(shape, jnp.iinfo(dtype).min, dtype)
   elif jnp.issubdtype(dtype, jnp.bool):
     return jnp.full(shape, False, dtype)
+  elif jnp.issubdtype(dtype, pallas_core.semaphore_dtype):
+    return jnp.full(shape, 0, dtype)
   raise NotImplementedError(dtype)
 
 def _pad_values_to_avoid_dynamic_slice_oob_shift(value,
@@ -334,6 +470,8 @@ def _pad_values_to_avoid_dynamic_slice_oob_shift(value,
 _unpad_values_to_avoid_dynamic_slice_oob_shift = partial(
   _pad_values_to_avoid_dynamic_slice_oob_shift, unpad=True)
 
+
+@state_discharge.register_discharge_rule(load_p)
 def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   del out_avals  # Unused.
   ref, indexers, mask, other = args_tree.unflatten(args_flat)
@@ -342,7 +480,7 @@ def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
     raise NotImplementedError("Only one indexer supported in discharge rule.")
   idx = indexers[0]
   if all((isinstance(s, Slice) or not s.shape) for s in idx.indices):
-    # TODO(b/329733289): support strided load/store in interpret mode.
+    # TODO(ayx): support strided load/store in interpret mode.
     for s in idx.indices:
       if isinstance(s, Slice) and s.stride > 1:
         raise NotImplementedError("Unimplemented stride support.")
@@ -354,7 +492,12 @@ def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
     # of bounds, it will instead move the start_index backwards so the slice
     # will fit in memory.
     ref = _pad_values_to_avoid_dynamic_slice_oob_shift(ref, slice_sizes)
-    out_ones = lax.dynamic_slice(ref, slice_starts, slice_sizes=slice_sizes)
+    idx_dtype = dtypes.canonicalize_dtype(jnp.int64)
+    out_ones = lax.dynamic_slice(
+      ref,
+      [jnp.astype(s, idx_dtype) for s in slice_starts],
+      slice_sizes=slice_sizes,
+    )
     out_indexer = tuple(0 if scalar else slice(None) for scalar in scalar_dims)
     out = out_ones[out_indexer]
   elif all(not isinstance(s, Slice) for s in idx.indices):
@@ -366,11 +509,10 @@ def _load_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   return (None,) * len(in_avals), out
 
 
-state_discharge.register_discharge_rule(load_p)(_load_discharge_rule)
-
 swap_p = jax_core.Primitive('masked_swap')
 
 
+@swap_p.def_effectful_abstract_eval
 def _swap_abstract_eval(*avals_flat, args_tree, **_):
   ref, indexers, val, _ = args_tree.unflatten(avals_flat)
   expected_output_shape = indexers[-1].get_indexer_shape()
@@ -390,15 +532,13 @@ def _swap_abstract_eval(*avals_flat, args_tree, **_):
   )
 
 
-swap_p.def_effectful_abstract_eval(_swap_abstract_eval)
-
 def _swap_pp_rule(eqn, context, settings):
   # Pretty prints `a = swap x v i` as `a, x[i] <- x[i], v`
   # or:
   # Pretty prints `_ = swap x v i` as `x[i] <- v`
   y, = eqn.outvars
   x, indexers, val, mask = eqn.params["args_tree"].unflatten(eqn.invars)
-  x_i = sp.pp_ref_indexers(context, x, indexers)
+  x_i = sp.pp_ref_transforms(context, x, indexers)
   if isinstance(y, jax_core.DropVar):
     return pp.concat([
         x_i,
@@ -444,6 +584,7 @@ def _swap_jvp(primals, tangents, *, args_tree, **params):
 ad.primitive_jvps[swap_p] = _swap_jvp
 
 
+@state_discharge.register_discharge_rule(swap_p)
 def _swap_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   del out_avals  # Unused.
   ref, indexers, val, mask = args_tree.unflatten(args_flat)
@@ -451,7 +592,7 @@ def _swap_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
     raise NotImplementedError("Only one indexer supported in discharge rule.")
   idx = indexers[0]
   if all((isinstance(s, Slice) or not s.shape) for s in idx.indices):
-    # TODO(b/329733289): support strided load/store in interpret mode.
+    # TODO(ayx): support strided load/store in interpret mode.
     for s in idx.indices:
       if isinstance(s, Slice) and s.stride > 1:
         raise NotImplementedError("Unimplemented stride support.")
@@ -463,7 +604,7 @@ def _swap_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
     ]
     slice_starts = [s.start if isinstance(s, Slice) else s for s in indices]
     slice_sizes = tuple(s.size if isinstance(s, Slice) else 1 for s in indices)
-    # fixes an inconstency with lax.dynamic_update_slice where if the slice
+    # Fixes an inconsistency with lax.dynamic_update_slice where if the slice
     # goes out of bounds, it will instead move the start_index backwards so the
     # slice will fit in memory.
     ref = _pad_values_to_avoid_dynamic_slice_oob_shift(ref, slice_sizes)
@@ -488,13 +629,28 @@ def _swap_discharge_rule(in_avals, out_avals, *args_flat, args_tree, **_):
   return (x_new,) + (None,) * (len(in_avals) - 1), out
 
 
-state_discharge.register_discharge_rule(swap_p)(_swap_discharge_rule)
-
-
 def load(x_ref_or_view, idx, *, mask=None, other=None, cache_modifier=None,
          eviction_policy=None, volatile=False) -> jax.Array:
-  x_ref, indexers = sp.get_ref_and_indexers(x_ref_or_view, idx, "load")
-  args_flat, args_tree = tree_util.tree_flatten((x_ref, indexers, mask, other))
+  """Returns an array loaded from the given index.
+
+  If neither ``mask`` nor ``other`` is specified, this function has the same
+  semantics as ``x_ref_or_view[idx]`` in JAX.
+
+  Args:
+    x_ref_or_view: The ref to load from.
+    idx: The indexer to use.
+    mask: An optional boolean mask specifying which indices to load.
+      If mask is ``False`` and ``other`` is not given, no assumptions can
+      be made about the value in the resulting array.
+    other: An optional value to use for indices where mask is ``False``.
+    cache_modifier: TO BE DOCUMENTED.
+    eviction_policy: TO BE DOCUMENTED.
+    volatile: TO BE DOCUMENTED.
+  """
+  x_ref, transforms = sp.get_ref_and_transforms(x_ref_or_view, idx, "load")
+  args_flat, args_tree = tree_util.tree_flatten(
+      (x_ref, transforms, mask, other)
+  )
   return load_p.bind(
       *args_flat,
       args_tree=args_tree,
@@ -504,14 +660,27 @@ def load(x_ref_or_view, idx, *, mask=None, other=None, cache_modifier=None,
   )
 
 def swap(x_ref_or_view, idx, val, *, mask=None, eviction_policy=None,
-         _function_name="swap") -> Any:
-  x_ref, indexers = sp.get_ref_and_indexers(x_ref_or_view, idx, _function_name)
-  args_flat, args_tree = tree_util.tree_flatten((x_ref, indexers, val, mask))
+         _function_name="swap") -> jax.Array:
+  """Swaps the value at the given index and returns the old value.
+
+  See :func:`~jax.experimental.pallas.load` for the meaning of the arguments.
+
+  Returns:
+    The value stored in the ref prior to the swap.
+  """
+  x_ref, transforms = sp.get_ref_and_transforms(
+      x_ref_or_view, idx, _function_name
+  )
+  args_flat, args_tree = tree_util.tree_flatten((x_ref, transforms, val, mask))
   return swap_p.bind(
       *args_flat, args_tree=args_tree, eviction_policy=eviction_policy
   )
 
 def store(x_ref_or_view, idx, val, *, mask=None, eviction_policy=None) -> None:
+  """Stores a value at the given index.
+
+  See :func:`~jax.experimental.pallas.load` for the meaning of the arguments.
+  """
   _ = swap(x_ref_or_view, idx, val, mask=mask, eviction_policy=eviction_policy,
            _function_name="store")
 
@@ -551,31 +720,35 @@ debug_print_p = jax_core.Primitive("debug_print")
 debug_print_p.multiple_results = True
 
 
-def debug_print(fmt: str, *args: jax.ArrayLike):
-  """Prints scalar values from inside a Pallas kernel.
+def debug_print(fmt: str, *args: jax.typing.ArrayLike):
+  """Prints values from inside a Pallas kernel.
 
   Args:
     fmt: A format string to be included in the output. The restrictions on the
       format string depend on the backend:
-        * On GPU, when using Triton, ``fmt`` must not contain any placeholders
-          (``{...}``), since it is always printed before any of the values.
-        * On GPU, when using the experimental Mosaic GPU backend, ``fmt`` must
-          contain a placeholder for each value to be printed. Format specs and
-          conversions are not supported.
-        * In TPU, if ``fmt`` contains placeholders, all values must be 32-bit
-          integers. If there are no placeholders, the values are printed after
-          the format string.
-    *args: The scalar values to print.
+
+      * On GPU, when using Triton, ``fmt`` must not contain any placeholders
+        (``{...}``), since it is always printed before any of the values.
+      * On GPU, when using the experimental Mosaic GPU backend, ``fmt`` must
+        contain a placeholder for each value to be printed. Format specs and
+        conversions are not supported. All values must be scalars.
+      * On TPU, if all inputs are scalars: If ``fmt`` contains placeholders,
+        all values must be 32-bit integers. If there are no placeholders, the
+        values are printed after the format string.
+      * On TPU, if the input is a single vector, the vector is printed after
+        the format string. The format string must end with a single placeholder
+        ``{}``.
+    *args: The values to print.
   """  # fmt: skip
   has_placeholders = False
   if fmt:
-    _, field_name, *_ = next(string.Formatter().parse(fmt))
+    _, field_name, *_ = next(iter(string.Formatter().parse(fmt)))
     has_placeholders = field_name is not None
   return debug_print_p.bind(*args, fmt=fmt, has_placeholders=has_placeholders)
 
 
 def check_debug_print_format(
-    fmt: str, *args: jax.ArrayLike
+    fmt: str, *args: jax.typing.ArrayLike
 ):
   n_placeholders = 0
   for _, field, spec, conversion in string.Formatter().parse(fmt):
@@ -608,9 +781,7 @@ def debug_print_impl(*args: Any, fmt: str, has_placeholders: bool):
 
 @debug_print_p.def_effectful_abstract_eval
 def debug_print_abstract_eval(*avals: Any, fmt: str, has_placeholders: bool):
-  del fmt, has_placeholders
-  if any(aval.shape for aval in avals):
-    raise ValueError("Only scalar values are supported")
+  del avals, fmt, has_placeholders  # Unused.
   return [], {debug_print_effect}
 
 
@@ -650,3 +821,127 @@ def debug_print_lowering_rule(ctx, *args, **params):
       has_side_effect=True,
   )
   return result
+
+
+# All of those shenanigans are because we can't make TransformedRef a PyTree,
+# because they should appear as atomic JAX values to the users.
+# TODO(apaszke): This can be deleted once we make transforms in Mosaic GPU
+# inferred by the compiler.
+@lu.transformation2
+def wrap_with_transforms(f, transforms, *args):
+  new_args = tuple(
+      state_types.TransformedRef(a, t) if t else a
+      for a, t in zip(args, transforms)
+  )
+  return f(*new_args)
+
+
+run_scoped_p = jax_core.Primitive("run_scoped")
+run_scoped_p.multiple_results = True
+
+
+def run_scoped(f: Callable[..., Any], *types: Any, **kw_types: Any) -> Any:
+  """Calls the function with allocated references and returns the result.
+
+  The positional and keyword arguments describe which reference types
+  to allocate for each argument. Each backend has its own set of reference
+  types in addition to :class:`jax.experimental.pallas.MemoryRef`.
+  """
+  flat_types, in_tree = tree_util.tree_flatten((types, kw_types))
+  flat_fun, out_tree_thunk = api_util.flatten_fun(lu.wrap_init(f), in_tree)
+  # We allow ref avals to be transformed references.
+  ref_avals = [t.get_ref_aval() for t in flat_types]
+  avals = [
+      t.ref if isinstance(t, state_types.TransformedRef) else t
+      for t in ref_avals
+  ]
+  ref_transforms = tuple(
+      t.transforms if isinstance(t, state_types.TransformedRef) else ()
+      for t in ref_avals
+  )
+  flat_fun = wrap_with_transforms(flat_fun, ref_transforms)
+  # Turn the function into a jaxpr. The body of run_scoped may have
+  # effects (IO) on constvars (i.e. variables inherited from the
+  # parent scope). Jax can't reason about effects to references that
+  # are not in the invars of an operation so we just put them all
+  # there.
+  jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(flat_fun, avals)
+  out = run_scoped_p.bind(*consts, jaxpr=jaxpr)
+  return tree_util.tree_unflatten(out_tree_thunk(), out)
+
+
+@run_scoped_p.def_effectful_abstract_eval
+def _run_scoped_abstract_eval(*args, jaxpr):
+  del args
+  # jaxpr will have effects for its inputs (Refs that are allocated) and for
+  # constvars (closed over Refs). The effects for the allocated Refs are local
+  # to the jaxpr and shouldn't propagate out.
+  nonlocal_effects = {
+      eff
+      for eff in jaxpr.effects
+      if not (
+          isinstance(eff, effects.JaxprInputEffect)
+          and eff.input_index >= len(jaxpr.constvars)
+      )
+  }
+  return [v.aval for v in jaxpr.outvars], nonlocal_effects
+
+
+def _run_scoped_discharge_rule(
+    should_discharge,
+    in_avals,
+    out_avals,
+    *args_flat,
+    jaxpr,
+    **_):
+  del out_avals
+  num_consts = len(args_flat)
+  jaxpr_noconst = pe.convert_constvars_jaxpr(jaxpr)
+  num_return_values = len(jaxpr_noconst.outvars)
+  should_discharge = should_discharge + [
+      isinstance(var.aval, state.AbstractRef) for var in jaxpr.invars
+  ]
+  discharged_body, new_consts = state_discharge.discharge_state(
+      jaxpr_noconst, [], should_discharge=should_discharge)
+  if new_consts:
+    raise NotImplementedError(
+        "Cannot handle new consts created by state discharge.")
+  # Create inputs filled with uninitialized values to the body.
+  body_avals = [v.aval for v in discharged_body.invars[num_consts:]]
+  init_vals = [uninitialized_value(
+      aval.shape, aval.dtype) for aval in body_avals]
+  init_vals_with_consts = args_flat + tuple(init_vals)
+  out = jax_core.eval_jaxpr(discharged_body, [], *init_vals_with_consts)
+  # Order of outputs:
+  # (1) return values, (2) closed refs, (3) scoped refs.
+  return_values = out[:num_return_values]
+  ref_outputs = out[num_return_values:]
+  # We update all ref values with their updated values from the discharged
+  # body. For other values we leave them in place.
+  updates = [
+      ref_outputs.pop(0) if isinstance(aval, pallas_core.AbstractMemoryRef)
+      else None for aval in in_avals]
+  assert len(updates) == len(in_avals), f'{len(updates)} != {len(in_avals)}'
+  return updates, return_values
+
+
+state_discharge.register_partial_discharge_rule(run_scoped_p)(
+    _run_scoped_discharge_rule)
+
+
+@functools.partial(mlir.register_lowering, run_scoped_p)
+def _run_scoped_lowering_rule(ctx, *args, jaxpr):
+  # This lowering rule gets triggered when run_scoped is not discharged.
+  # In this case there are no stateful effects to handle.
+  should_discharge = [
+      isinstance(aval, state.AbstractRef) for aval in ctx.avals_in
+  ]
+
+  def _lower_fun(*lower_fun_args):
+    updates, out = _run_scoped_discharge_rule(
+        should_discharge,
+        [], [], *lower_fun_args,
+        jaxpr=jaxpr)
+    assert len(updates) == 0, 'Cannot lower run_scoped with effects.'
+    return out
+  return mlir.lower_fun(_lower_fun, multiple_results=True)(ctx, *args)
