@@ -78,7 +78,7 @@ def cholesky(a: ArrayLike, lower: bool = False, overwrite_a: bool = False,
    - :func:`jax.scipy.linalg.cho_factor`
    - :func:`jax.scipy.linalg.cho_solve`
 
-  Example:
+  Examples:
     A small real Hermitian positive-definite matrix:
 
     >>> x = jnp.array([[2., 1.],
@@ -130,7 +130,7 @@ def cho_factor(a: ArrayLike, lower: bool = False, overwrite_a: bool = False,
     - :func:`jax.scipy.linalg.cholesky`
     - :func:`jax.scipy.linalg.cho_solve`
 
-  Example:
+  Examples:
     A small real Hermitian positive-definite matrix:
 
     >>> x = jnp.array([[2., 1.],
@@ -186,7 +186,7 @@ def cho_solve(c_and_lower: tuple[ArrayLike, bool], b: ArrayLike,
     - :func:`jax.scipy.linalg.cholesky`
     - :func:`jax.scipy.linalg.cho_factor`
 
-  Example:
+  Examples:
     A small real Hermitian positive-definite matrix:
 
     >>> x = jnp.array([[2., 1.],
@@ -288,7 +288,7 @@ def svd(a: ArrayLike, full_matrices: bool = True, compute_uv: bool = True,
     - :func:`jax.numpy.linalg.svd`: NumPy-style SVD API
     - :func:`jax.lax.linalg.svd`: XLA-style SVD API
 
-  Example:
+  Examples:
     Consider the SVD of a small real-valued array:
 
     >>> x = jnp.array([[1., 2., 3.],
@@ -415,7 +415,7 @@ def eigh(a: ArrayLike, b: ArrayLike | None = None, lower: bool = True,
          type: int = 1, check_finite: bool = True) -> Array | tuple[Array, Array]:
   """Compute eigenvalues and eigenvectors for a Hermitian matrix
 
-  JAX implementation of :func:`jax.scipy.linalg.eigh`.
+  JAX implementation of :func:`scipy.linalg.eigh`.
 
   Args:
     a: Hermitian input array of shape ``(..., N, N)``
@@ -516,7 +516,7 @@ def schur(a: ArrayLike, output: str = 'real') -> tuple[Array, Array]:
     - :func:`jax.scipy.linalg.rsf2csf`: convert real Schur form to complex Schur form.
     - :func:`jax.lax.linalg.schur`: XLA-style API for Schur decomposition.
 
-  Example:
+  Examples:
     A Schur decomposition of a 3x3 matrix:
 
     >>> a = jnp.array([[1., 2., 3.],
@@ -570,7 +570,7 @@ def inv(a: ArrayLike, overwrite_a: bool = False, check_finite: bool = True) -> A
     - :func:`jax.numpy.linalg.inv`: NumPy-style API for matrix inverse
     - :func:`jax.scipy.linalg.solve`: direct linear solver
 
-  Example:
+  Examples:
     Compute the inverse of a 3x3 matrix
 
     >>> a = jnp.array([[1., 2., 3.],
@@ -630,7 +630,7 @@ def lu_factor(a: ArrayLike, overwrite_a: bool = False, check_finite: bool = True
     - :func:`jax.scipy.linalg.lu`
     - :func:`jax.scipy.linalg.lu_solve`
 
-  Example:
+  Examples:
     Solving a small linear system via LU factorization:
 
     >>> a = jnp.array([[2., 1.],
@@ -686,7 +686,7 @@ def lu_solve(lu_and_piv: tuple[Array, ArrayLike], b: ArrayLike, trans: int = 0,
     - :func:`jax.scipy.linalg.lu`
     - :func:`jax.scipy.linalg.lu_factor`
 
-  Example:
+  Examples:
     Solving a small linear system via LU factorization:
 
     >>> a = jnp.array([[2., 1.],
@@ -784,7 +784,7 @@ def lu(a: ArrayLike, permute_l: bool = False, overwrite_a: bool = False,
     - :func:`jax.lax.linalg.lu`: XLA-style API for LU decomposition.
     - :func:`jax.scipy.linalg.lu_solve`: LU-based linear solver.
 
-  Example:
+  Examples:
     An LU decomposition of a 3x3 matrix:
 
     >>> a = jnp.array([[1., 2., 3.],
@@ -820,20 +820,39 @@ def lu(a: ArrayLike, permute_l: bool = False, overwrite_a: bool = False,
   del overwrite_a, check_finite  # unused
   return _lu(a, permute_l)
 
-@overload
-def _qr(a: ArrayLike, mode: Literal["r"], pivoting: bool) -> tuple[Array]: ...
 
 @overload
-def _qr(a: ArrayLike, mode: Literal["full", "economic"], pivoting: bool) -> tuple[Array, Array]: ...
+def _qr(a: ArrayLike, mode: Literal["r"], pivoting: Literal[False]
+       ) -> tuple[Array]: ...
 
 @overload
-def _qr(a: ArrayLike, mode: str, pivoting: bool) -> tuple[Array] | tuple[Array, Array]: ...
+def _qr(a: ArrayLike, mode: Literal["r"], pivoting: Literal[True]
+       ) -> tuple[Array, Array]: ...
+
+@overload
+def _qr(a: ArrayLike, mode: Literal["full", "economic"], pivoting: Literal[False]
+       ) -> tuple[Array, Array]: ...
+
+@overload
+def _qr(a: ArrayLike, mode: Literal["full", "economic"], pivoting: Literal[True]
+       ) -> tuple[Array, Array, Array]: ...
+
+@overload
+def _qr(a: ArrayLike, mode: str, pivoting: Literal[False]
+       ) -> tuple[Array] | tuple[Array, Array]: ...
+
+@overload
+def _qr(a: ArrayLike, mode: str, pivoting: Literal[True]
+       ) -> tuple[Array, Array] | tuple[Array, Array, Array]: ...
+
+@overload
+def _qr(a: ArrayLike, mode: str, pivoting: bool
+       ) -> tuple[Array] | tuple[Array, Array] | tuple[Array, Array, Array]: ...
+
 
 @partial(jit, static_argnames=('mode', 'pivoting'))
-def _qr(a: ArrayLike, mode: str, pivoting: bool) -> tuple[Array] | tuple[Array, Array]:
-  if pivoting:
-    raise NotImplementedError(
-        "The pivoting=True case of qr is not implemented.")
+def _qr(a: ArrayLike, mode: str, pivoting: bool
+       ) -> tuple[Array] | tuple[Array, Array] | tuple[Array, Array, Array]:
   if mode in ("full", "r"):
     full_matrices = True
   elif mode == "economic":
@@ -841,31 +860,56 @@ def _qr(a: ArrayLike, mode: str, pivoting: bool) -> tuple[Array] | tuple[Array,
   else:
     raise ValueError(f"Unsupported QR decomposition mode '{mode}'")
   a, = promote_dtypes_inexact(jnp.asarray(a))
-  q, r = lax_linalg.qr(a, full_matrices=full_matrices)
+  q, r, *p = lax_linalg.qr(a, pivoting=pivoting, full_matrices=full_matrices)
   if mode == "r":
+    if pivoting:
+      return r, p[0]
     return (r,)
+  if pivoting:
+    return q, r, p[0]
   return q, r
 
 
 @overload
-def qr(a: ArrayLike, overwrite_a: bool = False, lwork: Any = None, mode: Literal["full", "economic"] = "full",
-       pivoting: bool = False, check_finite: bool = True) -> tuple[Array, Array]: ...
+def qr(a: ArrayLike,  overwrite_a: bool = False, lwork: Any = None, *,
+       mode: Literal["full", "economic"], pivoting: Literal[False] = False,
+       check_finite: bool = True) -> tuple[Array, Array]: ...
 
 @overload
-def qr(a: ArrayLike,  overwrite_a: bool, lwork: Any, mode: Literal["r"],
-       pivoting: bool = False, check_finite: bool = True) -> tuple[Array]: ...
+def qr(a: ArrayLike,  overwrite_a: bool = False, lwork: Any = None, *,
+       mode: Literal["full", "economic"], pivoting: Literal[True] = True,
+       check_finite: bool = True) -> tuple[Array, Array, Array]: ...
 
 @overload
-def qr(a: ArrayLike,  overwrite_a: bool = False, lwork: Any = None, *, mode: Literal["r"],
-       pivoting: bool = False, check_finite: bool = True) -> tuple[Array]: ...
+def qr(a: ArrayLike,  overwrite_a: bool = False, lwork: Any = None, *,
+       mode: Literal["full", "economic"], pivoting: bool = False,
+       check_finite: bool = True
+      ) -> tuple[Array, Array] | tuple[Array, Array, Array]: ...
+
+@overload
+def qr(a: ArrayLike,  overwrite_a: bool = False, lwork: Any = None, *,
+       mode: Literal["r"], pivoting: Literal[False] = False, check_finite: bool = True
+      ) -> tuple[Array]: ...
+
+@overload
+def qr(a: ArrayLike,  overwrite_a: bool = False, lwork: Any = None, *,
+       mode: Literal["r"], pivoting: Literal[True] = True, check_finite: bool = True
+      ) -> tuple[Array, Array]: ...
+
+@overload
+def qr(a: ArrayLike,  overwrite_a: bool = False, lwork: Any = None, *,
+       mode: Literal["r"], pivoting: bool = False, check_finite: bool = True
+      ) -> tuple[Array] | tuple[Array, Array]: ...
 
 @overload
 def qr(a: ArrayLike, overwrite_a: bool = False, lwork: Any = None, mode: str = "full",
-       pivoting: bool = False, check_finite: bool = True) -> tuple[Array] | tuple[Array, Array]: ...
+       pivoting: bool = False, check_finite: bool = True
+      ) -> tuple[Array] | tuple[Array, Array] | tuple[Array, Array, Array]: ...
 
 
 def qr(a: ArrayLike, overwrite_a: bool = False, lwork: Any = None, mode: str = "full",
-       pivoting: bool = False, check_finite: bool = True) -> tuple[Array] | tuple[Array, Array]:
+       pivoting: bool = False, check_finite: bool = True
+      ) -> tuple[Array] | tuple[Array, Array] | tuple[Array, Array, Array]:
   """Compute the QR decomposition of an array
 
   JAX implementation of :func:`scipy.linalg.qr`.
@@ -888,22 +932,29 @@ def qr(a: ArrayLike, overwrite_a: bool = False, lwork: Any = None, mode: str = "
       - ``"economic"``: return `Q` of shape ``(M, K)`` and `R` of shape ``(K, N)``,
         where K = min(M, N).
 
-    pivoting: Not implemented in JAX.
+    pivoting: Allows the QR decomposition to be rank-revealing. If ``True``, compute
+      the column-pivoted decomposition ``A[:, P] = Q @ R``, where ``P`` is chosen such
+      that the diagonal of ``R`` is non-increasing.
     overwrite_a: unused in JAX
     lwork: unused in JAX
     check_finite: unused in JAX
 
   Returns:
-    A tuple ``(Q, R)`` (if ``mode`` is not ``"r"``) otherwise an array ``R``,
-    where:
+    A tuple ``(Q, R)`` or ``(Q, R, P)``, if ``mode`` is not ``"r"`` and ``pivoting`` is
+    respectively ``False`` or ``True``, otherwise an array ``R`` or tuple ``(R, P)`` if
+    mode is ``"r"``, and ``pivoting`` is respectively ``False`` or ``True``, where:
 
     - ``Q`` is an orthogonal matrix of shape ``(..., M, M)`` (if ``mode`` is ``"full"``)
-      or ``(..., M, K)`` (if ``mode`` is ``"economic"``).
+      or ``(..., M, K)`` (if ``mode`` is ``"economic"``),
     - ``R`` is an upper-triangular matrix of shape ``(..., M, N)`` (if ``mode`` is
-      ``"r"`` or ``"full"``) or ``(..., K, N)`` (if ``mode`` is ``"economic"``)
+      ``"r"`` or ``"full"``) or ``(..., K, N)`` (if ``mode`` is ``"economic"``),
+    - ``P`` is an index vector of shape ``(..., N)``.
 
     with ``K = min(M, N)``.
 
+  Notes:
+    - At present, pivoting is only implemented on CPU backends.
+
   See also:
     - :func:`jax.numpy.linalg.qr`: NumPy-style QR decomposition API
     - :func:`jax.lax.linalg.qr`: XLA-style QR decomposition API
@@ -951,7 +1002,7 @@ def _solve(a: ArrayLike, b: ArrayLike, assume_a: str, lower: bool) -> Array:
   factors = cho_factor(lax.stop_gradient(a), lower=lower)
   custom_solve = partial(
       lax.custom_linear_solve,
-      lambda x: lax_linalg._matvec_multiply(a, x),
+      lambda x: lax_linalg._broadcasted_matvec(a, x),
       solve=lambda _, x: cho_solve(factors, x),
       symmetric=True)
   if a.ndim == b.ndim + 1:
@@ -999,7 +1050,7 @@ def solve(a: ArrayLike, b: ArrayLike, lower: bool = False,
     - :func:`jax.numpy.linalg.solve`: NumPy-style API for solving linear systems.
     - :func:`jax.lax.custom_linear_solve`: matrix-free linear solver.
 
-  Example:
+  Examples:
     A simple 3x3 linear system:
 
     >>> A = jnp.array([[1., 2., 3.],
@@ -1083,7 +1134,7 @@ def solve_triangular(a: ArrayLike, b: ArrayLike, trans: int | str = 0, lower: bo
   See also:
     :func:`jax.scipy.linalg.solve`: Solve a general linear system.
 
-  Example:
+  Examples:
     A simple 3x3 triangular linear system:
 
     >>> A = jnp.array([[1., 2., 3.],
@@ -1137,6 +1188,32 @@ def expm(A: ArrayLike, *, upper_triangular: bool = False, max_squarings: int = 1
 
   See Also:
     :func:`jax.scipy.linalg.expm_frechet`
+
+  Examples:
+
+    ``expm`` is the matrix exponential, and has similar properties to the more
+    familiar scalar exponential. For scalars ``a`` and ``b``, :math:`e^{a + b}
+    = e^a e^b`. However, for matrices, this property only holds when ``A`` and
+    ``B`` commute (``AB = BA``). In this case, ``expm(A+B) = expm(A) @ expm(B)``
+
+    >>> A = jnp.array([[2, 0],
+    ...                [0, 1]])
+    >>> B = jnp.array([[3, 0],
+    ...                [0, 4]])
+    >>> jnp.allclose(jax.scipy.linalg.expm(A+B),
+    ...              jax.scipy.linalg.expm(A) @ jax.scipy.linalg.expm(B),
+    ...              rtol=0.0001)
+    Array(True, dtype=bool)
+
+    If a matrix ``X`` is invertible, then
+    ``expm(X @ A @ inv(X)) = X @ expm(A) @ inv(X)``
+
+    >>> X = jnp.array([[3, 1],
+    ...                [2, 5]])
+    >>> X_inv = jax.scipy.linalg.inv(X)
+    >>> jnp.allclose(jax.scipy.linalg.expm(X @ A @ X_inv),
+    ...              X @ jax.scipy.linalg.expm(A) @ X_inv)
+    Array(True, dtype=bool)
   """
   A, = promote_dtypes_inexact(A)
 
@@ -1358,7 +1435,7 @@ def block_diag(*arrs: ArrayLike) -> Array:
     2D block-diagonal array constructed by placing the input arrays
     along the diagonal.
 
-  Example:
+  Examples:
     >>> A = jnp.ones((1, 1))
     >>> B = jnp.ones((2, 2))
     >>> C = jnp.ones((3, 3))
@@ -1642,6 +1719,36 @@ def polar(a: ArrayLike, side: str = 'right', *, method: str = 'qdwh', eps: float
     ``posdef`` is either :math:`n \times n` or :math:`m \times m` depending on
     whether ``side`` is ``"right"`` or ``"left"``, respectively.
 
+  Examples:
+
+    Polar decomposition of a 3x3 matrix:
+
+    >>> a = jnp.array([[1., 2., 3.],
+    ...                [5., 4., 2.],
+    ...                [3., 2., 1.]])
+    >>> U, P = jax.scipy.linalg.polar(a)
+
+    U is a Unitary Matrix:
+
+    >>> jnp.round(U.T @ U)  # doctest: +SKIP
+    Array([[ 1., -0., -0.],
+           [-0.,  1.,  0.],
+           [-0.,  0.,  1.]], dtype=float32)
+
+    P is positive-semidefinite Matrix:
+
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...     print(P)
+    [[4.79 3.25 1.23]
+     [3.25 3.06 2.01]
+     [1.23 2.01 2.91]]
+
+    The original matrix can be reconstructed by multiplying the U and P:
+
+    >>> a_reconstructed = U @ P
+    >>> jnp.allclose(a, a_reconstructed)
+    Array(True, dtype=bool)
+
   .. _QDWH: https://epubs.siam.org/doi/abs/10.1137/090774999
   """
   arr = jnp.asarray(a)
@@ -1730,7 +1837,7 @@ def sqrtm(A: ArrayLike, blocksize: int = 1) -> Array:
   See Also:
     :func:`jax.scipy.linalg.expm`
 
-  Example:
+  Examples:
     >>> a = jnp.array([[1., 2., 3.],
     ...                [2., 4., 2.],
     ...                [3., 2., 1.]])
@@ -1780,7 +1887,7 @@ def rsf2csf(T: ArrayLike, Z: ArrayLike, check_finite: bool = True) -> tuple[Arra
   See Also:
     :func:`jax.scipy.linalg.schur`: Schur decomposition
 
-  Example:
+  Examples:
     >>> A = jnp.array([[0., 3., 3.],
     ...                [0., 1., 2.],
     ...                [2., 0., 1.]])
@@ -1905,7 +2012,7 @@ def hessenberg(a: ArrayLike, *, calc_q: bool = False, overwrite_a: bool = False,
     - ``H`` has shape ``(..., N, N)`` and is the Hessenberg form of ``a``
     - ``Q`` has shape ``(..., N, N)`` and is the associated unitary matrix
 
-  Example:
+  Examples:
     Computing the Hessenberg form of a 4x4 matrix
 
     >>> a = jnp.array([[1., 2., 3., 4.],
@@ -1948,7 +2055,7 @@ def hessenberg(a: ArrayLike, *, calc_q: bool = False, overwrite_a: bool = False,
 
 
 def toeplitz(c: ArrayLike, r: ArrayLike | None = None) -> Array:
-  r"""Construct a Toeplitz matrix
+  r"""Construct a Toeplitz matrix.
 
   JAX implementation of :func:`scipy.linalg.toeplitz`.
 
@@ -1967,13 +2074,13 @@ def toeplitz(c: ArrayLike, r: ArrayLike | None = None) -> Array:
   Notice this implies that :math:`r_0` is ignored.
 
   Args:
-    c: array specifying the first column. Will be flattened
-      if not 1-dimensional.
-    r: (optional) array specifying the first row. If not specified, defaults
-      to ``conj(c)``. Will be flattened if not 1-dimensional.
+    c: array of shape ``(..., N)`` specifying the first column.
+    r: (optional) array of shape ``(..., M)`` specifying the first row. Leading
+      dimensions must be broadcast-compatible with those of ``c``. If not specified,
+      ``r`` defaults to ``conj(c)``.
 
   Returns:
-    toeplitz matrix of shape ``(c.size, r.size)``.
+    A Toeplitz matrix of shape ``(... N, M)``.
 
   Examples:
     Specifying ``c`` only:
@@ -2003,32 +2110,40 @@ def toeplitz(c: ArrayLike, r: ArrayLike | None = None) -> Array:
            [1.+2.j, 2.+1.j, 1.+0.j]], dtype=complex64)
     >>> print("M is Hermitian:", jnp.all(M == M.conj().T))
     M is Hermitian: True
+
+    For N-dimensional ``c`` and/or ``r``, the result is a batch of Toeplitz matrices:
+
+    >>> c = jnp.array([[1, 2, 3], [4, 5, 6]])
+    >>> jax.scipy.linalg.toeplitz(c)
+    Array([[[1, 2, 3],
+            [2, 1, 2],
+            [3, 2, 1]],
+    <BLANKLINE>
+           [[4, 5, 6],
+            [5, 4, 5],
+            [6, 5, 4]]], dtype=int32)
   """
   if r is None:
     check_arraylike("toeplitz", c)
     r = jnp.conjugate(jnp.asarray(c))
   else:
     check_arraylike("toeplitz", c, r)
+  return _toeplitz(jnp.atleast_1d(jnp.asarray(c)), jnp.atleast_1d(jnp.asarray(r)))
 
-  c_arr = jnp.asarray(c).flatten()
-  r_arr = jnp.asarray(r).flatten()
-
-  ncols, = c_arr.shape
-  nrows, = r_arr.shape
-
+@partial(jnp.vectorize, signature="(m),(n)->(m,n)")
+def _toeplitz(c: Array, r: Array) -> Array:
+  ncols, = c.shape
+  nrows, = r.shape
   if ncols == 0 or nrows == 0:
-    return jnp.empty((ncols, nrows),
-                     dtype=jnp.promote_types(c_arr.dtype, r_arr.dtype))
-
+    return jnp.empty((ncols, nrows), dtype=jnp.promote_types(c.dtype, r.dtype))
   nelems = ncols + nrows - 1
-  elems = jnp.concatenate((c_arr[::-1], r_arr[1:]))
+  elems = jnp.concatenate((c[::-1], r[1:]))
   patches = lax.conv_general_dilated_patches(
       elems.reshape((1, nelems, 1)),
       (nrows,), (1,), 'VALID', dimension_numbers=('NTC', 'IOT', 'NTC'),
       precision=lax.Precision.HIGHEST)[0]
   return jnp.flip(patches, axis=0)
 
-
 @partial(jit, static_argnames=("n",))
 def hilbert(n: int) -> Array:
   r"""Create a Hilbert matrix of order n.
