@@ -46,6 +46,40 @@ def mode(a: ArrayLike, axis: int | None = 0, nan_policy: str = "propagate", keep
   Returns:
     A tuple of arrays, ``(mode, count)``. ``mode`` is the array of modal values,
     and ``count`` is the number of times each value appears in the input array.
+
+  Examples:
+    >>> x = jnp.array([2, 4, 1, 1, 3, 4, 4, 2, 3])
+    >>> mode, count = jax.scipy.stats.mode(x)
+    >>> mode, count
+    (Array(4, dtype=int32), Array(3, dtype=int32))
+
+    For multi dimensional arrays, ``jax.scipy.stats.mode`` computes the ``mode``
+    and the corresponding ``count`` along ``axis=0``:
+
+    >>> x1 = jnp.array([[1, 2, 1, 3, 2, 1],
+    ...                 [3, 1, 3, 2, 1, 3],
+    ...                 [1, 2, 2, 3, 1, 2]])
+    >>> mode, count = jax.scipy.stats.mode(x1)
+    >>> mode, count
+    (Array([1, 2, 1, 3, 1, 1], dtype=int32), Array([2, 2, 1, 2, 2, 1], dtype=int32))
+
+    If ``axis=1``, ``mode`` and ``count`` will be computed along ``axis 1``.
+
+    >>> mode, count = jax.scipy.stats.mode(x1, axis=1)
+    >>> mode, count
+    (Array([1, 3, 2], dtype=int32), Array([3, 3, 3], dtype=int32))
+
+    By default, ``jax.scipy.stats.mode`` reduces the dimension of the result.
+    To keep the dimensions same as that of the input array, the argument
+    ``keepdims`` must be set to ``True``.
+
+    >>> mode, count = jax.scipy.stats.mode(x1, axis=1, keepdims=True)
+    >>> mode, count
+    (Array([[1],
+           [3],
+           [2]], dtype=int32), Array([[3],
+           [3],
+           [3]], dtype=int32))
   """
   check_arraylike("mode", a)
   x = jnp.atleast_1d(a)
@@ -164,17 +198,16 @@ def rankdata(
     return jnp.apply_along_axis(rankdata, axis, a, method)
 
   arr = jnp.ravel(a)
-  sorter = jnp.argsort(arr)
+  arr, sorter = jax.lax.sort_key_val(arr, jnp.arange(arr.size))
   inv = invert_permutation(sorter)
 
   if method == "ordinal":
     return inv + 1
-  arr = arr[sorter]
-  obs = jnp.insert(arr[1:] != arr[:-1], 0, True)
+  obs = jnp.concatenate([jnp.array([True]), arr[1:] != arr[:-1]])
   dense = obs.cumsum()[inv]
   if method == "dense":
     return dense
-  count = jnp.nonzero(obs, size=arr.size + 1, fill_value=len(obs))[0]
+  count = jnp.nonzero(obs, size=arr.size + 1, fill_value=obs.size)[0]
   if method == "max":
     return count[dense]
   if method == "min":
@@ -201,15 +234,69 @@ def sem(a: ArrayLike, axis: int | None = 0, ddof: int = 1, nan_policy: str = "pr
 
   Returns:
     array
+
+  Examples:
+    >>> x = jnp.array([2, 4, 1, 1, 3, 4, 4, 2, 3])
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...   jax.scipy.stats.sem(x)
+    Array(0.41, dtype=float32)
+
+    For multi dimensional arrays, ``sem`` computes standard error of mean along
+    ``axis=0``:
+
+    >>> x1 = jnp.array([[1, 2, 1, 3, 2, 1],
+    ...                 [3, 1, 3, 2, 1, 3],
+    ...                 [1, 2, 2, 3, 1, 2]])
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...   jax.scipy.stats.sem(x1)
+    Array([0.67, 0.33, 0.58, 0.33, 0.33, 0.58], dtype=float32)
+
+    If ``axis=1``, standard error of mean will be computed along ``axis 1``.
+
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...   jax.scipy.stats.sem(x1, axis=1)
+    Array([0.33, 0.4 , 0.31], dtype=float32)
+
+    If ``axis=None``, standard error of mean will be computed along all the axes.
+
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...   jax.scipy.stats.sem(x1, axis=None)
+    Array(0.2, dtype=float32)
+
+    By default, ``sem`` reduces the dimension of the result. To keep the
+    dimensions same as that of the input array, the argument ``keepdims`` must
+    be set to ``True``.
+
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...   jax.scipy.stats.sem(x1, axis=1, keepdims=True)
+    Array([[0.33],
+           [0.4 ],
+           [0.31]], dtype=float32)
+
+    Since, by default, ``nan_policy='propagate'``, ``sem`` propagates the ``nan``
+    values in the result.
+
+    >>> nan = jnp.nan
+    >>> x2 = jnp.array([[1, 2, 3, nan, 4, 2],
+    ...                 [4, 5, 4, 3, nan, 1],
+    ...                 [7, nan, 8, 7, 9, nan]])
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...   jax.scipy.stats.sem(x2)
+    Array([1.73,  nan, 1.53,  nan,  nan,  nan], dtype=float32)
+
+    If ``nan_policy='omit```, ``sem`` omits the ``nan`` values and computes the error
+    for the remainging values along the specified axis.
+
+    >>> with jnp.printoptions(precision=2, suppress=True):
+    ...   jax.scipy.stats.sem(x2, nan_policy='omit')
+    Array([1.73, 1.5 , 1.53, 2.  , 2.5 , 0.5 ], dtype=float32)
   """
   b, = promote_args_inexact("sem", a)
-  if axis is None:
-    b = b.ravel()
-    axis = 0
   if nan_policy == "propagate":
-    return b.std(axis, ddof=ddof) / jnp.sqrt(b.shape[axis]).astype(b.dtype)
+    size = b.size if axis is None else b.shape[axis]
+    return b.std(axis, ddof=ddof, keepdims=keepdims) / jnp.sqrt(size).astype(b.dtype)
   elif nan_policy == "omit":
-    count = (~jnp.isnan(b)).sum(axis)
-    return jnp.nanstd(b, axis, ddof=ddof) / jnp.sqrt(count).astype(b.dtype)
+    count = (~jnp.isnan(b)).sum(axis, keepdims=keepdims)
+    return jnp.nanstd(b, axis, ddof=ddof, keepdims=keepdims) / jnp.sqrt(count).astype(b.dtype)
   else:
     raise ValueError(f"{nan_policy} is not supported")
