@@ -30,10 +30,10 @@ executable protocols described above.
 """
 from __future__ import annotations
 
+import functools
 from collections.abc import Sequence
 from dataclasses import dataclass
-from typing import Any, NamedTuple, Protocol, Union
-import warnings
+from typing import Any, NamedTuple, Protocol, Union, runtime_checkable
 
 import jax
 
@@ -42,9 +42,9 @@ from jax._src import config
 from jax._src import source_info_util
 from jax._src import traceback_util
 from jax._src import tree_util
-from jax._src.tree_util import tree_unflatten, keystr
 from jax._src import util
-from jax._src.sharding_impls import is_unspecified_or_auto
+from jax._src import mesh as mesh_lib
+from jax._src.sharding_impls import UnspecifiedValue, AUTO
 from jax._src.layout import Layout
 from jax._src.interpreters import mlir
 from jax._src.lib.mlir import ir
@@ -162,7 +162,8 @@ class Lowering(Protocol):
     """Compile and return a corresponding ``Executable``."""
     raise NotImplementedError
 
-  def as_text(self, dialect: str | None = None) -> str:
+  def as_text(self, dialect: str | None = None, *,
+              debug_info: bool = False) -> str:
     """A human-readable text representation of this lowering.
 
     Intended for visualization and debugging purposes. This need not be a valid
@@ -249,15 +250,13 @@ class XlaExecutable(Executable):
       else:
         raise
 
-    # TODO(skyewm): this should return a single dict (I think returning a list
-    # was to support MPMD executables, which never fully landed)
-  def cost_analysis(self) -> list[dict[str, float]]:
+  def cost_analysis(self) -> dict[str, float]:
     xla_ext_exe = self.xla_extension_executable()
 
     # TODO(b/259255524): Unify/merge the two cost_analysis calls below.
     if hasattr(xla_ext_exe, "cost_analysis"):
       try:
-        return [xla_ext_exe.cost_analysis()]
+        return xla_ext_exe.cost_analysis()
       except xla_extension.XlaRuntimeError as e:
         msg, *_ = e.args
         if not (type(msg) is str and msg.startswith("UNIMPLEMENTED")):
@@ -266,13 +265,24 @@ class XlaExecutable(Executable):
     # Try client method if executable cost_analysis method is unimplemented
     if hasattr(xla_ext_exe, "client"):
       try:
-        return [
-            xla_extension.hlo_module_cost_analysis(xla_ext_exe.client, m)
-            for m in xla_ext_exe.hlo_modules()
-        ]
+        # TODO(b/384741132): We expect that the executable has only one
+        # HloModule. We should be able to remove this check once we update the
+        # Executable class to have only a single HloModule (see bug).
+        hlo_modules = xla_ext_exe.hlo_modules()
+        assert len(hlo_modules) == 1, (
+            f"Exectuable should have only one HloModule ({len(hlo_modules)})"
+            " were found)."
+        )
+
+        return xla_extension.hlo_module_cost_analysis(
+            xla_ext_exe.client, hlo_modules[0]
+        )
       except xla_extension.XlaRuntimeError as e:
         msg, *_ = e.args
-        if not (type(msg) is str and msg.startswith("UNIMPLEMENTED")):
+        supported = not (type(msg) is str and
+                         (msg.startswith("UNIMPLEMENTED") or
+                          "PjRt-compatible backend only" in msg))
+        if supported:
           raise
 
     if (
@@ -281,7 +291,7 @@ class XlaExecutable(Executable):
         and hasattr(self.unsafe_call, "compiled")
         and hasattr(self.unsafe_call.compiled, "cost_analysis")
     ):
-      return [self.unsafe_call.compiled.cost_analysis()]
+      return self.unsafe_call.compiled.cost_analysis()
 
     raise NotImplementedError(
         f"cost analysis unsupported on current XLA backend: {type(xla_ext_exe)}"
@@ -314,23 +324,11 @@ class XlaLowering(Lowering):
   def hlo(self) -> xc.XlaComputation:
     """Return an HLO representation of this computation."""
     hlo = self.stablehlo()
-    m: Union[str, bytes]
+    m: str | bytes
     m = mlir.module_to_bytecode(hlo)
     return xla_extension.mlir.mlir_module_to_xla_computation(
         m, use_tuple_args=self.compile_args["tuple_args"])
 
-  def mhlo(self) -> ir.Module:
-    """Return an MHLO representation of this computation."""
-    warnings.warn(
-        "mhlo support is deprecated and will be removed "
-        "from a future release of JAX. Use stablehlo instead.",
-        DeprecationWarning,
-    )
-    module_str = xla_extension.mlir.stablehlo_to_mhlo(
-        mlir.module_to_bytecode(self.stablehlo()))
-    with self.stablehlo().context:
-      return ir.Module.parse(module_str)
-
   def stablehlo(self) -> ir.Module:
     """Return a StableHLO representation of this computation."""
     raise NotImplementedError("must override")
@@ -339,24 +337,25 @@ class XlaLowering(Lowering):
       self, compiler_options: CompilerOptions | None = None) -> Executable:
     raise NotImplementedError("must override")
 
-  def as_text(self, dialect: str | None = None) -> str:
+  def as_text(self, dialect: str | None = None,
+              *,
+              debug_info: bool = False) -> str:
     if dialect is None:
       dialect = "stablehlo"
-    if dialect == "mhlo":
-      return str(self.mhlo())
-    elif dialect == "stablehlo":
-      return str(self.stablehlo())
+    if dialect == "stablehlo":
+      return mlir.module_to_string(self.stablehlo(),
+                                   enable_debug_info=debug_info)
     elif dialect == "hlo":
-      return self.hlo().as_hlo_text()
+      print_opts = xc._xla.HloPrintOptions.short_parsable()
+      print_opts.print_metadata = debug_info
+      return self.hlo().as_hlo_module().to_string(print_opts)
     else:
       raise ValueError(f"unknown dialect: {dialect}")
 
   def compiler_ir(self, dialect: str | None = None) -> Any:
     if dialect is None:
       dialect = "stablehlo"
-    if dialect == "mhlo":
-      return self.mhlo()
-    elif dialect == "stablehlo":
+    if dialect == "stablehlo":
       return self.stablehlo()
     elif dialect == "hlo":
       return self.hlo()
@@ -418,6 +417,7 @@ def make_args_info(in_tree, in_avals, donate_argnums):
       ArgInfo(aval, i in donate_argnums)
       for i, aval in enumerate(flat_avals)])
 
+
 class CompiledCallParams(NamedTuple):
   executable: Executable
   no_kwargs: bool
@@ -425,32 +425,6 @@ class CompiledCallParams(NamedTuple):
   out_tree: tree_util.PyTreeDef
 
 
-class Traced(Stage):
-  __slots__ = ["jaxpr", "args_info", "fun_name", "_out_tree", "_lower_callable",
-               "_args_flat", "_arg_names", "_num_consts"]
-
-  def __init__(self, jaxpr: core.ClosedJaxpr, args_info, fun_name, out_tree,
-               lower_callable, args_flat=None, arg_names=None,
-               num_consts: int = 0):
-    self.jaxpr = jaxpr
-    self.args_info = args_info
-    self.fun_name = fun_name
-    self._out_tree = out_tree
-    self._lower_callable = lower_callable
-    self._args_flat = args_flat
-    self._arg_names = arg_names
-    self._num_consts = num_consts
-
-  @property
-  def out_info(self):
-    return self._out_tree.unflatten(
-        [OutInfo(o.shape, o.dtype) for o in self.jaxpr.out_avals])
-
-  def lower(self):
-    lowering = self._lower_callable()
-    return Lowered(lowering, self.args_info, self._out_tree)
-
-
 class Compiled(Stage):
   """Compiled representation of a function specialized to types/values.
 
@@ -552,6 +526,7 @@ class Compiled(Stage):
     shardings_flat = self._executable.output_shardings()
     return tree_util.tree_unflatten(self.out_tree, shardings_flat)  # pytype: disable=attribute-error
 
+  @property
   def input_layouts(self):
     layouts_flat = self._executable.input_layouts()
     assert all(isinstance(l, Layout) for l in layouts_flat)
@@ -562,6 +537,7 @@ class Compiled(Stage):
                       else Layout() for i in range(self.in_tree.num_leaves)]
     return tree_util.tree_unflatten(self.in_tree, layouts_flat)  # pytype: disable=attribute-error
 
+  @property
   def output_layouts(self):
     layouts_flat = self._executable.output_layouts()
     assert all(isinstance(l, Layout) for l in layouts_flat)
@@ -569,6 +545,7 @@ class Compiled(Stage):
 
   @staticmethod
   def call(*args, **kwargs):
+    util.test_event("stages_compiled_call")
     # This is because `__call__` passes in `self._params` as the first argument.
     # Instead of making the call signature `call(params, *args, **kwargs)`
     # extract it from args because `params` can be passed as a kwarg by users
@@ -584,11 +561,7 @@ class Compiled(Stage):
           f"keyword arguments, but called with keyword arguments: {kws}")
     args_flat, in_tree = tree_util.tree_flatten((args, kwargs))
     if in_tree != params.in_tree:
-      leaf = PytreeLeaf()
-      this_dummy = tree_unflatten(in_tree, [leaf] * in_tree.num_leaves)
-      other_dummy = tree_unflatten(
-          params.in_tree, [leaf] * params.in_tree.num_leaves)
-      errs = list(tree_util.equality_errors(this_dummy, other_dummy))
+      errs = list(tree_util.equality_errors_pytreedef(in_tree, params.in_tree))
       msg = []
       msg.append(
           "Function compiled with input pytree does not match the input pytree"
@@ -597,7 +570,7 @@ class Compiled(Stage):
         fst, *rest = path
         base = ['args', 'kwargs'][fst.idx]
         msg.append(
-            f"    * at {base}{keystr(tuple(rest))}, seen {thing2} but now"
+            f"    * at {base}{tree_util.keystr(tuple(rest))}, seen {thing2} but now"
             f" given {thing1}, so {explanation}")
       raise TypeError('\n'.join(msg))
     try:
@@ -635,10 +608,6 @@ class Compiled(Stage):
     return self._call(*args, **kwargs)
 
 
-class PytreeLeaf:
-  def __repr__(self): return "pytree leaf"
-
-
 class Lowered(Stage):
   """Lowering of a function specialized to argument types and values.
 
@@ -694,7 +663,7 @@ class Lowered(Stage):
     out_avals = self._lowering.compile_args["global_out_avals"]
     out_shardings = self._lowering.compile_args["out_shardings"]
     return self.out_tree.unflatten(
-        [OutInfo(o.shape, o.dtype, None if is_unspecified_or_auto(s) else s)
+        [OutInfo(o.shape, o.dtype, None if isinstance(s, (UnspecifiedValue, AUTO)) else s)
          for o, s in zip(out_avals, out_shardings)])
 
   def compile(
@@ -708,16 +677,21 @@ class Lowered(Stage):
         no_kwargs=self._no_kwargs,
     )
 
-  def as_text(self, dialect: str | None = None) -> str:
+  def as_text(self, dialect: str | None = None, *,
+              debug_info: bool = False) -> str:
     """A human-readable text representation of this lowering.
 
     Intended for visualization and debugging purposes. This need not be a valid
-    nor reliable serialization. It is relayed directly to external callers.
+    nor reliable serialization.
+    Use `jax.export` if you want reliable and portable serialization.
 
     Args:
-      dialect: Optional string specifying a lowering dialect (e.g. "stablehlo")
+      dialect: Optional string specifying a lowering dialect (e.g. "stablehlo",
+        or "hlo").
+      debug_info: Whether to include debugging information,
+        e.g., source location.
     """
-    return self._lowering.as_text(dialect)
+    return self._lowering.as_text(dialect, debug_info=debug_info)
 
   def compiler_ir(self, dialect: str | None = None) -> Any | None:
     """An arbitrary object representation of this lowering.
@@ -725,12 +699,14 @@ class Lowered(Stage):
     Intended for debugging purposes. This is not a valid nor reliable
     serialization. The output has no guarantee of consistency across
     invocations.
+    Use `jax.export` if you want reliable and portable serialization.
 
     Returns ``None`` if unavailable, e.g. based on backend, compiler, or
     runtime.
 
     Args:
-      dialect: Optional string specifying a lowering dialect (e.g. "stablehlo")
+      dialect: Optional string specifying a lowering dialect (e.g. "stablehlo",
+        or "hlo").
     """
     try:
       return self._lowering.compiler_ir(dialect)
@@ -756,8 +732,54 @@ class Lowered(Stage):
       return None
 
 
+class Traced(Stage):
+  __slots__ = ["jaxpr", "args_info", "fun_name", "_out_tree", "_lower_callable",
+               "_args_flat", "_arg_names", "_num_consts"]
+
+  def __init__(self, jaxpr: core.ClosedJaxpr, args_info, fun_name, out_tree,
+               lower_callable, abstract_mesh=None,
+               args_flat=None, arg_names=None, num_consts: int = 0):
+    self.jaxpr = jaxpr
+    self.args_info = args_info
+    self.fun_name = fun_name
+    self._out_tree = out_tree
+    self._lower_callable = lower_callable
+    self._abstract_mesh = abstract_mesh
+    self._args_flat = args_flat
+    self._arg_names = arg_names
+    self._num_consts = num_consts
+
+  @property
+  def out_info(self):
+    return self._out_tree.unflatten(
+        [OutInfo(o.shape, o.dtype) for o in self.jaxpr.out_avals])
+
+  def lower(self, *, lowering_platforms: tuple[str, ...] | None = None,
+            _private_parameters: mlir.LoweringParameters | None = None):
+    from jax._src.interpreters import pxla
+    from jax._src import pjit
+
+    if _private_parameters is None:
+      _private_parameters = mlir.LoweringParameters()
+    new_callable = functools.partial(
+        self._lower_callable, lowering_platforms=lowering_platforms,
+        lowering_parameters=_private_parameters)
+    try:
+      # TODO(yashkatariya): Maybe thread this into pjit params like resource_env
+      # and set the context manager down the stack?
+      with mesh_lib.set_abstract_mesh(self._abstract_mesh):
+        lowering = new_callable()
+    except pxla.DeviceAssignmentMismatchError as e:
+      fails, = e.args
+      msg = pjit._device_assignment_mismatch_error(
+          self.fun_name, fails, self._args_flat, 'jit', self._arg_names)
+      raise ValueError(msg) from None
+    return Lowered(lowering, self.args_info, self._out_tree)
+
+
+@runtime_checkable
 class Wrapped(Protocol):
-  """A function ready to be specialized, lowered, and compiled.
+  """A function ready to be traced, lowered, and compiled.
 
   This protocol reflects the output of functions such as
   ``jax.jit``. Calling it results in JIT (just-in-time) lowering,
@@ -769,6 +791,17 @@ class Wrapped(Protocol):
     """Executes the wrapped function, lowering and compiling as needed."""
     raise NotImplementedError
 
+  def trace(self, *args, **kwargs) -> Traced:
+    """Trace this function explicitly for the given arguments.
+
+    A traced function is staged out of Python and translated to a jaxpr. It is
+    ready for lowering but not yet lowered.
+
+    Returns:
+      A ``Traced`` instance representing the tracing.
+    """
+    raise NotImplementedError
+
   def lower(self, *args, **kwargs) -> Lowered:
     """Lower this function explicitly for the given arguments.
 
