@@ -14,17 +14,15 @@
 """Module for discharging state primitives."""
 from __future__ import annotations
 
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import dataclasses
 from functools import partial
+import math
 import operator
-from typing import Any, Callable, Protocol
+from typing import Any, Protocol, TypeVar
 
-import numpy as np
-
-from jax._src import api_util
 from jax._src import ad_util
-from jax._src import config
+from jax._src import api_util
 from jax._src import core
 from jax._src import linear_util as lu
 from jax._src import source_info_util
@@ -35,12 +33,28 @@ from jax._src.interpreters import partial_eval as pe
 from jax._src.lax import lax
 from jax._src.lax import slicing as lax_slicing
 from jax._src.state import indexing
-from jax._src.state.types import AbstractRef, RefEffect
-from jax._src.state.primitives import get_p, swap_p, addupdate_p
-from jax._src.state.utils import hoist_consts_to_refs
+from jax._src.state.primitives import addupdate_p, get_p, swap_p
+from jax._src.state.types import (
+    AbstractRef,
+    RefBitcaster,
+    RefEffect,
+    RefReshaper,
+    get_ref_aval_from_value,
+    uninitialized,
+)
+from jax._src.state.utils import bitcast, hoist_consts_to_refs
 from jax._src.typing import Array
-from jax._src.util import (safe_map, safe_zip, split_list, weakref_lru_cache,
-                           partition_list, merge_lists, split_dict)
+from jax._src.util import (
+    merge_lists,
+    partition_list,
+    safe_map,
+    safe_zip,
+    split_dict,
+    split_list,
+    unzip2,
+    weakref_lru_cache,
+)
+import numpy as np
 
 ## JAX utilities
 
@@ -91,11 +105,35 @@ class DischargeRule(Protocol):
 
 _discharge_rules: dict[core.Primitive, DischargeRule] = {}
 
+class PartialDischargeRule(Protocol):
+  """A partial discharge rule.
+
+  Exactly like a discharge rule only it accepts a `should_discharge`
+  argument that indicates which inputs should be discharged and the
+  return value returns a tuple of which the first element is the new
+  inputs or none but only the ones that correspond to `True` entries
+  in `should_charge`.
+  """
+
+  def __call__(self, should_discharge: Sequence[bool],
+      in_avals: Sequence[core.AbstractValue],
+      out_avals: Sequence[core.AbstractValue], *args: Any,
+      **params: Any) -> tuple[Sequence[Any | None], Sequence[Any]]:
+    ...
+
+_partial_discharge_rules: dict[core.Primitive, PartialDischargeRule] = {}
+
 def register_discharge_rule(prim: core.Primitive):
   def register(f: DischargeRule):
     _discharge_rules[prim] = f
   return register
 
+def register_partial_discharge_rule(prim: core.Primitive):
+  def register(f: PartialDischargeRule):
+    _partial_discharge_rules[prim] = f
+  return register
+
+
 def _eval_jaxpr_discharge_state(
     jaxpr: core.Jaxpr, should_discharge: Sequence[bool], consts: Sequence[Any],
     *args: Any):
@@ -110,30 +148,51 @@ def _eval_jaxpr_discharge_state(
                        if d and isinstance(v.aval, AbstractRef)}
 
   for eqn in jaxpr.eqns:
-    if eqn.primitive is core.mutable_array_p:
-      [invar], [outvar] = eqn.invars, eqn.outvars
-      ans = env.read(invar)
-      refs_to_discharge.add(id(outvar.aval))
-    elif (any(id(v.aval) in refs_to_discharge for v in eqn.invars)
-         or core.internal_mutable_array_effect in eqn.effects ):
-      if eqn.primitive not in _discharge_rules:
-        raise NotImplementedError("No state discharge rule implemented for "
-            f"primitive: {eqn.primitive}")
-      invals = map(env.read, eqn.invars)
-      in_avals = [v.aval for v in eqn.invars]
-      out_avals = [v.aval for v in eqn.outvars]
-      new_invals, ans = _discharge_rules[eqn.primitive](
-          in_avals, out_avals, *invals, **eqn.params)
-      for new_inval, invar in zip(new_invals, eqn.invars):
-        if new_inval is not None:
-          env.write(invar, new_inval)  # type: ignore[arg-type]
-    else:
-      # Default primitive rule, similar to `core.eval_jaxpr`. Note that here
-      # we assume any higher-order primitives inside of the jaxpr are *not*
-      # stateful.
-      subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
-      ans = eqn.primitive.bind(*subfuns, *map(env.read, eqn.invars),
-                               **bind_params)
+    name_stack = (
+        source_info_util.current_name_stack() + eqn.source_info.name_stack
+    )
+    traceback = eqn.source_info.traceback
+    with source_info_util.user_context(
+        traceback, name_stack=name_stack), eqn.ctx.manager:
+      should_discharge = [id(v.aval) in refs_to_discharge for v in eqn.invars]
+      if eqn.primitive is core.mutable_array_p:
+        [invar], [outvar] = eqn.invars, eqn.outvars
+        ans = env.read(invar)
+        refs_to_discharge.add(id(outvar.aval))
+      elif eqn.primitive is core.freeze_p:
+        [invar], [outvar] = eqn.invars, eqn.outvars
+        ans = env.read(invar)
+        refs_to_discharge.remove(id(invar.aval))
+      elif (any(should_discharge)
+            or core.internal_mutable_array_effect in eqn.effects
+        ):
+        if eqn.primitive in _partial_discharge_rules:
+          rule: DischargeRule = partial(_partial_discharge_rules[eqn.primitive], should_discharge)
+        elif eqn.primitive in _discharge_rules:
+          rule = _discharge_rules[eqn.primitive]
+        else:
+          raise NotImplementedError("No state discharge rule implemented for "
+              f"primitive: {eqn.primitive}")
+        invals = map(env.read, eqn.invars)
+        in_avals = [v.aval for v in eqn.invars]
+        out_avals = [v.aval for v in eqn.outvars]
+        new_invals, ans = rule(
+            in_avals, out_avals, *invals, **eqn.params)
+        for invar, should, new_inval in zip(eqn.invars, should_discharge, new_invals):
+          if new_inval is not None:
+            if not should:
+              raise ValueError(
+                  f"Did not ask for inval to be discharged but it was. ({invar=},"
+                  f" {new_inval=})"
+              )
+            env.write(invar, new_inval)  # type: ignore[arg-type]
+      else:
+        # Default primitive rule, similar to `core.eval_jaxpr`. Note that here
+        # we assume any higher-order primitives inside of the jaxpr are *not*
+        # stateful.
+        subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
+        ans = eqn.primitive.bind(*subfuns, *map(env.read, eqn.invars),
+                                **bind_params)
     if eqn.primitive.multiple_results:
       map(env.write, eqn.outvars, ans)
     else:
@@ -158,6 +217,62 @@ def _is_trivial_indexer(indexer: indexing.NDIndexer):
       return False
   return True
 
+
+def _maybe_convert_to_slice(
+    indexer: indexing.NDIndexer
+) -> list[tuple[int, int, int]] | None:
+  args = []
+
+  for i in indexer.indices:
+    if not isinstance(i, indexing.Slice):
+      return None
+
+    start = i.start
+    end = i.start + (i.size - 1) * i.stride + 1
+    stride = i.stride
+
+    # cannot convert to static `slice` if `start` or `end` is dynamic
+    if not isinstance(start, int) or not isinstance(end, int):
+      return None
+
+    args.append((start, end, stride))
+
+  return args
+
+
+def _maybe_convert_to_dynamic_slice(
+    indexer: indexing.NDIndexer,
+) -> (
+    tuple[tuple[Array | int, ...], tuple[Array | int, ...], tuple[int, ...]]
+    | None
+):
+  # An NDIndexer only corresponds to a `dynamic_slice` or `dynamic_update_slice`
+  # if each of the indexers is a `Slice` or a ()-shaped value.
+  if not all(isinstance(i, indexing.Slice) or not np.shape(i)
+             for i in indexer.indices):
+    return None
+
+  # `lax.dynamic_slice` does not handle striding
+  for i in indexer.indices:
+    if isinstance(i, indexing.Slice) and i.stride > 1:
+      return None
+
+  _convert_i32 = lambda x: lax.convert_element_type(x, np.dtype("int32"))
+  starts = tuple(
+      _convert_i32(i.start) if isinstance(i, indexing.Slice)
+      else _convert_i32(i) for i in indexer.indices
+  )
+  sizes = tuple(
+      i.size if isinstance(i, indexing.Slice) else 1 for i in indexer.indices
+  )
+  squeeze_dims = tuple(
+      i
+      for i, idx in enumerate(indexer.indices)
+      if not isinstance(idx, indexing.Slice)
+  )
+  return starts, sizes, squeeze_dims
+
+
 def _convert_to_array_indexer(indexer: indexing.NDIndexer
                               ) -> tuple[int | Array, ...]:
   # This is the general gather case. We need to create the gather arrays.
@@ -177,7 +292,7 @@ def _convert_to_array_indexer(indexer: indexing.NDIndexer
       assert isinstance(idx, indexing.Slice)
       slice_indices = lax.broadcasted_iota(
           np.dtype("int32"), total_shape, next(slice_dim_iter)
-      ) + idx.start
+      ) * idx.stride + idx.start
       slice_indexer.append(slice_indices)
       integer_indexer = tuple(
           lax.expand_dims(idx, (-1,)) for idx in integer_indexer
@@ -187,37 +302,6 @@ def _convert_to_array_indexer(indexer: indexing.NDIndexer
   return tuple(merge_lists(is_integer_indexer, slice_indexer, integer_indexer))
 
 
-def _maybe_convert_to_dynamic_slice(
-    indexer: indexing.NDIndexer,
-) -> (
-    tuple[tuple[Array | int, ...], tuple[Array | int, ...], tuple[int, ...]]
-    | None
-):
-  # An NDIndexer only corresponds to a `dynamic_slice` or `dynamic_update_slice`
-  # if each of the indexers is a `Slice` or a ()-shaped value.
-  if not all(isinstance(i, indexing.Slice) or not np.shape(i)
-             for i in indexer.indices):
-    return None
-  # TODO(b/329733289): support strided load/store in interpret mode.
-  for i in indexer.indices:
-    if isinstance(i, indexing.Slice) and i.stride > 1:
-      raise NotImplementedError("Unimplemented stride support.")
-  _convert_i32 = lambda x: lax.convert_element_type(x, np.dtype("int32"))
-  starts = tuple(
-      _convert_i32(i.start) if isinstance(i, indexing.Slice)
-      else _convert_i32(i) for i in indexer.indices
-  )
-  sizes = tuple(
-      i.size if isinstance(i, indexing.Slice) else 1 for i in indexer.indices
-  )
-  squeeze_dims = tuple(
-      i
-      for i, idx in enumerate(indexer.indices)
-      if not isinstance(idx, indexing.Slice)
-  )
-  return starts, sizes, squeeze_dims
-
-
 @register_discharge_rule(get_p)
 def _get_discharge_rule(
     in_avals: Sequence[core.AbstractValue],
@@ -240,31 +324,98 @@ def _prepend_scatter(x, indexer, val, *, add=False):
   return x[None].at[(0, *indexer)].set(val)[0]
 
 
-def _get_discharge(x, idx, tree):
-  indexers = tree_util.tree_unflatten(tree, idx)
+def _index_array(x, indexer):
+  if _is_trivial_indexer(indexer):
+    return x
+  # Try the three APIs in the following order: `lax.slice`,
+  # `lax.dynamic_slice` and gather
+  if maybe_slice := _maybe_convert_to_slice(indexer):
+    x = lax_slicing.slice(x, *zip(*maybe_slice))
+  # If everything in the indexer is a slice or ()-shaped, we can also
+  # use `lax.dynamic_slice` with 1-sized slices for ()-shaped indices.
+  # We need to squeeze out the 1-sized slices at the end.
+  elif maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
+    starts, sizes, squeeze_dims = maybe_slice
+    y = lax_slicing.dynamic_slice(x, starts, sizes)
+    x = lax.squeeze(y, squeeze_dims)
+  else:
+    indexer = _convert_to_array_indexer(indexer)
+    x = x[None][(np.array(0, "int32"), *indexer)]
+  return x
+
+
+def transform_array(x, transforms):
+  if transforms is None:
+    transforms = []
   result = x
-  for indexer in indexers:
-    if _is_trivial_indexer(indexer):
+  for transform in transforms:
+    if transform is None:
       continue
-    if indexer is None:
-      continue
-    # If everything in the indexer is a slice or ()-shaped, we can also
-    # use `lax.dynamic_slice` with 1-sized slices for ()-shaped indices.
-    # We need to squeeze out the 1-sized slices at the end.
-    if maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
-      starts, sizes, squeeze_dims = maybe_slice
-      y = lax_slicing.dynamic_slice(result, starts, sizes)
-      result = lax.squeeze(y, squeeze_dims)
-    else:
-      indexer = _convert_to_array_indexer(indexer)
-      result = result[None][(np.array(0, "int32"), *indexer)]
+    match transform:
+      case indexing.NDIndexer():
+        result = _index_array(result, transform)
+      case RefBitcaster():
+        result = bitcast(result, transform.dtype)
+      case RefReshaper():
+        result = result.reshape(transform.shape)
+      case _:
+        raise NotImplementedError(f"Unsupported transform: {transform}")
   return result
 
-def _indexer(idx, indexed_dims):
-  idx_ = iter(idx)
-  indexer = tuple(next(idx_) if b else slice(None) for b in indexed_dims)
-  assert next(idx_, None) is None
-  return indexer
+def transform_swap_array(x, transforms, val):
+  if transforms is None:
+    transforms = []
+  result = x
+  result_val = val
+  # Compute updated "val" (result).
+  _results = [x]
+  for transform in transforms:
+    match transform:
+      case indexing.NDIndexer():
+        indexer = transform
+        if _is_trivial_indexer(indexer):
+          _results.append(_results[-1])
+          continue
+        # If everything in the indexer is a slice or ()-shaped, we can also
+        # use `lax.dynamic_slice` with 1-sized slices for ()-shaped indices.
+        # We need to squeeze out the 1-sized slices at the end.
+        if maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
+          starts, sizes, squeeze_dims = maybe_slice
+          result_old = lax_slicing.dynamic_slice(result, starts, sizes)
+          result = lax.squeeze(result_old, squeeze_dims)
+        else:
+          indexer = _convert_to_array_indexer(indexer)
+          result = _prepend_gather(result, indexer)
+        _results.append(result)
+      case RefBitcaster():
+        _results.append(bitcast(result, transform.dtype))
+      case RefReshaper():
+        _results.append(result.reshape(transform.shape))
+      case _:
+        raise NotImplementedError(f"Unsupported transform: {transform}")
+
+  # Compute updated "x" (result_val)
+  for i, transform in reversed(list(enumerate(transforms))):
+    if isinstance(transform, indexing.NDIndexer):
+      indexer = transform
+      if _is_trivial_indexer(indexer):
+        continue
+      if maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
+        starts, _, squeeze_dims = maybe_slice
+        result_val = lax.expand_dims(result_val, squeeze_dims)
+        result_val = lax_slicing.dynamic_update_slice(
+            _results[i], result_val, starts
+        )
+      else:
+        indexer = _convert_to_array_indexer(indexer)
+        result_val = _prepend_scatter(_results[i], indexer, result_val)
+    else:
+      raise NotImplementedError(f"Unsupported transform: {transform}")
+  return result, result_val
+
+def _get_discharge(x, idx, tree):
+  transforms = tree_util.tree_unflatten(tree, idx)
+  return transform_array(x, transforms)
 
 @register_discharge_rule(swap_p)
 def _swap_discharge_rule(
@@ -276,29 +427,8 @@ def _swap_discharge_rule(
   return (x_new, None) + (None,) * len(idx), z
 
 def _swap_discharge(x, val, idx, tree):
-  indexers = tree_util.tree_unflatten(tree, idx)
-
-  result = x
-  result_val = val
-  for indexer in indexers:
-    if _is_trivial_indexer(indexer):
-      continue
-    # If everything in the indexer is a slice or ()-shaped, we can also
-    # use `lax.dynamic_slice` with 1-sized slices for ()-shaped indices.
-    # We need to squeeze out the 1-sized slices at the end.
-    if maybe_slice := _maybe_convert_to_dynamic_slice(indexer):
-      starts, sizes, squeeze_dims = maybe_slice
-      result_old = lax_slicing.dynamic_slice(result, starts, sizes)
-      result_val = lax.expand_dims(result_val, squeeze_dims)
-      y = lax_slicing.dynamic_update_slice(result, result_val, starts)
-      result = lax.squeeze(result_old, squeeze_dims)
-      result_val = y
-    else:
-      indexer = _convert_to_array_indexer(indexer)
-      result_old = _prepend_gather(result, indexer)
-      result_val = _prepend_scatter(result, indexer, result_val)
-      result = result_old
-  return result, result_val
+  transforms = tree_util.tree_unflatten(tree, idx)
+  return transform_swap_array(x, transforms, val)
 
 @register_discharge_rule(addupdate_p)
 def _addupdate_discharge_rule(
@@ -310,10 +440,10 @@ def _addupdate_discharge_rule(
   return (ans, None) + (None,) * len(idx), []
 
 def _addupdate_discharge(x, val, idx, tree):
-  indexers = tree_util.tree_unflatten(tree, idx)
-  if len(indexers) > 1:
+  transforms = tree_util.tree_unflatten(tree, idx)
+  if len(transforms) > 1:
     raise NotImplementedError("Only single indexer is supported.")
-  indexer = indexers[0]
+  indexer = transforms[0]
   if _is_trivial_indexer(indexer):
     return x + val
   # If everything in the indexer is a slice or ()-shaped, we can also
@@ -357,40 +487,73 @@ def _closed_call_discharge_rule(
 run_state_p = core.Primitive("run_state")
 run_state_p.multiple_results = True
 
-def _run_state_bind(*args: Any, jaxpr: core.Jaxpr,
-                    which_linear: tuple[bool, ...]):
-  if config.enable_checks.value:
-    core.check_jaxpr(jaxpr)
-    assert len(jaxpr.invars) == len(args)
-    assert len(which_linear) == len(args)
-  return core.Primitive.bind(run_state_p, *args, jaxpr=jaxpr,
-                             which_linear=which_linear)
-run_state_p.def_custom_bind(_run_state_bind)
+def _default_initialization(x):
+  assert hasattr(x, 'shape')
+  assert hasattr(x, 'dtype')
+  dtype = np.dtype(x)
+  if np.issubdtype(dtype, np.integer):
+    value = np.iinfo(dtype).min
+  else:
+    value = math.nan
+  return lax.full(x.shape, value, dtype)
 
 def _run_state_impl(*args: Any, jaxpr: core.Jaxpr,
-                    which_linear: tuple[bool, ...]):
+                    which_linear: tuple[bool, ...],
+                    is_initialized: tuple[bool, ...]):
   del which_linear
   discharged_jaxpr, consts = discharge_state(jaxpr, ())
+  # Initialize the args that are not initialized.
+  args_it = iter(args)
+  args = tuple(
+      next(args_it) if is_init else _default_initialization(var.aval)
+      for is_init, var in zip(is_initialized, discharged_jaxpr.invars)
+  )
   return core.eval_jaxpr(discharged_jaxpr, consts, *args)
 run_state_p.def_impl(_run_state_impl)
 mlir.register_lowering(run_state_p, mlir.lower_fun(_run_state_impl))
 
 def _run_state_abstract_eval(*avals: core.AbstractValue, jaxpr: core.Jaxpr,
-                             which_linear: tuple[bool, ...]):
+                             which_linear: tuple[bool, ...],
+                             is_initialized: tuple[bool, ...]):
   del which_linear
+  assert sum(is_initialized) == len(avals)
   # When we abstractly evaluate `run_state`, we want to keep track of which
   # input avals are `Ref`s and which are not. If an aval is a `Ref`, we want to
   # "propagate" out its inner effects. Otherwise, the effects are local to this
   # `run_state`.
+  inner_to_outer_aval_mapping = {}
+  outer_ref_index = 0
+  for i, is_init in enumerate(is_initialized):
+    if not is_init:
+      pass
+    inner_to_outer_aval_mapping[i] = outer_ref_index
+    outer_ref_index += 1
+  nonlocal_effects = set()
   is_ref = {i for i, aval in enumerate(avals) if isinstance(aval, AbstractRef)}
-  nonlocal_effects = {e for e in jaxpr.effects
-                      if (isinstance(e, RefEffect) and e.input_index in is_ref)
-                      or not isinstance(e, RefEffect)}
+  for eff in jaxpr.effects:
+    if not isinstance(eff, RefEffect):
+      nonlocal_effects.add(eff)
+      continue
+    if eff.input_index not in inner_to_outer_aval_mapping:
+      # This means that this effect corresponds to an uninitialized Ref and
+      # should not propagate out of the primitive.
+      continue
+    # If we do propagate the effect, we need to update the input index to
+    # correspond to the outer index.
+    outer_index = inner_to_outer_aval_mapping[eff.input_index]
+    if outer_index in is_ref:
+      # This means that the effect corresponds to a Ref from an outside scope.
+      nonlocal_effects.add(
+          eff.replace(input_index=inner_to_outer_aval_mapping[eff.input_index])
+      )
   return avals, nonlocal_effects
 run_state_p.def_effectful_abstract_eval(_run_state_abstract_eval)
 
 def _run_state_jvp(primals: Sequence[Any], tangents: Sequence[Any], *,
-                   jaxpr: core.Jaxpr, which_linear: tuple[bool, ...]):
+                   jaxpr: core.Jaxpr, which_linear: tuple[bool, ...],
+                   is_initialized: tuple[bool, ...]):
+  if not all(is_initialized):
+    raise NotImplementedError("Uninitialized Refs are not supported in jvp.")
   nonzero_tangents = [not isinstance(t, ad_util.Zero) for t in tangents]
   discharged_jaxpr, body_consts = discharge_state(jaxpr, ())
   for _ in range(len(nonzero_tangents)):
@@ -412,12 +575,14 @@ def _run_state_jvp(primals: Sequence[Any], tangents: Sequence[Any], *,
   jvp_jaxpr = hoist_consts_to_refs(jvp_jaxpr_)
   jvp_which_linear = (*(False,) * len(jvp_consts), *which_linear, *(True,) * len(tangents))
   out = run_state_p.bind(*jvp_consts, *primals, *tangents, jaxpr=jvp_jaxpr,
-                         which_linear=jvp_which_linear)
+                         which_linear=jvp_which_linear,
+                         # TODO(sharadmv): compute this in the general case
+                         is_initialized=(True,) * len(jvp_jaxpr.invars))
   out_consts, out_primals, out_tangents = split_list(out, [len(jvp_consts),
                                                            len(primals)])
   del out_consts
   out_tangents_iter = iter(out_tangents)
-  out_tangents = [next(out_tangents_iter) if nz else ad_util.Zero.from_value(p)
+  out_tangents = [next(out_tangents_iter) if nz else ad_util.Zero.from_primal_value(p)
                   for p, nz in zip(out_primals, nonzero_tangents)]
   return out_primals, out_tangents
 ad.primitive_jvps[run_state_p] = _run_state_jvp
@@ -443,8 +608,7 @@ def _convert_outputs_to_writes(
   jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(
       eval_jaxpr, [*in_avals, *res_ref_avals])
   assert not consts
-  return jaxpr, [core.ShapedArray(a.inner_aval.shape, a.inner_aval.dtype)  # pytype: disable=attribute-error
-                 for a in res_ref_avals]
+  return jaxpr, [core.ShapedArray(a.shape, a.dtype) for a in res_ref_avals]
 
 def _convert_inputs_to_reads(num_res: int, jaxpr: core.Jaxpr) -> core.Jaxpr:
   assert not jaxpr.constvars, "Jaxpr should not have constvars"
@@ -465,7 +629,12 @@ def _convert_inputs_to_reads(num_res: int, jaxpr: core.Jaxpr) -> core.Jaxpr:
   return jaxpr
 
 def _run_state_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
-                            jaxpr: core.Jaxpr, which_linear: tuple[bool, ...]):
+                            jaxpr: core.Jaxpr, which_linear: tuple[bool, ...],
+                            is_initialized: tuple[bool, ...]):
+  if not all(is_initialized):
+    raise NotImplementedError(
+        "Uninitialized Refs are not supported in partial_eval."
+    )
   num_inputs = len(tracers)
   assert num_inputs == len(jaxpr.invars)
   in_unknowns = [not t.pval.is_known() for t in tracers]
@@ -525,7 +694,9 @@ def _run_state_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
 
   jaxpr_known_which_linear = (*known_which_linear, *(False,) * num_res)
   out_flat = run_state_p.bind(*jaxpr_known_args, jaxpr=jaxpr_known,
-                              which_linear=jaxpr_known_which_linear)
+                              which_linear=jaxpr_known_which_linear,
+                              # TODO(sharadmv): compute this in the general case
+                              is_initialized=(True,) * len(jaxpr_known.invars))
   known_outputs, residuals = split_list(out_flat, [len(known_tracers)])
   residuals = map(trace.new_instantiated_const, residuals)
   ref_res, nonref_res = split_list(residuals, [num_res_ref])
@@ -553,7 +724,9 @@ def _run_state_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
 
   assert len(unknown_inputs) == len(res_ref_unknown_outputs)
   assert len(unknown_inputs) == len(jaxpr_unknown.invars)
-  uk_params = dict(jaxpr=jaxpr_unknown, which_linear=unknown_which_linear)
+  uk_params = dict(jaxpr=jaxpr_unknown, which_linear=unknown_which_linear,
+                   # TODO(sharadmv); compute this in the general case
+                   is_initialized=(True,) * len(jaxpr_unknown.invars))
   _, eqn_effects = run_state_p.abstract_eval(*[v.aval for v in unknown_inputs],
                                              **uk_params)
   eqn = pe.new_eqn_recipe(unknown_inputs, res_ref_unknown_outputs,
@@ -571,7 +744,13 @@ def _run_state_partial_eval_custom(
     eqn: core.JaxprEqn):
   if not any(in_unknowns):
     return eqn, None, in_unknowns, [False] * len(in_unknowns), []
-  jaxpr, which_linear = split_dict(eqn.params, ["jaxpr", "which_linear"])
+  jaxpr, which_linear, is_initialized = split_dict(
+      eqn.params, ["jaxpr", "which_linear", "is_initialized"]
+  )
+  if not all(is_initialized):
+    raise NotImplementedError(
+        "Uninitialized Refs are not supported in partial_eval_custom."
+    )
   num_inputs = len(eqn.invars)
   # We first need to run a fixpoint to determine which of the `Ref`s are unknown
   # after running the for loop. However, the jaxpr has no outputs. Instead, we
@@ -598,7 +777,8 @@ def _run_state_partial_eval_custom(
       break
     in_unknowns = map(operator.or_, in_unknowns, out_unknowns)
   else:
-    if num_inputs > 0: raise Exception("Invalid fixpoint")
+    if num_inputs > 0:
+      raise Exception("Invalid fixpoint")
   del out_unknowns # Redundant since it's the same as `in_unknowns`
   new_inst = [x for x, already, inst in zip(eqn.invars, in_inst, out_inst)
               if type(x) is core.Var and inst and not already]
@@ -637,7 +817,9 @@ def _run_state_partial_eval_custom(
   jaxpr_known_which_linear = (*known_which_linear, *(False,) * num_res)
   known_and_res_invars = [*known_invars, *ref_resvars, *nonref_resvars]
 
-  known_params = dict(jaxpr=jaxpr_known, which_linear=jaxpr_known_which_linear)
+  known_params = dict(jaxpr=jaxpr_known, which_linear=jaxpr_known_which_linear,
+                      # TODO(sharadmv): compute this in the general case
+                      is_initialized=(True,) * len(jaxpr_known.invars))
   _, known_effects = run_state_p.abstract_eval(
       *[v.aval for v in known_and_res_invars], **known_params)
   eqn_known = pe.new_jaxpr_eqn(known_and_res_invars,
@@ -649,7 +831,9 @@ def _run_state_partial_eval_custom(
 
   _, staged_which_linear = partition_list(in_unknowns, which_linear)
   which_linear_unknown = (*[False] * num_res, *staged_which_linear)
-  staged_params = dict(jaxpr=jaxpr_staged, which_linear=which_linear_unknown)
+  staged_params = dict(jaxpr=jaxpr_staged, which_linear=which_linear_unknown,
+                       # TODO(sharadmv): compute this in the general case
+                       is_initialized=(True,) * len(jaxpr_staged.invars))
   rejiggered_resvars = [*nonref_resvars, *ref_resvars]
   _, staged_invars = partition_list(in_unknowns, eqn.invars)
   res_staged_invars = [*rejiggered_resvars, *staged_invars]
@@ -662,7 +846,7 @@ def _run_state_partial_eval_custom(
       out = run_state_p.bind(*args, **staged_params)
       return out[num_res:]
     staged_call_jaxpr, _, (), () = pe.trace_to_jaxpr_dynamic(staged,
-        [v.aval for v in res_staged_invars])
+                                                             [v.aval for v in res_staged_invars])
     eqn_staged = pe.new_jaxpr_eqn(res_staged_invars,
                                   staged_outvars,
                                   core.closed_call_p,
@@ -680,8 +864,12 @@ def _run_state_partial_eval_custom(
   return eqn_known, eqn_staged, in_unknowns, in_unknowns, new_vars
 pe.partial_eval_jaxpr_custom_rules[run_state_p] = _run_state_partial_eval_custom
 
-def _transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: Sequence[bool]
-                    ) -> tuple[core.Jaxpr, Any]:
+def _transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: Sequence[bool],
+                     is_initialized: tuple[bool, ...]) -> tuple[core.Jaxpr, Any]:
+  if not all(is_initialized):
+    raise NotImplementedError(
+        "Uninitialized Refs are not supported in transpose."
+    )
   def trans(*args):
     # First we want to run the computation to read all the residual refs. We can
     # do that by using partial evaluation with all linear inputs unknown.
@@ -700,8 +888,14 @@ def _transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: Sequence[bool]
     all_avals = [*res_avals, *[v.aval for v in res_jaxpr_.outvars]]
     empty_res = map(ad.zeros_like_aval, all_avals)
     res_jaxpr, _ = _convert_outputs_to_writes(res_jaxpr_)
-    res = run_state_p.bind(*res_args, *empty_res, jaxpr=res_jaxpr,
-                           which_linear=(False,) * (len(res_args) + len(empty_res)))
+    res = run_state_p.bind(
+        *res_args,
+        *empty_res,
+        jaxpr=res_jaxpr,
+        which_linear=(False,) * (len(res_args) + len(empty_res)),
+        # TODO(sharadmv): compute this in the general case
+        is_initialized=(True,) * len(res_jaxpr.invars),
+    )
     res = res[len(res_args):]
     ref_res_, nonref_res_ = split_list(res, [num_res_ref])
 
@@ -724,7 +918,12 @@ def _transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: Sequence[bool]
   return jaxpr_trans, consts
 
 def _run_state_transpose(in_cts, *args, jaxpr: core.Jaxpr,
-                         which_linear: tuple[bool, ...]):
+                         which_linear: tuple[bool, ...],
+                         is_initialized: tuple[bool, ...]):
+  if not all(is_initialized):
+    raise NotImplementedError(
+        "Uninitialized Refs are not supported in transpose."
+    )
   # if any in_ct is nonzero, we definitely want it in args_ (and the
   # corresponding x in args could be an undefined primal, but doesn't have to be)
   # for non-res stuff:
@@ -748,12 +947,19 @@ def _run_state_transpose(in_cts, *args, jaxpr: core.Jaxpr,
     elif type(ct) is not ad_util.Zero and     ad.is_undefined_primal(x):
       # the loop was 'getting and setting', grab that cotangent!
       transpose_args.append(ct)
-  jaxpr_transpose_, consts = _transpose_jaxpr(jaxpr, which_linear)
+  jaxpr_transpose_, consts = _transpose_jaxpr(
+      jaxpr, which_linear, is_initialized
+  )
   jaxpr_transpose = hoist_consts_to_refs(jaxpr_transpose_)
   which_linear = (*[False] * len(consts), *which_linear)
-  const_all_outs = run_state_p.bind(*consts, *transpose_args,
-                                    jaxpr=jaxpr_transpose,
-                                    which_linear=which_linear)
+  const_all_outs = run_state_p.bind(
+      *consts,
+      *transpose_args,
+      jaxpr=jaxpr_transpose,
+      which_linear=which_linear,
+      # TODO(sharadmv): compute this in the general case
+      is_initialized=(True,) * len(jaxpr_transpose.invars),
+  )
   _, all_outs = split_list(const_all_outs, [len(consts)])
   ct_outs = [ct if ad.is_undefined_primal(x) else None
              for x, ct in zip(args, all_outs)]
@@ -764,9 +970,15 @@ ad.primitive_transposes[run_state_p] = _run_state_transpose
 def _run_state_discharge_rule(in_avals: Sequence[core.AbstractValue],
                               out_avals: Sequence[core.AbstractValue],
                               *args: Any, jaxpr: core.Jaxpr,
-                              which_linear: Sequence[bool]):
+                              which_linear: Sequence[bool],
+                              is_initialized: tuple[bool, ...]):
+  if not all(is_initialized):
+    raise NotImplementedError(
+        "Uninitialized Refs are not supported in discharge."
+    )
   del out_avals
-  out_vals = run_state_p.bind(*args, jaxpr=jaxpr, which_linear=which_linear)
+  out_vals = run_state_p.bind(*args, jaxpr=jaxpr, which_linear=which_linear,
+                              is_initialized=is_initialized)
   new_invals = []
   for aval, out_val in zip(in_avals, out_vals):
     new_invals.append(out_val if isinstance(aval, AbstractRef) else None)
@@ -781,19 +993,27 @@ def initial_style_jaxpr(
 def _initial_style_jaxpr(fun, in_tree, in_avals):
   fun_, out_tree_thunk = api_util.flatten_fun_nokwargs(lu.wrap_init(fun),
       tree_util.treedef_tuple((in_tree,)))
-  debug = pe.debug_info(fun_, in_tree, out_tree_thunk, False, 'run_state')
+  debug = pe.tracing_debug_info(fun_, in_tree, out_tree_thunk, False, 'run_state')
   jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(fun_, in_avals, debug)
   return jaxpr, consts, out_tree_thunk()
 
-def run_state(f: Callable[..., None]):
+
+T = TypeVar('T')
+def run_state(f: Callable[..., None]) -> Callable[[T], T]:
   def wrapped(args):
     flat_args, in_tree = tree_util.tree_flatten(args)
-    avals = [core.raise_to_shaped(core.get_aval(arg)) for arg in flat_args]
-    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, map(AbstractRef, avals))
+    ref_avals, ref_args = unzip2(map(get_ref_aval_from_value, flat_args))
+    # There may be some uninitialized values here in ref_args.
+    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, ref_avals)
     jaxpr = hoist_consts_to_refs(jaxpr_)
-    which_linear = (False,) * (len(consts) + len(flat_args))
-    out_const_flat = run_state_p.bind(*consts, *flat_args, jaxpr=jaxpr,
-                                      which_linear=which_linear)
+    which_linear = (False,) * (len(consts) + len(ref_args))
+    refs_is_initialized = tuple(r is not uninitialized for r in ref_args)
+    init_args = tuple(r for r in ref_args if r is not uninitialized)
+    # Consts are always initialized.
+    is_initialized = (True,) * len(consts) + refs_is_initialized
+    out_const_flat = run_state_p.bind(*consts, *init_args, jaxpr=jaxpr,
+                                      which_linear=which_linear,
+                                      is_initialized=is_initialized)
     _, out_flat = split_list(out_const_flat, [len(consts)])
     return in_tree.unflatten(out_flat)
   return wrapped
@@ -801,12 +1021,19 @@ def run_state(f: Callable[..., None]):
 def run_state_reference(f: Callable[..., None]):
   def wrapped(args):
     flat_args, in_tree = tree_util.tree_flatten(args)
-    avals = [core.raise_to_shaped(core.get_aval(arg)) for arg in flat_args]
-    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, map(AbstractRef, avals))
+    ref_avals, ref_args = unzip2(map(get_ref_aval_from_value, flat_args))
+    jaxpr_, consts, _ = initial_style_jaxpr(f, in_tree, ref_avals)
     jaxpr = hoist_consts_to_refs(jaxpr_)
     discharged_jaxpr, discharged_consts = discharge_state(jaxpr, ())
+
+    # Initialize any uninitialized values here in ref_args in the reference.
+    ref_args = [
+        _default_initialization(aval) if r is uninitialized else r
+        for r, aval in zip(ref_args, ref_avals)
+    ]
+
     out_const_flat = core.eval_jaxpr(discharged_jaxpr, discharged_consts,
-                                     *consts, *args)
+                                     *consts, *ref_args)
     _, out_flat = split_list(out_const_flat, [len(consts)])
     return in_tree.unflatten(out_flat)
   return wrapped
