@@ -19,27 +19,26 @@ from __future__ import annotations
 
 import base64
 import collections.abc
+from collections.abc import Callable, Sequence
 import dataclasses
+import enum
 import functools
 import io
 import os
 import time
-from typing import Any, Callable
+from typing import Any
 
 import jax
-from jax import core
+from jax._src import core
 from jax._src import config
 from jax._src import sharding_impls
 from jax._src.interpreters import mlir
 from jax._src.lib import tpu
 from jax._src.lib import xla_client
-from jax._src.lib.mlir.dialects import hlo
 from jax.interpreters import xla
 from jaxlib.mlir import ir
 from jaxlib.mlir.dialects import mhlo
-from jaxlib.mlir.dialects import stablehlo
 from jaxlib.mlir.passmanager import PassManager
-import numpy as np
 
 try:
   from absl import flags
@@ -47,7 +46,7 @@ try:
 except ImportError:
   FLAGS = {}
 
-_MOSAIC_USE_PYTHON_PIPELINE = config.define_bool_state(
+_MOSAIC_USE_PYTHON_PIPELINE = config.bool_state(
     name="mosaic_use_python_pipeline",
     default=False,
     help=(
@@ -57,18 +56,49 @@ _MOSAIC_USE_PYTHON_PIPELINE = config.define_bool_state(
     ),
 )
 
-_MOSAIC_ALLOW_HLO = config.define_bool_state(
+_MOSAIC_ALLOW_HLO = config.bool_state(
     name="jax_mosaic_allow_hlo",
     default=False,
     help="Allow hlo dialects in Mosaic",
 )
 
+
+# This tracks the latest Mosaic IR version with a monthly delay.
+FWD_COMPAT_IR_VERSION = 3
+
+
 tpu_custom_call_p = core.Primitive("tpu_custom_call")
 tpu_custom_call_p.def_impl(
     functools.partial(xla.apply_primitive, tpu_custom_call_p))
 tpu_custom_call_p.multiple_results = True
 
 
+def get_target_shape(hardware_generation: int) -> tuple[int, int]:
+  """Returns the target shape for the given hardware generation."""
+  del hardware_generation
+  return (8, 128)
+
+
+class MemorySpace(enum.Enum):
+  HBM = enum.auto()
+  VMEM = enum.auto()
+  SEMAPHORE_MEM = enum.auto()
+  SMEM = enum.auto()
+
+  @property
+  def color(self) -> int:
+    if self == MemorySpace.HBM:
+      return 0
+    elif self == MemorySpace.VMEM:
+      return 1
+    elif self == MemorySpace.SEMAPHORE_MEM:
+      return 2
+    elif self == MemorySpace.SMEM:
+      return 4
+    else:
+      raise ValueError("invalid memory space: " + str(self))
+
+
 @dataclasses.dataclass(frozen=True)
 class CostEstimate:
   flops: int
@@ -96,6 +126,8 @@ class CustomCallBackendConfig:
   flags: dict[str, bool | int | float] | None
   allow_input_fusion: list[bool] | None
   serialization_format: int | None
+  internal_scratch_in_bytes: int | None
+  output_memory_spaces: tuple[MemorySpace | None, ...] | None
 
   # We omit the body while printing, because primitive params get embedded
   # in HLO metadata, and the body blows up its size.
@@ -135,6 +167,17 @@ class CustomCallBackendConfig:
         if i + 1 != len(self.allow_input_fusion):
           config.write(b",")
       config.write(b"]")
+    if self.internal_scratch_in_bytes is not None:
+      config.write(b', "internal_scratch_in_bytes": ')
+      config.write(str(self.internal_scratch_in_bytes).encode("ascii"))
+    if self.output_memory_spaces is not None:
+      config.write(b', "output_memory_colors": [')
+      for i, memory_space in enumerate(self.output_memory_spaces):
+        if i:
+          config.write(b",")
+        color = memory_space.color if memory_space is not None else -1
+        config.write(str(color).encode("ascii"))
+      config.write(b"]")
     config.write(b"}")  # End of custom_call_config.
     if self.device_type is not None:
       config.write(b', "device_type": ')
@@ -169,9 +212,6 @@ class CustomCallBackendConfig:
         if i + 1 != len(self.flags):
           config.write(b",")
       config.write(b"]")
-    # Prevent the compiler from sharding the custom call beyond what Mosaic does
-    # based on user annotations
-    config.write(b', "implicit_sharding": {"type": "MANUAL"}')
     config.write(b"}")
     return config.getvalue()
 
@@ -181,13 +221,8 @@ def _tpu_custom_call_abstract_eval(*_, out_avals, **__):
   return out_avals
 
 
-def _aval_to_layout(aval):
-  arange = np.arange(aval.ndim, dtype=np.dtype(np.int64))[::-1].copy()
-  return ir.DenseIntElementsAttr.get(arange, type=ir.IndexType.get())
-
-
-def _avals_to_layouts(avals):
-  return ir.ArrayAttr.get([_aval_to_layout(a) for a in avals])
+def _avals_to_layouts(avals) -> Sequence[Sequence[int]]:
+  return [tuple(range(a.ndim - 1, -1, -1)) for a in avals]
 
 
 def _tpu_custom_call_lowering(
@@ -195,74 +230,52 @@ def _tpu_custom_call_lowering(
     *in_nodes,  # pylint: disable=missing-function-docstring
     config: CustomCallBackendConfig,
     kernel_name: str | None,
-    kernel_regeneration_metadata: bytes | None,
     out_avals: Any,
     input_output_aliases: tuple[tuple[int, int], ...],
 ) -> ...:
-  i32_type = ir.IntegerType.get_signless(32)
-  multiple_results = len(out_avals) > 1
-  if multiple_results:
-    result_type = ir.TupleType.get_tuple(
-        [mlir.aval_to_ir_type(aval) for aval in out_avals]
-    )
-  else:
-    result_type = mlir.aval_to_ir_type(out_avals[0])
+  result_types = [mlir.aval_to_ir_type(aval) for aval in out_avals]
   axis_context = ctx.module_context.axis_context
   if isinstance(axis_context, sharding_impls.SPMDAxisContext):
     if axis_context.manual_axes != frozenset(axis_context.mesh.axis_names):
       raise NotImplementedError(
           "Mosaic kernels cannot be automatically partitioned. Please wrap the"
-          " call in a shard_map or xmap."
+          " call in a shard_map."
       )
   elif isinstance(axis_context, sharding_impls.ShardingContext):
     if axis_context.num_devices != 1:
       raise NotImplementedError(
           "Mosaic kernels cannot be automatically partitioned. Please wrap the"
-          " call in a shard_map or xmap."
+          " call in a shard_map."
       )
   elif config.has_communication:
     raise NotImplementedError(
         "Replica lowering for Mosaic kernels not implemented."
     )
-  call = stablehlo.CustomCallOp(
-      [result_type],
-      in_nodes,
-      call_target_name=ir.StringAttr.get(b"tpu_custom_call"),
-      has_side_effect=ir.BoolAttr.get(False),
-      backend_config=ir.StringAttr.get(config.to_json()),
-      api_version=ir.IntegerAttr.get(i32_type, 1),
-      called_computations=ir.ArrayAttr.get([]),
+  if all(core.is_constant_shape(aval_out.shape) for aval_out in ctx.avals_out):
+    result_shapes = None
+  else:
+    result_shapes = [
+        mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, aval_out.shape))
+        for aval_out in ctx.avals_out]
+  extra_attributes = None
+  # Add kernel_name and kernel_metadata as attributes to the custom call op.
+  # This is because we do not want to pollute the backend_config with this
+  # information.
+  if kernel_name is not None:
+    extra_attributes = dict(kernel_name=ir.StringAttr.get(kernel_name))
+  call = mlir.custom_call(
+      "tpu_custom_call",
+      result_types=result_types,
+      operands=in_nodes,
+      backend_config=config.to_json(),
+      api_version=1,
+      operand_output_aliases=dict(input_output_aliases),
       operand_layouts=_avals_to_layouts(ctx.avals_in),
       result_layouts=_avals_to_layouts(ctx.avals_out),
-      output_operand_aliases=ir.ArrayAttr.get([
-          hlo.OutputOperandAlias.get(
-              # if len(result_types) == 1 then the aliasing refers implicitly to
-              # the only output.
-              output_tuple_indices=[output_idx]
-              if len(out_avals) > 1
-              else [],
-              operand_index=input_idx,
-              operand_tuple_indices=[],
-          )
-          for input_idx, output_idx in input_output_aliases
-      ]),
-  )
+      result_shapes=result_shapes,
+      extra_attributes=extra_attributes)
 
-  # Add kernel_name and kernel_regeneration_metadata as attributes to the
-  # custom call op. This is because we do not want to pollute the backend_config
-  # with this information.
-  if kernel_name is not None:
-    call.attributes["kernel_name"] = ir.StringAttr.get(kernel_name)
-  if kernel_regeneration_metadata is not None:
-    call.attributes["kernel_regeneration_metadata"] = ir.StringAttr.get(
-        base64.b64encode(kernel_regeneration_metadata)
-    )
-  if multiple_results:
-    results = [stablehlo.get_tuple_element(call, mlir.i32_attr(i))
-               for i in range(len(out_avals))]
-  else:
-    results = call.results
-  return results
+  return call.results
 
 
 mlir.register_lowering(tpu_custom_call_p, _tpu_custom_call_lowering,
@@ -272,14 +285,18 @@ mlir.register_lowering(tpu_custom_call_p, _tpu_custom_call_lowering,
 def _lower_tpu_kernel(
     module: ir.Module,
     hardware_generation: int,
+    target_shape: tuple[int, int],
+    kernel_name: str | None = None,
 ) -> ir.Module:
   """Runs MLIR passes lowering the given module to an MLIR module.
 
-  Uses Python versions of infer-memref-layout and apply-vector-layout.
+  Uses Python versions of canonicalize-mosaic,infer-memref-layout and
+    apply-vector-layout.
 
   Args:
     module: The MLIR module to lower.
     hardware_generation: The TPU hardware generation to target.
+    target_shape: The target shape of (sublane_count, lane_count).
 
   Returns:
     An MLIR module implementing the kernel.
@@ -289,14 +306,21 @@ def _lower_tpu_kernel(
   except ir.MLIRError as e:
     raise ValueError("The compiled module fails MLIR verification") from e
 
+  timestamp = time.time_ns()
+  dump_cnt = [0]
+
+  def get_dump_file_prefix() -> str:
+    s = f"{timestamp}-{dump_cnt[0]:04}"
+    dump_cnt[0] += 1
+    return s
+
   with module.context as ctx, module.operation.location as _:
     ctx.append_dialect_registry(mlir.upstream_dialects)
     ctx.load_all_available_dialects()
     tpu.register_dialect(ctx)
     mhlo.register_mhlo_dialect(ctx)
     mhlo.register_mhlo_passes()
-
-    dump_mlir(module, "original")
+    dump_mlir(module, "original", get_dump_file_prefix(), kernel_name)
 
     if _MOSAIC_ALLOW_HLO.value:
       # Run hlo dialect conversion: hlo -> linalg -> vector.
@@ -307,14 +331,22 @@ def _lower_tpu_kernel(
       ]
       pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
       pipeline.run(module.operation)
-      dump_mlir(module, "post-hlo-conversion")
+      dump_mlir(module, "post-hlo-conversion", get_dump_file_prefix(), kernel_name)
 
+    sl_cnt, l_cnt = target_shape
+    # Note: we don't pass the TpuTilingFlags here, since we don't know the
+    # tiling decisions made by the compiler / what flags are enabled at this
+    # point, so we assume everything can be tiled up to default tiling.
     pipeline = [
-        f"func.func(tpu-infer-memref-layout{{hardware-generation={hardware_generation}}})"
+        "func.func(tpu-infer-memref-layout{"
+        f" hardware-generation={hardware_generation}"
+        f" sublane-count={sl_cnt}"
+        f" lane-count={l_cnt}"
+        "})"
     ]
     pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
     pipeline.run(module.operation)
-    dump_mlir(module, "post-infer-memref-layout")
+    dump_mlir(module, "post-infer-memref-layout", get_dump_file_prefix(), kernel_name)
 
     pipeline = [
         "canonicalize",
@@ -322,7 +354,12 @@ def _lower_tpu_kernel(
     ]
     pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
     pipeline.run(module.operation)
-    dump_mlir(module, "post-simplify")
+    dump_mlir(
+        module,
+        "post-infer-memref-layout-simplify",
+        get_dump_file_prefix(),
+        kernel_name,
+    )
 
     try:
       on_device_checks = FLAGS["xla_mosaic_on_device_checks"].value
@@ -336,60 +373,84 @@ def _lower_tpu_kernel(
             "builtin.module(func.func(debug-assert-insertion))"
         )
         pipeline.run(module.operation)
-        dump_mlir(module, "post-assert-insertion")
+        dump_mlir(module, "post-assert-insertion", get_dump_file_prefix(), kernel_name)
       elif checks:
         checks.discard("bounds")
         raise ValueError(
             f"Unrecognized on-device check categories: {', '.join(checks)}"
         )
 
+    # Legacy pipeline always runs in compatibility mode.
+    compatibility_mode = True
     pipeline = [
-        "func.func(tpu-infer-vector-layout{sublane-count=8 lane-count=128})",
+        (
+            f"func.func(tpu-canonicalize-mosaic{{hardware-generation={hardware_generation} compatibility-mode={compatibility_mode}}})"
+        ),
     ]
     pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
     pipeline.run(module.operation)
-    dump_mlir(module, "post-infer-vector-layout")
+    dump_mlir(module, "post-canonicalize-mosaic", get_dump_file_prefix(), kernel_name)
+
+    pipeline = [
+        (
+            "func.func(tpu-infer-vector-layout{"
+            f" hardware-generation={hardware_generation}"
+            f" sublane-count={sl_cnt} lane-count={l_cnt}"
+            "})"
+        ),
+    ]
+    pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
+    pipeline.run(module.operation)
+    dump_mlir(module, "post-infer-vector-layout", get_dump_file_prefix(), kernel_name)
+
+    pipeline = [
+        (
+            "func.func(tpu-relayout-insertion{"
+            f" sublane-count={sl_cnt} lane-count={l_cnt}"
+            "})"
+        ),
+    ]
+    pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
+    pipeline.run(module.operation)
+    dump_mlir(module, "post-relayout-insertion", get_dump_file_prefix(), kernel_name)
 
     mxu_size = 128 if hardware_generation < 6 else 256
     pipeline = [
-        "func.func(tpu-apply-vector-layout{sublane-count=8 lane-count=128"
+        "func.func(tpu-apply-vector-layout{"
+        f" sublane-count={sl_cnt} lane-count={l_cnt}"
         f" hardware-generation={hardware_generation}"
         f" mxu-contracting-size={mxu_size} mxu-noncontracting-size={mxu_size}"
+        f" max-sublanes-in-scratch={sl_cnt * (sl_cnt + 1)}"
         "})"
     ]
     pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
     pipeline.run(module.operation)
-    dump_mlir(module, "post-apply-vector-layout")
+    dump_mlir(module, "post-apply-vector-layout", get_dump_file_prefix(), kernel_name)
 
-    pipeline = PassManager.parse("builtin.module(canonicalize)")
+    pipeline = [
+        "canonicalize",
+        "cse",
+    ]
+    pipeline = PassManager.parse(f"builtin.module({','.join(pipeline)})")
     pipeline.run(module.operation)
-    dump_mlir(module, "pre-lower-to-llo")
+    dump_mlir(
+        module,
+        "post-apply-vector-layout-simplify",
+        get_dump_file_prefix(),
+        kernel_name,
+    )
 
     return module
 
 
-def as_tpu_kernel(
+def _lower_mosaic_module_to_asm(
     module: ir.Module,
-    out_type: Any,
     *,
-    cost_estimate: CostEstimate | None = None,
-    backend: str | xla_client.Client = "tpu",
-    device_type: str | None = None,
-    kernel_name: str | None = None,
-    kernel_regeneration_metadata: bytes | None = None,
-    vmem_limit_bytes: int | None = None,
-    flags: dict[str, bool | int | float] | None = None,
-    allow_input_fusion: list[bool] | None = None,
-    input_output_aliases: tuple[tuple[int, int], ...] = (),
-) -> Callable[..., Any]:
-  """Turns an MLIR Mosaic kernel into a JAX-compatible function."""
-  # We use jax.jit to make sure we hit the fast compilation cache.
-
-  if vmem_limit_bytes is not None and not isinstance(vmem_limit_bytes, int):
-    raise ValueError(
-        "vmem_limit_bytes must be an int: provided with a"
-        f" {type(vmem_limit_bytes)}."
-    )
+    backend: str,
+    device_type: str | None,
+    kernel_name: str | None,
+    ir_version: int | None = None,
+) -> tuple[ir.Module, tuple[bool, bool, bool, bool]]:
   has_communication, has_custom_barrier = tpu.private_has_communication(
       module.operation
   )
@@ -397,10 +458,11 @@ def as_tpu_kernel(
   needs_layout_passes = not device_type
   # We'll mutate the module, so clone it
   with module.context as ctx, module.operation.location as _:
-    module = ir.Module.parse(
-        module.operation.get_asm(binary=True, enable_debug_info=True)
-    )
     if needs_layout_passes and _MOSAIC_USE_PYTHON_PIPELINE.value:
+      module = ir.Module.parse(
+          module.operation.get_asm(binary=True, enable_debug_info=True)
+      )
+      module_op = module.operation
       some_tpu = jax.devices(backend)[0]
       device_kind = some_tpu.device_kind
       if not device_kind.startswith("TPU v"):
@@ -409,102 +471,333 @@ def as_tpu_kernel(
             "tpu_custom_call cannot be lowered on a machine without TPUs "
             "when mosaic_use_python_pipeline=True.")
       hardware_generation = int(device_kind[len("TPU v")])
-      module = _lower_tpu_kernel(module, hardware_generation)
+      target_shape = get_target_shape(hardware_generation)
+      module = _lower_tpu_kernel(
+          module, hardware_generation, target_shape=target_shape, kernel_name=kernel_name,
+      )
       needs_hlo_passes = False
       needs_layout_passes = False
+    else:
+      module_op = module.operation.clone()
     prev_allow_unregistered_dialects = ctx.allow_unregistered_dialects
     ctx.allow_unregistered_dialects = True
+    # TODO(apaszke): Remove once the minimum jaxlib version is at least 0.4.37.
+    if jax.version._version_as_tuple(jax.lib.__version__) < (0, 4, 37):
+      target_version = ""
+    else:
+      target_version = (
+          f"target-version={ir_version}" if ir_version is not None else ""
+      )
     try:
-      pipeline = PassManager.parse("builtin.module(mosaic-serde{serialize=true})")
-      pipeline.run(module.operation)
+      pipeline = PassManager.parse(
+          "builtin.module(mosaic-serde{serialize=true " + target_version + "})"
+      )
+      pipeline.run(module_op)
     finally:
       ctx.allow_unregistered_dialects = prev_allow_unregistered_dialects
     bytecode_buffer = io.BytesIO()
-    module.operation.write_bytecode(bytecode_buffer, desired_version=0)
+    module_op.write_bytecode(bytecode_buffer, desired_version=0)
     asm = bytecode_buffer.getvalue()
+    return asm, (
+        has_communication,
+        has_custom_barrier,
+        needs_hlo_passes,
+        needs_layout_passes,
+    )
 
-  # TODO(amagni): Kernel name and regeneration metadata could alternatively be
-  # added as a custom attribute to the MLIR call op rather than including them
-  # in the backend_config.
-  return _lowered_as_tpu_kernel(
-      asm,
-      out_type,
-      needs_hlo_passes=needs_hlo_passes,
-      needs_layout_passes=needs_layout_passes,
+
+def _get_device_type(module: ir.Module) -> str | None:
+  """Determines the device type based on the core_type annotations."""
+  sparsecore_func_found = False
+  tensorcore_func_found = False
+
+  def assign_device_type_based_on_core_type(op: ir.Operation) -> ir.WalkResult:
+    nonlocal sparsecore_func_found
+    nonlocal tensorcore_func_found
+    if op.name == "func.func":
+      if "tpu.core_type" in op.attributes:
+        core_type = op.attributes["tpu.core_type"]
+        if str(core_type) in [
+            f"#tpu.core_type<{c}>"
+            for c in ["sc_scalar_subcore", "sc_vector_subcore"]
+        ]:
+          sparsecore_func_found = True
+          if tensorcore_func_found:
+            return ir.WalkResult.INTERRUPT
+          return ir.WalkResult.SKIP
+        if str(core_type) == "#tpu.core_type<tc>":
+          tensorcore_func_found = True
+          return ir.WalkResult.SKIP
+        raise ValueError(f"Unknown core type: {core_type}")
+    return ir.WalkResult.ADVANCE
+
+  module.operation.walk(
+      assign_device_type_based_on_core_type, walk_order=ir.WalkOrder.PRE_ORDER
+  )
+  if tensorcore_func_found and sparsecore_func_found:
+    raise ValueError(
+        "A single Mosaic kernel cannot contain both "
+        "TensorCore and SparseCore functions."
+    )
+  if sparsecore_func_found:
+    return "sparsecore"
+  return None
+
+
+def _lower_to_custom_call_config(
+    module: ir.Module,
+    *,
+    backend: str,
+    device_type: str | None,
+    vmem_limit_bytes: int | None,
+    cost_estimate: CostEstimate | None,
+    flags: dict[str, bool | int | float] | None,
+    allow_input_fusion: list[bool] | None,
+    internal_scratch_in_bytes: int | None,
+    collective_id: int | None,
+    serialization_format: int | None,
+    output_memory_spaces: tuple[MemorySpace | None, ...] | None = None,
+    kernel_name: str | None = None,
+    ir_version: int | None = None,
+) -> CustomCallBackendConfig:
+  lowered_module_asm, (
+      has_communication,
+      has_custom_barrier,
+      needs_hlo_passes,
+      needs_layout_passes,
+  ) = _lower_mosaic_module_to_asm(
+      module,
+      backend=backend,
       device_type=device_type,
-      has_communication=has_communication,
-      has_custom_barrier=has_custom_barrier,
       kernel_name=kernel_name,
-      kernel_regeneration_metadata=kernel_regeneration_metadata,
-      cost_estimate=cost_estimate,
+      ir_version=ir_version,
+  )
+  return _lowered_to_custom_call_config(
+      lowered_module_asm,
       vmem_limit_bytes=vmem_limit_bytes,
+      cost_estimate=cost_estimate,
       flags=flags,
       allow_input_fusion=allow_input_fusion,
+      internal_scratch_in_bytes=internal_scratch_in_bytes,
+      collective_id=collective_id,
+      device_type=device_type,
+      serialization_format=serialization_format,
+      has_custom_barrier=has_custom_barrier,
+      has_communication=has_communication,
+      needs_hlo_passes=needs_hlo_passes,
+      needs_layout_passes=needs_layout_passes,
+      output_memory_spaces=output_memory_spaces,
+  )
+
+
+def _lowered_to_custom_call_config(
+    lowered_module_asm: bytes,
+    *,
+    vmem_limit_bytes: int | None,
+    cost_estimate: CostEstimate | None,
+    flags: dict[str, bool | int | float] | None,
+    allow_input_fusion: list[bool] | None,
+    internal_scratch_in_bytes: int | None,
+    collective_id: int | None,
+    serialization_format: int | None,
+    has_custom_barrier: bool,
+    has_communication: bool,
+    needs_hlo_passes: bool,
+    needs_layout_passes: bool,
+    device_type: str | None,
+    output_memory_spaces: tuple[MemorySpace | None, ...] | None = None,
+):
+  if has_custom_barrier:
+    if collective_id is None:
+      raise ValueError(
+          "collective_id has to be specified when using a custom barrier"
+      )
+  elif collective_id is not None:
+    raise ValueError(
+        "collective_id has to be unspecified or None when not using a custom"
+        " barrier"
+    )
+  if vmem_limit_bytes is not None and not isinstance(vmem_limit_bytes, int):
+    raise ValueError(
+        "vmem_limit_bytes must be an int: provided with a"
+        f" {type(vmem_limit_bytes)}."
+    )
+  config = CustomCallBackendConfig(
+      lowered_module_asm,
+      has_communication,
+      collective_id,
+      device_type,
+      cost_estimate,
+      needs_hlo_passes,
+      needs_layout_passes,
+      vmem_limit_bytes,
+      flags,
+      allow_input_fusion,
+      serialization_format,
+      internal_scratch_in_bytes,
+      output_memory_spaces,
+  )
+  return config
+
+
+def lower_module_to_custom_call(
+    ctx: mlir.LoweringRuleContext,
+    *in_nodes: ir.Value,
+    module: ir.Module,
+    out_type: Any,
+    backend: str,
+    kernel_name: str,
+    cost_estimate: CostEstimate | None,
+    vmem_limit_bytes: int | None,
+    flags: dict[str, bool | int | float] | None,
+    allow_input_fusion: list[bool] | None,
+    input_output_aliases: tuple[tuple[int, int], ...],
+    internal_scratch_in_bytes: int | None,
+    collective_id: int | None,
+    serialization_format: int | None,
+    output_memory_spaces: tuple[MemorySpace | None, ...] | None,
+    device_type: str | None,
+) -> Sequence[ir.Value]:
+  config = _lower_to_custom_call_config(
+      module,
+      backend=backend,
+      vmem_limit_bytes=vmem_limit_bytes,
+      cost_estimate=cost_estimate,
+      flags=flags,
+      allow_input_fusion=allow_input_fusion,
+      internal_scratch_in_bytes=internal_scratch_in_bytes,
+      collective_id=collective_id,
+      device_type=device_type,
+      serialization_format=serialization_format,
+      output_memory_spaces=output_memory_spaces,
+      kernel_name=kernel_name,
+      ir_version=FWD_COMPAT_IR_VERSION if ctx.is_forward_compat() else None,
+  )
+  return _tpu_custom_call_lowering(
+      ctx,
+      *in_nodes,
+      config=config,
+      kernel_name=kernel_name,
+      out_avals=out_type,
       input_output_aliases=input_output_aliases,
   )
 
 
-def _lowered_as_tpu_kernel(
-    lowered_module_asm: bytes,
+def as_tpu_kernel(
+    module: ir.Module,
     out_type: Any,
     *,
     cost_estimate: CostEstimate | None = None,
+    backend: str | xla_client.Client = "tpu",
+    kernel_name: str | None = None,
+    vmem_limit_bytes: int | None = None,
+    flags: dict[str, bool | int | float] | None = None,
+    allow_input_fusion: list[bool] | None = None,
+    input_output_aliases: tuple[tuple[int, int], ...] = (),
+    internal_scratch_in_bytes: int | None = None,
+    collective_id: int | None = None,
+    serialization_format: int | None = 1,
+    output_memory_spaces: tuple[MemorySpace | None, ...] | None = None,
+) -> Callable[..., Any]:
+  """Turns an MLIR Mosaic kernel into a JAX-compatible function."""
+  device_type = _get_device_type(module)
+  config = _lower_to_custom_call_config(
+      module,
+      backend=backend,
+      device_type=device_type,
+      vmem_limit_bytes=vmem_limit_bytes,
+      cost_estimate=cost_estimate,
+      flags=flags,
+      allow_input_fusion=allow_input_fusion,
+      internal_scratch_in_bytes=internal_scratch_in_bytes,
+      collective_id=collective_id,
+      serialization_format=serialization_format,
+      output_memory_spaces=output_memory_spaces,
+      kernel_name=kernel_name,
+  )
+  return _as_jax_callable(
+      config,
+      out_type,
+      kernel_name=kernel_name,
+      input_output_aliases=input_output_aliases,
+  )
+
+
+def lowered_as_tpu_kernel(
+    lowered_module: ir.Module,
+    out_type: Any,
+    *,
+    collective_id: int | None = None,
+    cost_estimate: CostEstimate | None = None,
     needs_hlo_passes: bool = False,
     needs_layout_passes: bool = False,
     device_type: str | None = None,
     has_communication: bool = False,
     has_custom_barrier: bool = False,
     kernel_name: str | None = None,
-    kernel_regeneration_metadata: bytes | None = None,
     vmem_limit_bytes: int | None = None,
     flags: dict[str, bool | int | float] | None = None,
     allow_input_fusion: list[bool] | None = None,
     input_output_aliases: tuple[tuple[int, int], ...] = (),
-    serialization_format: int | None = 1,
-):
-  """Turns a low-level MLIR Mosaic kernel into a JAX-compatible function."""
+    serialization_format: int | None = None,
+    internal_scratch_in_bytes: int | None = None,
+) -> Callable[..., Any]:
+  lowered_module_asm = lowered_module.operation.get_asm(
+      binary=True, enable_debug_info=True
+  )
+  config = _lowered_to_custom_call_config(
+      lowered_module_asm,
+      vmem_limit_bytes=vmem_limit_bytes,
+      cost_estimate=cost_estimate,
+      flags=flags,
+      allow_input_fusion=allow_input_fusion,
+      internal_scratch_in_bytes=internal_scratch_in_bytes,
+      collective_id=collective_id,
+      device_type=device_type,
+      serialization_format=serialization_format,
+      has_custom_barrier=has_custom_barrier,
+      has_communication=has_communication,
+      needs_hlo_passes=needs_hlo_passes,
+      needs_layout_passes=needs_layout_passes,
+  )
+  return _as_jax_callable(
+      config,
+      out_type,
+      kernel_name=kernel_name,
+      input_output_aliases=input_output_aliases,
+  )
+
+
+def _as_jax_callable(
+    config: CustomCallBackendConfig,
+    out_type: Any,
+    *,
+    kernel_name: str | None,
+    input_output_aliases: tuple[tuple[int, int], ...],
+) -> Callable[..., Any]:
   unpack = False
   if not isinstance(out_type, collections.abc.Iterable):
     out_type = (out_type,)
     unpack = True
   out_avals = tuple(core.ShapedArray(ty.shape, ty.dtype) for ty in out_type)
-  def apply_kernel(*args, collective_id: int | None = None):
-    if has_custom_barrier:
-      if collective_id is None:
-        raise ValueError(
-            "collective_id has to be specified when using a custom barrier"
-        )
-    elif collective_id is not None:
-      raise ValueError(
-          "collective_id has to be unspecified or None when not using a custom"
-          " barrier"
-      )
-    config = CustomCallBackendConfig(
-        lowered_module_asm,
-        has_communication,
-        collective_id,
-        device_type,
-        cost_estimate,
-        needs_hlo_passes,
-        needs_layout_passes,
-        vmem_limit_bytes,
-        flags,
-        allow_input_fusion,
-        serialization_format=serialization_format,
-    )
+
+  # We use jax.jit to make sure we hit the fast compilation cache.
+  def apply_kernel(*args):
     result = tpu_custom_call_p.bind(
         *args,
         config=config,
         kernel_name=kernel_name,
-        kernel_regeneration_metadata=kernel_regeneration_metadata,
         out_avals=out_avals,
         input_output_aliases=input_output_aliases,
     )
     return result[0] if unpack else result
-  return jax.jit(apply_kernel, static_argnames=["collective_id"])
+
+  return jax.jit(apply_kernel)
 
 
-def dump_mlir(module: ir.Module, name: str):
+def dump_mlir(
+    module: ir.Module, name: str, prefix: str, kernel_name: str | None = None
+):
   """A helper function to dump mosaic mlir module"""
   try:
     should_dump = FLAGS["xla_mosaic_dump_to"].value
@@ -513,6 +806,8 @@ def dump_mlir(module: ir.Module, name: str):
   if should_dump == "sponge":
     outdir = os.environ.get("TEST_UNDECLARED_OUTPUTS_DIR", None)
     if outdir:
-      path = os.path.join(outdir, f"{time.time_ns()}-mosaic-dump-{name}.txt")
+      if kernel_name:
+        name = f"{kernel_name}-{name}"
+      path = os.path.join(outdir, f"{prefix}-mosaic-dump-{name}-py.txt")
       with open(path, "w") as f:
         f.write(str(module))
