@@ -14,14 +14,14 @@
 from __future__ import annotations
 
 import collections
-from collections.abc import Hashable, Iterable
-from dataclasses import dataclass
+from collections.abc import Callable, Hashable, Iterable, Sequence
+import dataclasses
 import difflib
 import functools
 from functools import partial
 import operator as op
 import textwrap
-from typing import Any, Callable, NamedTuple, Sequence, TypeVar, Union, overload
+from typing import Any, NamedTuple, TypeVar, overload
 
 from jax._src import traceback_util
 from jax._src.lib import pytree
@@ -74,66 +74,13 @@ dispatch_registry.__name__ = "dispatch_registry"
 def tree_flatten(tree: Any,
                  is_leaf: Callable[[Any], bool] | None = None
                  ) -> tuple[list[Leaf], PyTreeDef]:
-  """Flattens a pytree.
-
-  The flattening order (i.e. the order of elements in the output list)
-  is deterministic, corresponding to a left-to-right depth-first tree
-  traversal.
-
-  Args:
-    tree: a pytree to flatten.
-    is_leaf: an optionally specified function that will be called at each
-      flattening step. It should return a boolean, with true stopping the
-      traversal and the whole subtree being treated as a leaf, and false
-      indicating the flattening should traverse the current object.
-
-  Returns:
-    A pair where the first element is a list of leaf values and the second
-    element is a treedef representing the structure of the flattened tree.
-
-  Example:
-    >>> import jax
-    >>> vals, treedef = jax.tree.flatten([1, (2, 3), [4, 5]])
-    >>> vals
-    [1, 2, 3, 4, 5]
-    >>> treedef
-    PyTreeDef([*, (*, *), [*, *]])
-
-  See Also:
-    - :func:`jax.tree.leaves`
-    - :func:`jax.tree.structure`
-    - :func:`jax.tree.unflatten`
-  """
+  """Alias of :func:`jax.tree.flatten`."""
   return default_registry.flatten(tree, is_leaf)
 
 
 @export
 def tree_unflatten(treedef: PyTreeDef, leaves: Iterable[Leaf]) -> Any:
-  """Reconstructs a pytree from the treedef and the leaves.
-
-  The inverse of :func:`tree_flatten`.
-
-  Args:
-    treedef: the treedef to reconstruct
-    leaves: the iterable of leaves to use for reconstruction. The iterable must
-      match the leaves of the treedef.
-
-  Returns:
-    The reconstructed pytree, containing the ``leaves`` placed in the structure
-    described by ``treedef``.
-
-  Example:
-    >>> import jax
-    >>> vals, treedef = jax.tree.flatten([1, (2, 3), [4, 5]])
-    >>> newvals = [100, 200, 300, 400, 500]
-    >>> jax.tree.unflatten(treedef, newvals)
-    [100, (200, 300), [400, 500]]
-
-  See Also:
-    - :func:`jax.tree.flatten`
-    - :func:`jax.tree.leaves`
-    - :func:`jax.tree.structure`
-  """
+  """Alias of :func:`jax.tree.unflatten`."""
   return treedef.unflatten(leaves)
 
 
@@ -141,28 +88,7 @@ def tree_unflatten(treedef: PyTreeDef, leaves: Iterable[Leaf]) -> Any:
 def tree_leaves(tree: Any,
                 is_leaf: Callable[[Any], bool] | None = None
                 ) -> list[Leaf]:
-  """Gets the leaves of a pytree.
-
-  Args:
-    tree: the pytree for which to get the leaves
-    is_leaf : an optionally specified function that will be called at each
-      flattening step. It should return a boolean, which indicates whether the
-      flattening should traverse the current object, or if it should be stopped
-      immediately, with the whole subtree being treated as a leaf.
-
-  Returns:
-    leaves: a list of tree leaves.
-
-  Example:
-    >>> import jax
-    >>> jax.tree.leaves([1, (2, 3), [4, 5]])
-    [1, 2, 3, 4, 5]
-
-  See Also:
-    - :func:`jax.tree.flatten`
-    - :func:`jax.tree.structure`
-    - :func:`jax.tree.unflatten`
-  """
+  """Alias of :func:`jax.tree.leaves`."""
   return default_registry.flatten(tree, is_leaf)[0]
 
 
@@ -170,28 +96,7 @@ def tree_leaves(tree: Any,
 def tree_structure(tree: Any,
                    is_leaf: None | (Callable[[Any],
                                               bool]) = None) -> PyTreeDef:
-  """Gets the treedef for a pytree.
-
-  Args:
-    tree: the pytree for which to get the leaves
-    is_leaf : an optionally specified function that will be called at each
-      flattening step. It should return a boolean, which indicates whether the
-      flattening should traverse the current object, or if it should be stopped
-      immediately, with the whole subtree being treated as a leaf.
-
-  Returns:
-    pytreedef: a PyTreeDef representing the structure of the tree.
-
-  Example:
-    >>> import jax
-    >>> jax.tree.structure([1, (2, 3), [4, 5]])
-    PyTreeDef([*, (*, *), [*, *]])
-
-  See Also:
-    - :func:`jax.tree.flatten`
-    - :func:`jax.tree.leaves`
-    - :func:`jax.tree.unflatten`
-  """
+  """Alias of :func:`jax.tree.structure`."""
   return default_registry.flatten(tree, is_leaf)[1]
 
 
@@ -205,7 +110,7 @@ def treedef_tuple(treedefs: Iterable[PyTreeDef]) -> PyTreeDef:
   Returns:
     a single treedef representing a tuple of the structures
 
-  Example:
+  Examples:
     >>> import jax
     >>> x = [1, 2, 3]
     >>> y = {'a': 4, 'b': 5}
@@ -256,7 +161,7 @@ def treedef_is_leaf(treedef: PyTreeDef) -> bool:
   Returns:
     True if treedef is a leaf (i.e. has a single node); False otherwise.
 
-  Example:
+  Examples:
     >>> import jax
     >>> tree1 = jax.tree.structure(1)
     >>> jax.tree_util.treedef_is_leaf(tree1)
@@ -288,7 +193,7 @@ def all_leaves(iterable: Iterable[Any],
   Returns:
     A boolean indicating if all elements in the input are leaves.
 
-  Example:
+  Examples:
     >>> import jax
     >>> tree = {"a": [1, 2, 3]}
     >>> assert all_leaves(jax.tree_util.tree_leaves(tree))
@@ -303,12 +208,21 @@ def all_leaves(iterable: Iterable[Any],
 
 _Children = TypeVar("_Children", bound=Iterable[Any])
 _AuxData = TypeVar("_AuxData", bound=Hashable)
+KeyEntry = TypeVar("KeyEntry", bound=Any)
+KeyLeafPair = tuple[KeyEntry, Any]
+KeyLeafPairs = Iterable[KeyLeafPair]
+KeyPath = tuple[KeyEntry, ...]
 
 
 @export
-def register_pytree_node(nodetype: type[T],
-                         flatten_func: Callable[[T], tuple[_Children, _AuxData]],
-                         unflatten_func: Callable[[_AuxData, _Children], T]) -> None:
+def register_pytree_node(
+    nodetype: type[T],
+    flatten_func: Callable[[T], tuple[_Children, _AuxData]],
+    unflatten_func: Callable[[_AuxData, _Children], T],
+    flatten_with_keys_func: (
+        Callable[[T], tuple[KeyLeafPairs, _AuxData]] | None
+    ) = None,
+) -> None:
   """Extends the set of types that are considered internal nodes in pytrees.
 
   See :ref:`example usage <pytrees>`.
@@ -331,7 +245,7 @@ def register_pytree_node(nodetype: type[T],
     - :func:`~jax.tree_util.register_pytree_node_class`
     - :func:`~jax.tree_util.register_pytree_with_keys_class`
 
-  Example:
+  Examples:
     First we'll define a custom type:
 
     >>> class MyContainer:
@@ -373,9 +287,15 @@ def register_pytree_node(nodetype: type[T],
     >>> jax.jit(f)(m)
     Array([1., 2., 3., 4., 5.], dtype=float32)
   """
-  default_registry.register_node(nodetype, flatten_func, unflatten_func)
-  none_leaf_registry.register_node(nodetype, flatten_func, unflatten_func)
-  dispatch_registry.register_node(nodetype, flatten_func, unflatten_func)
+  default_registry.register_node(  # type: ignore[call-arg]
+      nodetype, flatten_func, unflatten_func, flatten_with_keys_func   # type: ignore[arg-type]
+  )
+  none_leaf_registry.register_node(  # type: ignore[call-arg]
+      nodetype, flatten_func, unflatten_func, flatten_with_keys_func   # type: ignore[arg-type]
+  )
+  dispatch_registry.register_node(  # type: ignore[call-arg]
+      nodetype, flatten_func, unflatten_func, flatten_with_keys_func   # type: ignore[arg-type]
+  )
   _registry[nodetype] = _RegistryEntry(flatten_func, unflatten_func)
 
 
@@ -401,7 +321,7 @@ def register_pytree_node_class(cls: Typ) -> Typ:
     - :func:`~jax.tree_util.register_pytree_with_keys`
     - :func:`~jax.tree_util.register_pytree_with_keys_class`
 
-  Example:
+  Examples:
     Here we'll define a custom container that will be compatible with :func:`jax.jit`
     and other JAX transformations:
 
@@ -432,42 +352,7 @@ def tree_map(f: Callable[..., Any],
              tree: Any,
              *rest: Any,
              is_leaf: Callable[[Any], bool] | None = None) -> Any:
-  """Maps a multi-input function over pytree args to produce a new pytree.
-
-  Args:
-    f: function that takes ``1 + len(rest)`` arguments, to be applied at the
-      corresponding leaves of the pytrees.
-    tree: a pytree to be mapped over, with each leaf providing the first
-      positional argument to ``f``.
-    rest: a tuple of pytrees, each of which has the same structure as ``tree``
-      or has ``tree`` as a prefix.
-    is_leaf: an optionally specified function that will be called at each
-      flattening step. It should return a boolean, which indicates whether the
-      flattening should traverse the current object, or if it should be stopped
-      immediately, with the whole subtree being treated as a leaf.
-
-  Returns:
-    A new pytree with the same structure as ``tree`` but with the value at each
-    leaf given by ``f(x, *xs)`` where ``x`` is the value at the corresponding
-    leaf in ``tree`` and ``xs`` is the tuple of values at corresponding nodes in
-    ``rest``.
-
-  Examples:
-
-    >>> import jax.tree_util
-    >>> jax.tree_util.tree_map(lambda x: x + 1, {"x": 7, "y": 42})
-    {'x': 8, 'y': 43}
-
-    If multiple inputs are passed, the structure of the tree is taken from the
-    first input; subsequent inputs need only have ``tree`` as a prefix:
-
-    >>> jax.tree_util.tree_map(lambda x, y: [x] + y, [5, 6], [[7, 9], [1, 2]])
-    [[5, 7, 9], [6, 1, 2]]
-
-  See Also:
-    - :func:`jax.tree.leaves`
-    - :func:`jax.tree.reduce`
-  """
+  """Alias of :func:`jax.tree.map`."""
   leaves, treedef = tree_flatten(tree, is_leaf)
   all_leaves = [leaves] + [treedef.flatten_up_to(r) for r in rest]
   return treedef.unflatten(f(*xs) for xs in zip(*all_leaves))
@@ -487,7 +372,7 @@ def build_tree(treedef: PyTreeDef, xs: Any) -> Any:
   See Also:
     - :func:`jax.tree.unflatten`
 
-  Example:
+  Examples:
     >>> import jax
     >>> tree = [(1, 2), {'a': 3, 'b': 4}]
     >>> treedef = jax.tree.structure(tree)
@@ -507,31 +392,7 @@ def build_tree(treedef: PyTreeDef, xs: Any) -> Any:
 @export
 def tree_transpose(outer_treedef: PyTreeDef, inner_treedef: PyTreeDef | None,
                    pytree_to_transpose: Any) -> Any:
-  """Transform a tree having tree structure (outer, inner) into one having structure (inner, outer).
-
-  Args:
-    outer_treedef: PyTreeDef representing the outer tree.
-    inner_treedef: PyTreeDef representing the inner tree.
-      If None, then it will be inferred from outer_treedef and the structure of
-      pytree_to_transpose.
-    pytree_to_transpose: the pytree to be transposed.
-
-  Returns:
-    transposed_pytree: the transposed pytree.
-
-  Examples:
-    >>> import jax
-    >>> tree = [(1, 2, 3), (4, 5, 6)]
-    >>> inner_structure = jax.tree.structure(('*', '*', '*'))
-    >>> outer_structure = jax.tree.structure(['*', '*'])
-    >>> jax.tree.transpose(outer_structure, inner_structure, tree)
-    ([1, 4], [2, 5], [3, 6])
-
-    Inferring the inner structure:
-
-    >>> jax.tree.transpose(outer_structure, None, tree)
-    ([1, 4], [2, 5], [3, 6])
-  """
+  """Alias of :func:`jax.tree.transpose`."""
   flat, treedef = tree_flatten(pytree_to_transpose)
   if inner_treedef is None:
     inner_treedef = tree_structure(outer_treedef.flatten_up_to(pytree_to_transpose)[0])
@@ -553,7 +414,7 @@ def tree_transpose(outer_treedef: PyTreeDef, inner_treedef: PyTreeDef | None,
 # sufficiently queryable that we can express _replace_nones. That may mean once
 # we have a flatten_one function.
 _RegistryEntry = collections.namedtuple("_RegistryEntry", ["to_iter", "from_iter"])
-_registry = {
+_registry: dict[type[Any], _RegistryEntry] = {
     tuple: _RegistryEntry(lambda xs: (xs, None), lambda _, xs: tuple(xs)),
     list: _RegistryEntry(lambda xs: (xs, None), lambda _, xs: list(xs)),
     dict: _RegistryEntry(lambda xs: unzip2(sorted(xs.items()))[::-1],
@@ -592,30 +453,7 @@ def tree_reduce(function: Callable[[T, Any], T],
                 tree: Any,
                 initializer: Any = no_initializer,
                 is_leaf: Callable[[Any], bool] | None = None) -> T:
-  """Call reduce() over the leaves of a tree.
-
-  Args:
-    function: the reduction function
-    tree: the pytree to reduce over
-    initializer: the optional initial value
-    is_leaf : an optionally specified function that will be called at each
-      flattening step. It should return a boolean, which indicates whether the
-      flattening should traverse the current object, or if it should be stopped
-      immediately, with the whole subtree being treated as a leaf.
-
-  Returns:
-    result: the reduced value.
-
-  Examples:
-    >>> import jax
-    >>> import operator
-    >>> jax.tree.reduce(operator.add, [1, (2, 3), [4, 5, 6]])
-    21
-
-  See Also:
-    - :func:`jax.tree.leaves`
-    - :func:`jax.tree.map`
-  """
+  """Alias of :func:`jax.tree.reduce`."""
   if initializer is no_initializer:
     return functools.reduce(function, tree_leaves(tree, is_leaf=is_leaf))
   else:
@@ -623,42 +461,9 @@ def tree_reduce(function: Callable[[T, Any], T],
 
 
 @export
-def tree_all(tree: Any) -> bool:
-  """Call all() over the leaves of a tree.
-
-  Args:
-    tree: the pytree to evaluate
-
-  Returns:
-    result: boolean True or False
-
-  Examples:
-    >>> import jax
-    >>> jax.tree.all([True, {'a': True, 'b': (True, True)}])
-    True
-    >>> jax.tree.all([False, (True, False)])
-    False
-
-  See Also:
-    - :func:`jax.tree_util.tree_reduce`
-    - :func:`jax.tree_util.tree_leaves`
-  """
-  return all(tree_leaves(tree))
-
-
-register_pytree_node(
-  collections.OrderedDict,
-  lambda x: (tuple(x.values()), tuple(x.keys())),
-  lambda keys, values: collections.OrderedDict(safe_zip(keys, values)))
-
-def _flatten_defaultdict(d):
-  keys = tuple(sorted(d))
-  return tuple(d[k] for k in keys), (d.default_factory, keys)
-
-register_pytree_node(
-  collections.defaultdict,
-  _flatten_defaultdict,
-  lambda s, values: collections.defaultdict(s[0], safe_zip(s[1], values)))
+def tree_all(tree: Any, *, is_leaf: Callable[[Any], bool] | None = None) -> bool:
+  """Alias of :func:`jax.tree.all`."""
+  return all(tree_leaves(tree, is_leaf=is_leaf))
 
 
 class _HashableCallableShim:
@@ -772,11 +577,11 @@ def broadcast_prefix(prefix_tree: Any, full_tree: Any,
 
 
 # flatten_one_level is not exported.
-def flatten_one_level(pytree: Any) -> tuple[Iterable[Any], Hashable]:
+def flatten_one_level(tree: Any) -> tuple[Iterable[Any], Hashable]:
   """Flatten the given pytree node by one level.
 
   Args:
-    pytree: A valid pytree node, either built-in or registered via
+    tree: A valid pytree node, either built-in or registered via
       :func:`register_pytree_node` or related functions.
 
   Returns:
@@ -786,7 +591,7 @@ def flatten_one_level(pytree: Any) -> tuple[Iterable[Any], Hashable]:
     ValueError: If the given pytree is not a built-in or registered container
     via ``register_pytree_node`` or ``register_pytree_with_keys``.
 
-  Example:
+  Examples:
     >>> import jax
     >>> from jax._src.tree_util import flatten_one_level
     >>> flattened, meta = flatten_one_level({'a': [1, 2], 'b': {'c': 3}})
@@ -795,9 +600,21 @@ def flatten_one_level(pytree: Any) -> tuple[Iterable[Any], Hashable]:
     >>> meta
     ('a', 'b')
   """
-  out = default_registry.flatten_one_level(pytree)
+  out = default_registry.flatten_one_level(tree)
   if out is None:
-    raise ValueError(f"can't tree-flatten type: {type(pytree)}")
+    raise ValueError(f"can't tree-flatten type: {type(tree)}")
+  else:
+    return out
+
+
+# flatten_one_level_with_keys is not exported.
+def flatten_one_level_with_keys(
+    tree: Any,
+) -> tuple[Iterable[KeyLeafPair], Hashable]:
+  """Flatten the given pytree node by one level, with keys."""
+  out = default_registry.flatten_one_level_with_keys(tree)  # type: ignore
+  if out is None:
+    raise ValueError(f"can't tree-flatten type: {type(tree)}")
   else:
     return out
 
@@ -816,7 +633,7 @@ def equality_errors(
   """Helper to describe structural differences between two pytrees.
 
   Args:
-    tree1, tree2: pytrees to compare.
+    tree1, tree2: pytrees known to have different structure.
 
   Usage:
 
@@ -831,6 +648,15 @@ def equality_errors(
   """
   yield from _equality_errors((), tree1, tree2, is_leaf)
 
+def equality_errors_pytreedef(
+    tree1: PyTreeDef,
+    tree2: PyTreeDef) -> Iterable[tuple[KeyPath, str, str, str]]:
+  """Like `equality_errors` but invoked on PyTreeDef."""
+  # TODO(mattjj): make equality_errors not print type name, avoid metaclass
+  leaf = type("LeafMeta", (type,), dict(__repr__=lambda _: "pytree leaf"))("Leaf", (), {})()
+  return equality_errors(tree_unflatten(tree1, [leaf] * tree1.num_leaves),
+                         tree_unflatten(tree2, [leaf] * tree2.num_leaves))
+
 # TODO(mattjj): maybe share some logic with _prefix_error?
 def _equality_errors(path, t1, t2, is_leaf):
   # If both are leaves, this isn't a structure equality error.
@@ -889,45 +715,10 @@ def _equality_errors(path, t1, t2, is_leaf):
     yield from _equality_errors((*path, k), c1, c2, is_leaf)
 
 
-@export
-@dataclass(frozen=True)
-class SequenceKey():
-  """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`."""
-  idx: int
-  def __str__(self):
-    return f'[{self.idx!r}]'
-
-
-@export
-@dataclass(frozen=True)
-class DictKey():
-  """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`."""
-  key: Hashable
-  def __str__(self):
-    return f'[{self.key!r}]'
-
-
-@export
-@dataclass(frozen=True)
-class GetAttrKey():
-  """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`."""
-  name: str
-  def __str__(self):
-    return f'.{self.name}'
-
-
-@export
-@dataclass(frozen=True)
-class FlattenedIndexKey():
-  """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`."""
-  key: int
-  def __str__(self):
-    return f'[<flat index {self.key}>]'
-
-BuiltInKeyEntry = Union[SequenceKey, DictKey, GetAttrKey, FlattenedIndexKey]
-
-KeyEntry = TypeVar("KeyEntry", bound=Hashable)
-KeyPath = tuple[KeyEntry, ...]
+SequenceKey: Any = pytree.SequenceKey  # type: ignore
+DictKey: Any = pytree.DictKey  # type: ignore
+GetAttrKey: Any = pytree.GetAttrKey  # type: ignore
+FlattenedIndexKey: Any = pytree.FlattenedIndexKey  # type: ignore
 
 
 @export
@@ -940,15 +731,16 @@ def keystr(keys: KeyPath):
   Returns:
     A string that joins all string representations of the keys.
 
-  Example:
+  Examples:
     >>> import jax
     >>> keys = (0, 1, 'a', 'b')
     >>> jax.tree_util.keystr(keys)
     '01ab'
   """
-  return ''.join([str(k) for k in keys])
+  return ''.join(map(str, keys))
 
 
+# TODO(ivyzheng): remove this after another jaxlib release.
 class _RegistryWithKeypathsEntry(NamedTuple):
   flatten_with_keys: Callable[..., Any]
   unflatten_func: Callable[..., Any]
@@ -965,8 +757,7 @@ def _register_keypaths(
         flatten_with_keys, _registry[ty].from_iter
     )
 
-
-_registry_with_keypaths = {}
+_registry_with_keypaths: dict[type[Any], _RegistryWithKeypathsEntry] = {}
 
 _register_keypaths(
     tuple, lambda xs: tuple(SequenceKey(i) for i in range(len(xs)))
@@ -988,13 +779,9 @@ _register_keypaths(
 @export
 def register_pytree_with_keys(
     nodetype: type[T],
-    flatten_with_keys: Callable[
-        [T], tuple[Iterable[tuple[KeyEntry, Any]], _AuxData]
-    ],
+    flatten_with_keys: Callable[[T], tuple[Iterable[KeyLeafPair], _AuxData]],
     unflatten_func: Callable[[_AuxData, Iterable[Any]], T],
-    flatten_func: None | (
-        Callable[[T], tuple[Iterable[Any], _AuxData]]
-    ) = None,
+    flatten_func: None | (Callable[[T], tuple[Iterable[Any], _AuxData]]) = None,
 ):
   """Extends the set of types that are considered internal nodes in pytrees.
 
@@ -1017,7 +804,7 @@ def register_pytree_with_keys(
       This argument is optional and only needed for faster traversal when
       calling functions without keys like ``tree_map`` and ``tree_flatten``.
 
-  Example:
+  Examples:
     First we'll define a custom type:
 
     >>> class MyContainer:
@@ -1055,7 +842,9 @@ def register_pytree_with_keys(
       return [c for _, c in key_children], treedef
     flatten_func = flatten_func_impl
 
-  register_pytree_node(nodetype, flatten_func, unflatten_func)
+  register_pytree_node(
+      nodetype, flatten_func, unflatten_func, flatten_with_keys
+  )
   _registry_with_keypaths[nodetype] = _RegistryWithKeypathsEntry(
       flatten_with_keys, unflatten_func
   )
@@ -1086,7 +875,7 @@ def register_pytree_with_keys_class(cls: Typ) -> Typ:
     - :func:`~jax.tree_util.register_pytree_with_keys`
     - :func:`~jax.tree_util.register_pytree_node_class`
 
-  Example:
+  Examples:
     >>> from jax.tree_util import register_pytree_with_keys_class, GetAttrKey
     >>> @register_pytree_with_keys_class
     ... class Special:
@@ -1111,7 +900,10 @@ def register_pytree_with_keys_class(cls: Typ) -> Typ:
 
 @export
 def register_dataclass(
-    nodetype: Typ, data_fields: Sequence[str], meta_fields: Sequence[str]
+    nodetype: Typ,
+    data_fields: Sequence[str] | None = None,
+    meta_fields: Sequence[str] | None = None,
+    drop_fields: Sequence[str] = (),
 ) -> Typ:
   """Extends the set of types that are considered internal nodes in pytrees.
 
@@ -1127,24 +919,33 @@ def register_dataclass(
       attributes represent the whole of the object state, and can be passed
       as keywords to the class constructor to create a copy of the object.
       All defined attributes should be listed among ``meta_fields`` or ``data_fields``.
-    meta_fields: auxiliary data field names. These fields *must* contain static,
-      hashable, immutable objects, as these objects are used to generate JIT cache
-      keys. In particular, ``meta_fields`` cannot contain :class:`jax.Array` or
-      :class:`numpy.ndarray` objects.
-    data_fields: data field names. These fields *must* be JAX-compatible objects
-      such as arrays (:class:`jax.Array` or :class:`numpy.ndarray`), scalars, or
-      pytrees whose leaves are arrays or scalars. Note that ``data_fields`` may be
-      ``None``, as this is recognized by JAX as an empty pytree.
+    meta_fields: metadata field names: these are attributes which will be treated as
+      {term}`static` when this pytree is passed to :func:`jax.jit`. ``meta_fields`` is
+      optional only if ``nodetype`` is a dataclass, in which case individual fields can
+      be marked static via :func:`dataclasses.field` (see examples below).
+      Metadata fields *must* be static, hashable, immutable objects, as these objects
+      are used to generate JIT cache keys. In particular, metadata fields cannot contain
+      :class:`jax.Array` or :class:`numpy.ndarray` objects.
+    data_fields: data field names: these are attributes which will be treated as non-static
+      when this pytree is passed to :func:`jax.jit`. ``data_fields`` is optional only if
+      ``nodetype`` is a dataclass, in which case fields are assumed data fields unless
+      marked via :func:`dataclasses.field` (see examples below).
+      Data fields *must* be JAX-compatible objects such as arrays (:class:`jax.Array`
+      or :class:`numpy.ndarray`), scalars, or pytrees whose leaves are arrays or scalars.
+      Note that ``None`` is a valid data field, as JAX recognizes this as an empty pytree.
 
   Returns:
     The input class ``nodetype`` is returned unchanged after being added to JAX's
-    pytree registry. This return value allows ``register_dataclass`` to be partially
-    evaluated and used as a decorator as in the example below.
+    pytree registry, so that :func:`register_dataclass` can be used as a decorator.
 
-  Example:
+  Examples:
+    In JAX v0.4.35 or older, you must specify ``data_fields`` and ``meta_fields``
+    in order to use this decorator:
+
+    >>> import jax
     >>> from dataclasses import dataclass
     >>> from functools import partial
-    >>>
+    ...
     >>> @partial(jax.tree_util.register_dataclass,
     ...          data_fields=['x', 'y'],
     ...          meta_fields=['op'])
@@ -1158,7 +959,26 @@ def register_dataclass(
     >>> m
     MyStruct(x=Array([1., 1., 1.], dtype=float32), y=Array([0, 1, 2], dtype=int32), op='add')
 
-    Now that this class is registered, it can be used with functions in :mod:`jax.tree_util`:
+    Starting in JAX v0.4.36, the ``data_fields`` and ``meta_fields`` arguments are optional
+    for :func:`~dataclasses.dataclass` inputs, with fields defaulting to ``data_fields``
+    unless marked as static using `static` metadata in :func:`dataclasses.field`.
+
+    >>> import jax
+    >>> from dataclasses import dataclass, field
+    ...
+    >>> @jax.tree_util.register_dataclass
+    ... @dataclass
+    ... class MyStruct:
+    ...   x: jax.Array  # defaults to non-static data field
+    ...   y: jax.Array  # defaults to non-static data field
+    ...   op: str = field(metadata=dict(static=True))  # marked as static meta field.
+    ...
+    >>> m = MyStruct(x=jnp.ones(3), y=jnp.arange(3), op='add')
+    >>> m
+    MyStruct(x=Array([1., 1., 1.], dtype=float32), y=Array([0, 1, 2], dtype=int32), op='add')
+
+    Once this class is registered, it can be used with functions in :mod:`jax.tree` and
+    :mod:`jax.tree_util`:
 
     >>> leaves, treedef = jax.tree.flatten(m)
     >>> leaves
@@ -1169,7 +989,8 @@ def register_dataclass(
     MyStruct(x=Array([1., 1., 1.], dtype=float32), y=Array([0, 1, 2], dtype=int32), op='add')
 
     In particular, this registration allows ``m`` to be passed seamlessly through code
-    wrapped in :func:`jax.jit` and other JAX transformations:
+    wrapped in :func:`jax.jit` and other JAX transformations, with ``data_fields`` being
+    treated as dynamic arguments, and ``meta_fields`` being treated as static arguments:
 
     >>> @jax.jit
     ... def compiled_func(m):
@@ -1181,12 +1002,44 @@ def register_dataclass(
     >>> compiled_func(m)
     Array([1., 2., 3.], dtype=float32)
   """
+  if data_fields is None or meta_fields is None:
+    if (data_fields is None) != (meta_fields is None):
+      raise TypeError("register_dataclass: data_fields and meta_fields must both be specified"
+                      f" when either is specified. Got {data_fields=} {meta_fields=}.")
+    if not dataclasses.is_dataclass(nodetype):
+      raise TypeError("register_dataclass: data_fields and meta_fields are required when"
+                      f" nodetype is not a dataclass. Got {nodetype=}.")
+    data_fields = [f.name for f in dataclasses.fields(nodetype)
+                   if not f.metadata.get('static', False)]
+    meta_fields = [f.name for f in dataclasses.fields(nodetype)
+                   if f.metadata.get('static', False)]
+
+  assert meta_fields is not None
+  assert data_fields is not None
+
   # Store inputs as immutable tuples in this scope, because we close over them
   # for later evaluation. This prevents potentially confusing behavior if the
   # caller were to pass in lists that are later mutated.
   meta_fields = tuple(meta_fields)
   data_fields = tuple(data_fields)
 
+  if dataclasses.is_dataclass(nodetype):
+    init_fields = {f.name for f in dataclasses.fields(nodetype) if f.init}
+    init_fields.difference_update(*drop_fields)
+    if {*meta_fields, *data_fields} != init_fields:
+      msg = (
+          "data_fields and meta_fields must include all dataclass fields with"
+          " ``init=True`` and only them."
+      )
+      if missing := init_fields - {*meta_fields, *data_fields}:
+        msg += (
+            f" Missing fields: {missing}. Add them to drop_fields to suppress"
+            " this error."
+        )
+      if unexpected := {*meta_fields, *data_fields} - init_fields:
+        msg += f" Unexpected fields: {unexpected}."
+      raise ValueError(msg)
+
   def flatten_with_keys(x):
     meta = tuple(getattr(x, name) for name in meta_fields)
     data = tuple((GetAttrKey(name), getattr(x, name)) for name in data_fields)
@@ -1213,6 +1066,23 @@ def register_dataclass(
   return nodetype
 
 
+register_pytree_with_keys(
+    collections.OrderedDict,
+    lambda x: (tuple((DictKey(k), x[k]) for k in x.keys()), tuple(x.keys())),
+    lambda keys, values: collections.OrderedDict(safe_zip(keys, values)),
+)
+
+def _flatten_defaultdict_with_keys(d):
+  keys = tuple(sorted(d))
+  return tuple((DictKey(k), d[k]) for k in keys), (d.default_factory, keys)
+
+register_pytree_with_keys(
+    collections.defaultdict,
+    _flatten_defaultdict_with_keys,
+    lambda s, values: collections.defaultdict(s[0], safe_zip(s[1], values)),
+)
+
+
 @export
 def register_static(cls: type[H]) -> type[H]:
   """Registers `cls` as a pytree with no leaves.
@@ -1255,110 +1125,31 @@ def register_static(cls: type[H]) -> type[H]:
 def tree_flatten_with_path(
     tree: Any, is_leaf: Callable[[Any], bool] | None = None
 ) -> tuple[list[tuple[KeyPath, Any]], PyTreeDef]:
-  """Flattens a pytree like ``tree_flatten``, but also returns each leaf's key path.
-
-  Args:
-    tree: a pytree to flatten. If it contains a custom type, it must be
-      registered with ``register_pytree_with_keys``.
-  Returns:
-    A pair which the first element is a list of key-leaf pairs, each of
-    which contains a leaf and its key path. The second element is a treedef
-    representing the structure of the flattened tree.
-  """
-  _, tree_def = tree_flatten(tree, is_leaf)
-  return _generate_key_paths(tree, is_leaf), tree_def
+  """Alias of :func:`jax.tree.flatten_with_path`."""
+  return default_registry.flatten_with_path(tree, is_leaf)
 
 
 @export
 def tree_leaves_with_path(
     tree: Any, is_leaf: Callable[[Any], bool] | None = None
 ) -> list[tuple[KeyPath, Any]]:
-  """Gets the leaves of a pytree like ``tree_leaves`` and returns each leaf's key path.
-
-  Args:
-    tree: a pytree. If it contains a custom type, it must be registered with
-      ``register_pytree_with_keys``.
-  Returns:
-    A list of key-leaf pairs, each of which contains a leaf and its key path.
-
-  See Also:
-    - :func:`jax.tree_util.tree_leaves`
-    - :func:`jax.tree_util.tree_flatten_with_path`
-  """
-  return _generate_key_paths(tree, is_leaf)
+  """Alias of :func:`jax.tree.leaves_with_path`."""
+  return tree_flatten_with_path(tree, is_leaf)[0]
 
 
 # generate_key_paths is not exported.
 def generate_key_paths(
     tree: Any, is_leaf: Callable[[Any], bool] | None = None
 ) -> list[tuple[KeyPath, Any]]:
-  return list(_generate_key_paths_((), tree, is_leaf))
+  return tree_leaves_with_path(tree, is_leaf)
 _generate_key_paths = generate_key_paths  # alias for backward compat
 
 
-# The overall logic should be same as PyTreeDef::FlattenIntoImpl
-def _generate_key_paths_(
-    key_path: KeyPath,
-    tree: Any,
-    is_leaf: Callable[[Any], bool] | None = None,
-) -> Iterable[tuple[KeyPath, Any]]:
-  if is_leaf and is_leaf(tree):
-    yield key_path, tree
-    return
-  key_handler = _registry_with_keypaths.get(type(tree))
-  if key_handler:
-    key_children, _ = key_handler.flatten_with_keys(tree)
-    for k, c in key_children:
-      yield from _generate_key_paths_((*key_path, k), c, is_leaf)
-    return
-
-  flat = default_registry.flatten_one_level(tree)
-  if flat is None:
-    yield key_path, tree  # strict leaf type
-    return
-
-  if (isinstance(tree, tuple) and hasattr(tree, '_fields') and
-      flat[1] == type(tree)):
-    # handle namedtuple as a special case, based on heuristic
-    key_children = [(GetAttrKey(s), getattr(tree, s)) for s in tree._fields]
-    for k, c in key_children:
-      yield from _generate_key_paths_((*key_path, k), c, is_leaf)
-    return
-
-  for i, c in enumerate(flat[0]):
-    k = FlattenedIndexKey(i)
-    yield from _generate_key_paths_((*key_path, k), c, is_leaf)
-
-
 @export
 def tree_map_with_path(f: Callable[..., Any],
                        tree: Any, *rest: Any,
                        is_leaf: Callable[[Any], bool] | None = None) -> Any:
-  """Maps a multi-input function over pytree key path and args to produce a new pytree.
-
-  This is a more powerful alternative of ``tree_map`` that can take the key path
-  of each leaf as input argument as well.
-
-  Args:
-    f: function that takes ``2 + len(rest)`` arguments, aka. the key path and
-      each corresponding leaves of the pytrees.
-    tree: a pytree to be mapped over, with each leaf's key path as the first
-      positional argument and the leaf itself as the second argument to ``f``.
-    *rest: a tuple of pytrees, each of which has the same structure as ``tree``
-      or has ``tree`` as a prefix.
-
-  Returns:
-    A new pytree with the same structure as ``tree`` but with the value at each
-    leaf given by ``f(kp, x, *xs)`` where ``kp`` is the key path of the leaf at
-    the corresponding leaf in ``tree``, ``x`` is the leaf value and ``xs`` is
-    the tuple of values at corresponding nodes in ``rest``.
-
-  See Also:
-    - :func:`jax.tree_util.tree_map`
-    - :func:`jax.tree_util.tree_flatten_with_path`
-    - :func:`jax.tree_util.tree_leaves_with_path`
-  """
-
+  """Alias of :func:`jax.tree.map_with_path`."""
   keypath_leaves, treedef = tree_flatten_with_path(tree, is_leaf)
   keypath_leaves = list(zip(*keypath_leaves))
   all_keypath_leaves = keypath_leaves + [treedef.flatten_up_to(r) for r in rest]
@@ -1367,15 +1158,7 @@ def tree_map_with_path(f: Callable[..., Any],
 
 def _child_keys(pytree: Any) -> KeyPath:
   assert not treedef_is_strict_leaf(tree_structure(pytree))
-  handler = _registry_with_keypaths.get(type(pytree))
-  if handler:
-    return tuple(k for k, _ in handler.flatten_with_keys(pytree)[0])
-  elif isinstance(pytree, tuple) and hasattr(pytree, '_fields'):
-    # handle namedtuple as a special case, based on heuristic
-    return tuple(GetAttrKey(s) for s in pytree._fields)
-  else:
-    num_children = len(treedef_children(tree_structure(pytree)))
-    return tuple(FlattenedIndexKey(i) for i in range(num_children))
+  return tuple(k for k, _ in flatten_one_level_with_keys(pytree)[0])
 
 
 def _prefix_error(
@@ -1392,11 +1175,11 @@ def _prefix_error(
   if type(prefix_tree) != type(full_tree):
     yield lambda name: ValueError(
       "pytree structure error: different types at key path\n"
-      f"    {{name}}{keystr(key_path)}\n"
-      f"At that key path, the prefix pytree {{name}} has a subtree of type\n"
+      f"    {name}{keystr(key_path)}\n"
+      f"At that key path, the prefix pytree {name} has a subtree of type\n"
       f"    {type(prefix_tree)}\n"
       f"but at the same key path the full pytree has a subtree of different type\n"
-      f"    {type(full_tree)}.".format(name=name))
+      f"    {type(full_tree)}.")
     return  # don't look for more errors in this subtree
 
   # Or they may disagree if their roots have different numbers or keys of
@@ -1416,11 +1199,10 @@ def _prefix_error(
       ty = type(prefix_tree)
       yield lambda name: ValueError(
           f"pytree structure error: different lengths of {ty.__name__} at key path\n"
-          f"    {{name}}{keystr(key_path)}\n"
-          f"At that key path, the prefix pytree {{name}} has a subtree of type "
+          f"    {name}{keystr(key_path)}\n"
+          f"At that key path, the prefix pytree {name} has a subtree of type "
           f"{ty.__name__} of length {len(prefix_tree)}, but the full pytree "
-          f"has a subtree of the same type but of length {len(full_tree)}."
-          .format(name=name))
+          f"has a subtree of the same type but of length {len(full_tree)}.")
       return  # don't look for more errors in this subtree
   else:
     # Next we handle the general case of checking child keys.
@@ -1431,15 +1213,14 @@ def _prefix_error(
     if len(prefix_tree_children) != len(full_tree_children):
       yield lambda name: ValueError(
         "pytree structure error: different numbers of pytree children at key path\n"
-        f"    {{name}}{keystr(key_path)}\n"
-        f"At that key path, the prefix pytree {{name}} has a subtree of type\n"
+        f"    {name}{keystr(key_path)}\n"
+        f"At that key path, the prefix pytree {name} has a subtree of type\n"
         f"    {type(prefix_tree)}\n"
         f"with {len(prefix_tree_children)} child keys\n"
         f"    {' '.join(str(k.key) for k in prefix_tree_keys)}\n"
         f"but at the same key path the full pytree has a subtree of the same "
         f"type but with {len(full_tree_children)} child keys\n"
         f"    {' '.join(str(k.key) for k in full_tree_keys)}\n"
-        .format(name=name)
         + ("" if diff is None else
            f"so the symmetric difference on key sets is\n"
            f"    {' '.join(str(k.key) for k in diff)}"))
@@ -1456,8 +1237,8 @@ def _prefix_error(
         prefix="    ")
     yield lambda name: ValueError(
       "pytree structure error: different pytree metadata at key path\n"
-      f"    {{name}}{keystr(key_path)}\n"
-      f"At that key path, the prefix pytree {{name}} has a subtree of type\n"
+      f"    {name}{keystr(key_path)}\n"
+      f"At that key path, the prefix pytree {name} has a subtree of type\n"
       f"    {type(prefix_tree)}\n"
       f"with metadata\n"
       f"    {prefix_tree_meta_str}\n"
@@ -1465,7 +1246,7 @@ def _prefix_error(
       f"type but with metadata\n"
       f"    {full_tree_meta_str}\n"
       f"so the diff in the metadata at these pytree nodes is\n"
-      f"{metadata_diff}".format(name=name))
+      f"{metadata_diff}")
     return  # don't look for more errors in this subtree
 
   # If the root types and numbers of children agree, there must be an error
