@@ -14,10 +14,8 @@
 """Tests for serialization and deserialization of GDA."""
 
 import asyncio
-import contextlib
 import math
 from functools import partial
-import re
 import os
 import pathlib
 import tracemalloc as tm
@@ -26,10 +24,10 @@ from absl.testing import absltest
 from absl.testing import parameterized
 import jax
 import jax.numpy as jnp
+from jax._src import config
 from jax._src import test_util as jtu
 from jax._src import array
-from jax._src import xla_bridge as xb
-from jax.sharding import NamedSharding, GSPMDSharding
+from jax.sharding import NamedSharding, GSPMDSharding, SingleDeviceSharding
 from jax.sharding import PartitionSpec as P
 from jax.experimental.array_serialization import serialization
 from jax.experimental.layout import Layout, DeviceLocalLayout as DLL
@@ -37,20 +35,7 @@ import numpy as np
 import tensorstore as ts
 
 jax.config.parse_flags_with_absl()
-_exit_stack = contextlib.ExitStack()
-
-def setUpModule():
-  _exit_stack.enter_context(jtu.set_host_platform_device_count(8))
-
-def tearDownModule():
-  _exit_stack.close()
-
-
-pattern = re.compile(r"\{(.*?):")
-
-def extract_minor_to_major(l):
-  match = re.search(pattern, str(l))
-  return tuple(int(i) for i in match.groups()[0].split(','))
+jtu.request_cpu_devices(8)
 
 
 class CheckpointTest(jtu.JaxTestCase):
@@ -58,9 +43,43 @@ class CheckpointTest(jtu.JaxTestCase):
   def _on_commit_callback(self, temp_ckpt_dir, final_ckpt_dir):
     os.rename(temp_ckpt_dir, final_ckpt_dir)
 
+  def test_deserialize_on_array_list(self):
+    global_mesh = jtu.create_mesh((2, 4), ('x', 'y'))
+    inp_shape = (16, 64)
+    pspec = P('x', 'y')
+    sharding = NamedSharding(global_mesh, pspec)
+    inputs = []
+    lambda_fn = lambda idx: src[idx]
+    num_arrays = 5
+    for _ in range(num_arrays):
+      src = jax.random.normal(jax.random.key(0), inp_shape)
+      inp = array.make_array_from_callback(inp_shape, sharding, lambda_fn)
+      inputs.append(inp)
+    ckpt_dir = pathlib.Path(self.create_tempdir().full_path)
+    tspecs = [
+        serialization.get_tensorstore_spec(f'{ckpt_dir}/array_{i}')
+        for i in range(num_arrays)
+    ]
+    inputs = tuple(inputs)
+    tspecs = tuple(tspecs)
+    manager = serialization.GlobalAsyncCheckpointManager()
+    manager.serialize(
+        inputs,
+        tspecs,
+        on_commit_callback=partial(
+            self._on_commit_callback, ckpt_dir, ckpt_dir
+        ),
+    )
+    manager.wait_until_finished()
+    shardings = tuple([sharding] * num_arrays)
+    restored_arrays = manager.deserialize(shardings, tspecs)
+    self.assertLen(restored_arrays, num_arrays)
+    for inp, deserialized_array in zip(inputs, restored_arrays):
+      self.assertArraysEqual(deserialized_array, inp)
+
   @jtu.skip_on_devices('cpu')
   def test_memory_consumption(self):
-    global_mesh = jtu.create_global_mesh((2, 4), ('x', 'y'))
+    global_mesh = jtu.create_mesh((2, 4), ('x', 'y'))
     inp_shape = (2_048, 4_096)
     pspec = P('x', 'y')
     num = math.prod(inp_shape)
@@ -104,8 +123,43 @@ class CheckpointTest(jtu.JaxTestCase):
     self.assertGreater(peak, 30_000_000)
     tm.stop()
 
+  def test_memory_consumption_for_save(self):
+    global_mesh = jtu.create_mesh((1, 1), ('x', 'y'))
+    inp_shape = (16 * 1024, 16 * 1024)
+    pspec = P('x', 'y')
+    num = math.prod(inp_shape)
+    sharding = NamedSharding(global_mesh, pspec)
+    src = jnp.arange(num, dtype=np.int32).reshape(inp_shape)
+    inp = array.make_array_from_callback(
+        inp_shape, sharding, lambda idx: src[idx]
+    )
+    ckpt_dir = pathlib.Path(self.create_tempdir('memprofsave').full_path)
+    tspec = serialization.get_tensorstore_spec(str(ckpt_dir))
+    tspec['metadata'] = {
+        'shape': inp.shape,
+        'compressor': None,
+        'chunks': inp.shape,
+    }
+
+    is_cpu = jtu.test_device_matches(['cpu'])
+    tm.start()
+    try:
+      manager = serialization.GlobalAsyncCheckpointManager()
+      manager.serialize(
+          [inp],
+          [tspec],
+          on_commit_callback=partial(
+              self._on_commit_callback, ckpt_dir, ckpt_dir
+          ),
+      )
+      manager.wait_until_finished()
+      unused_current, peak = tm.get_traced_memory()
+      self.assertLess(peak, src.nbytes * (1 * (not is_cpu) + 0.5))
+    finally:
+      tm.stop()
+
   def test_checkpointing_with_path_variant(self):
-    global_mesh = jtu.create_global_mesh((4, 2), ('x', 'y'))
+    global_mesh = jtu.create_mesh((4, 2), ('x', 'y'))
     inp_shape = (8, 2)
     pspec = P('x', 'y')
     num = math.prod(inp_shape)
@@ -137,7 +191,7 @@ class CheckpointTest(jtu.JaxTestCase):
     self.assertEqual(m1.dtype, np.int32)
 
   def test_checkpointing_jax_array(self):
-    global_mesh = jtu.create_global_mesh((4, 2), ('x', 'y'))
+    global_mesh = jtu.create_mesh((4, 2), ('x', 'y'))
     inp_shape = (8, 2)
     pspec = P('x', 'y')
     num = math.prod(inp_shape)
@@ -161,7 +215,7 @@ class CheckpointTest(jtu.JaxTestCase):
     # Third Array
     def cb3(_):
       return np.array([], dtype=np.float32)
-    global_mesh1d = jtu.create_global_mesh((8,), ('x',))
+    global_mesh1d = jtu.create_mesh((8,), ('x',))
     a3 = array.make_array_from_callback(
         (0,), NamedSharding(global_mesh1d, P(None)), cb3)
     ckpt_path3 = pathlib.Path(self.create_tempdir(f'{ckpt_dir}/third').full_path)
@@ -204,9 +258,103 @@ class CheckpointTest(jtu.JaxTestCase):
       self.assertArraysEqual(np.asarray(s.data), np.array([], dtype=np.float32))
     self.assertEqual(m3.dtype, np.float32)
 
+  def test_checkpointing_ocdbt_transaction(self):
+    global_mesh = jtu.create_mesh((4, 2), ('x', 'y'))
+    inp_shape = (8, 2)
+    pspec = P('x', 'y')
+    num = math.prod(inp_shape)
+
+    # First Array
+    global_input_data1 = np.arange(num, dtype=np.int32).reshape(inp_shape)
+    a1 = array.make_array_from_callback(
+        inp_shape,
+        NamedSharding(global_mesh, pspec),
+        lambda idx: global_input_data1[idx],
+    )
+    ckpt_dir = pathlib.Path(self.create_tempdir('ckpt').full_path)
+    ckpt_path1 = ckpt_dir / 'first'
+
+    # Second Array
+    global_input_data2 = np.arange(num, num + num, dtype=np.int32).reshape(
+        inp_shape
+    )
+    a2 = array.make_array_from_callback(
+        inp_shape,
+        NamedSharding(global_mesh, pspec),
+        lambda idx: global_input_data2[idx],
+    )
+    ckpt_path2 = ckpt_dir / 'second'
+
+    # Third Array
+    def cb3(_):
+      return np.array([], dtype=np.float32)
+
+    global_mesh1d = jtu.create_mesh((8,), ('x',))
+    a3 = array.make_array_from_callback(
+        (0,), NamedSharding(global_mesh1d, P(None)), cb3
+    )
+    ckpt_path3 = ckpt_dir / 'third'
+
+    ckpt_paths = [str(ckpt_path1), str(ckpt_path2), str(ckpt_path3)]
+    tspecs = jax.tree_util.tree_map(
+        lambda p: serialization.get_tensorstore_spec(p, ocdbt=True), ckpt_paths
+    )
+
+    manager = serialization.GlobalAsyncCheckpointManager()
+    with ts.Transaction(atomic=True) as transaction:
+      manager.serialize(
+          [a1, a2, a3],
+          tspecs,
+          on_commit_callback=partial(
+              self._on_commit_callback, ckpt_dir, ckpt_dir
+          ),
+          transaction=transaction,
+      )
+    manager.wait_until_finished()
+
+    m1, m2, m3 = serialization.run_deserialization(
+        [
+            NamedSharding(global_mesh, pspec),
+            NamedSharding(global_mesh, P('x')),
+            NamedSharding(global_mesh1d, P(None)),
+        ],
+        tspecs,
+    )
+
+    self.assertIsInstance(m1, array.ArrayImpl)
+    self.assertArraysEqual(
+        np.asarray(m1.addressable_shards[0].data),
+        np.array([[0], [2]], dtype=np.int32),
+    )
+    self.assertArraysEqual(
+        np.asarray(m1.addressable_shards[1].data),
+        np.array([[1], [3]], dtype=np.int32),
+    )
+    self.assertEqual(m1.addressable_shards[0].data.shape, (2, 1))
+    self.assertEqual(m1.dtype, np.int32)
+
+    self.assertIsInstance(m2, array.ArrayImpl)
+    self.assertArraysEqual(
+        np.asarray(m2.addressable_shards[0].data),
+        np.array([[16, 17], [18, 19]], dtype=np.int32),
+    )
+    self.assertArraysEqual(
+        np.asarray(m2.addressable_shards[1].data),
+        np.array([[16, 17], [18, 19]], dtype=np.int32),
+    )
+    self.assertEqual(m2.addressable_shards[0].data.shape, (2, 2))
+    self.assertEqual(m2.dtype, np.int32)
+
+    self.assertIsInstance(m3, array.ArrayImpl)
+    for i, s in enumerate(m3.addressable_shards):
+      self.assertEqual(s.index, (slice(None),))
+      self.assertEqual(s.replica_id, i)
+      self.assertArraysEqual(np.asarray(s.data), np.array([], dtype=np.float32))
+    self.assertEqual(m3.dtype, np.float32)
+
   @parameterized.product(input_dtype=[np.int32, jnp.bfloat16])
   def test_checkpointing_with_bigger_shape_jax_array(self, input_dtype):
-    global_mesh = jtu.create_global_mesh((2, 2), ('x', 'y'))
+    global_mesh = jtu.create_mesh((2, 2), ('x', 'y'), iota_order=True)
     global_input_shape = (8, 2)
     num = math.prod(global_input_shape)
 
@@ -228,7 +376,8 @@ class CheckpointTest(jtu.JaxTestCase):
         on_commit_callback=partial(self._on_commit_callback, ckpt_dir, ckpt_dir))
     manager.wait_until_finished()
 
-    ds = NamedSharding(jtu.create_global_mesh((4, 2), ('x', 'y')), P('x', 'y'))
+    ds = NamedSharding(jtu.create_mesh((4, 2), ('x', 'y'), iota_order=True),
+                       P('x', 'y'))
 
     m1, = serialization.run_deserialization([ds], tspecs, [(12, 2)],
                                             [np.float32])
@@ -254,7 +403,9 @@ class CheckpointTest(jtu.JaxTestCase):
 
   @parameterized.product(input_dtype=[jnp.int4, jnp.int8])
   def test_checkpointing_with_int4(self, input_dtype):
-    global_mesh = jtu.create_global_mesh((2, 2), ('x', 'y'))
+    if config.use_shardy_partitioner.value:
+      self.skipTest("TODO(b/376077396): Fix XlaRuntimeError: INVALID_ARGUMENT")
+    global_mesh = jtu.create_mesh((2, 2), ('x', 'y'), iota_order=True)
     global_input_shape = (8, 2)
     num = math.prod(global_input_shape)
 
@@ -276,7 +427,8 @@ class CheckpointTest(jtu.JaxTestCase):
         on_commit_callback=partial(self._on_commit_callback, ckpt_dir, ckpt_dir))
     manager.wait_until_finished()
 
-    ds = NamedSharding(jtu.create_global_mesh((4, 2), ('x', 'y')), P('x', 'y'))
+    ds = NamedSharding(jtu.create_mesh((4, 2), ('x', 'y'), iota_order=True),
+                       P('x', 'y'))
 
     target_dtype = jnp.dtype('int4')
     m1, = serialization.run_deserialization([ds], tspecs, [(12, 2)],
@@ -303,7 +455,7 @@ class CheckpointTest(jtu.JaxTestCase):
       self.assertArraysEqual(l.data, global_input_data.astype(target_dtype))
 
   def test_checkpointing_scalar_jax_array(self):
-    global_mesh = jtu.create_global_mesh((2,), ('x'))
+    global_mesh = jtu.create_mesh((2,), ('x'))
     global_input_shape = ()
     data = np.array(4)
     s = NamedSharding(global_mesh, P(None))
@@ -320,7 +472,7 @@ class CheckpointTest(jtu.JaxTestCase):
         on_commit_callback=partial(self._on_commit_callback, ckpt_dir, ckpt_dir))
     manager.wait_until_finished()
 
-    ds = NamedSharding(jtu.create_global_mesh((2,), ('x')), P(None))
+    ds = NamedSharding(jtu.create_mesh((2,), ('x')), P(None))
 
     m1, = serialization.run_deserialization(
         [ds],
@@ -333,7 +485,7 @@ class CheckpointTest(jtu.JaxTestCase):
       self.assertArraysEqual(np.asarray(l.data), data.astype(np.float32))
 
   def test_deserialize_tensorstore_array_jax_array(self):
-    global_mesh = jtu.create_global_mesh((2,), ('x'))
+    global_mesh = jtu.create_mesh((2,), ('x'))
     data = np.arange(1024)
     tspec = ts.array(data).spec()
     m1, = serialization.run_deserialization(
@@ -429,15 +581,15 @@ class CheckpointTest(jtu.JaxTestCase):
     if not jtu.test_device_matches(['tpu']):
       self.skipTest('Layouts are only supported on TPUs')
 
-    mesh = jtu.create_global_mesh((4, 2), ('x', 'y'))
+    mesh = jtu.create_mesh((4, 2), ('x', 'y'))
     np_inp = np.arange(32).reshape(8, 4)
     s = NamedSharding(mesh, P('x', 'y'))
     arr = jax.device_put(np_inp, s)
 
     out_layout = jax.jit(lambda x: x.T, out_shardings=Layout(DLL.AUTO)).lower(
-        arr).compile().output_layouts()
-    self.assertEqual(extract_minor_to_major(arr.layout),
-                     extract_minor_to_major(out_layout)[::-1])
+        arr).compile().output_layouts
+    self.assertEqual(arr.layout.device_local_layout.major_to_minor,
+                     out_layout.device_local_layout.major_to_minor[::-1])
 
     ckpt_dir = pathlib.Path(self.create_tempdir('ckpt').full_path)
     ckpt_path = pathlib.Path(self.create_tempdir(f'{ckpt_dir}/first').full_path)
@@ -458,6 +610,10 @@ class CheckpointTest(jtu.JaxTestCase):
       self.assertArraysEqual(s.data, np_inp[s.index])
 
   def test_deserialization_with_int4(self):
+    if config.use_shardy_partitioner.value:
+      self.skipTest("TODO(b/376077396): Fix XlaRuntimeError: INVALID_ARGUMENT")
+    if jtu.test_device_matches(['gpu']):
+      self.skipTest("Fails on GPU. Enable after it's fixed")
     dtype = jnp.int4
     shape = (8, 2)
     arr = jnp.arange(np.prod(shape)).reshape(shape).astype(dtype)
@@ -489,5 +645,19 @@ class CheckpointTest(jtu.JaxTestCase):
     self.assertEqual(out.dtype, jnp.int8)
     self.assertArraysEqual(out + out, out * 2)
 
+
+class TransferShardTest(jtu.JaxTestCase):
+
+  @jtu.skip_on_devices('cpu')
+  def test_transfer_shard_to_host(self):
+    np_inp = np.arange(16).reshape((4, 4))
+    sharding = SingleDeviceSharding(jax.devices()[0], memory_kind="device")
+    arr = jax.device_put(np_inp, sharding)
+    shard = arr.addressable_shards[0]
+
+    np_out = asyncio.run(serialization.transfer_shard_to_host(shard))
+
+    self.assertArraysEqual(np_out, np_inp)
+
 if __name__ == '__main__':
   absltest.main(testLoader=jtu.JaxTestLoader())
