@@ -45,14 +45,15 @@ r"""Jet is an experimental module for higher-order automatic differentiation
   and can thus be used for high-order
   automatic differentiation of :math:`f`.
   Details are explained in
-  `these notes <https://github.com/google/jax/files/6717197/jet.pdf>`__.
+  `these notes <https://github.com/jax-ml/jax/files/6717197/jet.pdf>`__.
 
   Note:
     Help improve :func:`jet` by contributing
-    `outstanding primitive rules <https://github.com/google/jax/issues/2431>`__.
+    `outstanding primitive rules <https://github.com/jax-ml/jax/issues/2431>`__.
 """
 
-from typing import Any, Callable
+from collections.abc import Callable
+from typing import Any
 
 from functools import partial
 
@@ -69,10 +70,9 @@ from jax._src import core
 from jax._src import dispatch
 from jax._src import linear_util as lu
 from jax._src import sharding_impls
-from jax._src.api_util import shaped_abstractify
 from jax._src.interpreters import partial_eval as pe
 from jax._src.lax import lax as lax_internal
-from jax._src.util import unzip2, weakref_lru_cache
+from jax._src.util import unzip2, weakref_lru_cache, safe_zip
 
 
 def jet(fun, primals, series):
@@ -140,40 +140,43 @@ def jet(fun, primals, series):
       if not treedef_is_leaf(treedef):
         raise ValueError(f"term {j} for argument {i} is not an array")
 
-  @lu.transformation_with_aux
-  def flatten_fun_output(*args):
-    ans = yield args, {}
-    yield tree_flatten(ans)
+  @lu.transformation_with_aux2
+  def flatten_fun_output(f, store, *args):
+    ans = f(*args)
+    ans, tree = tree_flatten(ans)
+    store.store(tree)
+    return ans
 
   f, out_tree = flatten_fun_output(lu.wrap_init(fun))
   out_primals, out_terms = jet_fun(jet_subtrace(f), order).call_wrapped(primals, series)
   return tree_unflatten(out_tree(), out_primals), tree_unflatten(out_tree(), out_terms)
 
-@lu.transformation
-def jet_fun(order, primals, series):
-  with core.new_main(JetTrace) as main:
-    main.order = order
-    out_primals, out_terms = yield (main, primals, series), {}
-    del main
+@lu.transformation2
+def jet_fun(f, order, primals, series):
+  tag = core.TraceTag()
+  out_primals, out_terms = f(tag, order, primals, series)
   out_terms = [[jnp.zeros_like(p)] * order if s is zero_series else s
                for p, s in zip(out_primals, out_terms)]
-  yield out_primals, out_terms
+  return out_primals, out_terms
 
-@lu.transformation
-def jet_subtrace(main, primals, series):
-  trace = JetTrace(main, core.cur_sublevel())
-  in_tracers = map(partial(JetTracer, trace), primals, series)
-  ans = yield in_tracers, {}
-  out_tracers = map(trace.full_raise, ans)
-  out_primals, out_terms = unzip2((t.primal, t.terms) for t in out_tracers)
-  yield out_primals, out_terms
+@lu.transformation2
+def jet_subtrace(f, tag, order, primals, series):
+  with core.take_current_trace() as parent_trace:
+    trace = JetTrace(tag, parent_trace, order)
+    in_tracers = map(partial(JetTracer, trace), primals, series)
+    with core.set_current_trace(trace):
+       ans = f(*in_tracers)
 
-@lu.transformation_with_aux
-def traceable(in_tree_def, *primals_and_series):
+    out_primals, out_terms = unzip2(map(trace.to_primal_terms_pair, ans))
+    return out_primals, out_terms
+
+@lu.transformation_with_aux2
+def traceable(f, store, in_tree_def, *primals_and_series):
   primals_in, series_in = tree_unflatten(in_tree_def, primals_and_series)
-  primals_out, series_out = yield (primals_in, series_in), {}
+  primals_out, series_out = f(primals_in, series_in)
   out_flat, out_tree_def = tree_flatten((primals_out, series_out))
-  yield out_flat, out_tree_def
+  store.store(out_tree_def)
+  return out_flat
 
 
 class JetTracer(core.Tracer):
@@ -197,33 +200,44 @@ class JetTracer(core.Tracer):
 
 class JetTrace(core.Trace):
 
-  def pure(self, val):
-    return JetTracer(self, val, zero_series)
+  def __init__(self, tag, parent_trace, order):
+    self.tag = tag
+    self.parent_trace = parent_trace
+    self.order = order
 
-  def lift(self, val):
-    return JetTracer(self, val, zero_series)
-
-  def sublift(self, val):
-    return JetTracer(self, val.primal, val.terms)
+  def to_primal_terms_pair(self, val):
+    if isinstance(val, JetTracer) and val._trace.tag is self.tag:
+      return val.primal, val.terms
+    else:
+      return val, zero_series
 
   def process_primitive(self, primitive, tracers, params):
-    order = self.main.order              # pytype: disable=attribute-error
-    primals_in, series_in = unzip2((t.primal, t.terms) for t in tracers)
+    order = self.order              # pytype: disable=attribute-error
+    primals_in, series_in = unzip2(map(self.to_primal_terms_pair, tracers))
+
+    if all(t is zero_series for t in series_in):
+      primal_out = primitive.bind_with_trace(self.parent_trace, primals_in, params)
+      if primitive.multiple_results:
+        return [JetTracer(self, p, zero_series) for p in primal_out]
+      else:
+        return JetTracer(self, primal_out, zero_series)
+
     series_in = [[zero_term] * order if s is zero_series else s
                  for s in series_in]
-    # TODO(mattjj): avoid always instantiating zeros
-    series_in = [[jnp.zeros(np.shape(x), dtype=jnp.result_type(x))
-                  if t is zero_term else t for t in series]
-                 for x, series in zip(primals_in, series_in)]
-    rule = jet_rules[primitive]
-    primal_out, terms_out = rule(primals_in, series_in, **params)
+    with core.set_current_trace(self.parent_trace):
+      # TODO(mattjj): avoid always instantiating zeros
+      series_in = [[jnp.zeros(np.shape(x), dtype=jnp.result_type(x))
+                    if t is zero_term else t for t in series]
+                   for x, series in zip(primals_in, series_in)]
+      rule = jet_rules[primitive]
+      primal_out, terms_out = rule(primals_in, series_in, **params)
     if not primitive.multiple_results:
       return JetTracer(self, primal_out, terms_out)
     else:
       return [JetTracer(self, p, ts) for p, ts in zip(primal_out, terms_out)]
 
   def process_call(self, call_primitive, f, tracers, params):
-    primals_in, series_in = unzip2((t.primal, t.terms) for t in tracers)
+    primals_in, series_in = unzip2(map(self.to_primal_terms_pair, tracers))
     primals_and_series, in_tree_def = tree_flatten((primals_in, series_in))
     f_jet, out_tree_def = traceable(jet_subtrace(f, self.main), in_tree_def)
     update_params = call_param_updaters.get(call_primitive)
@@ -233,17 +247,6 @@ class JetTrace(core.Trace):
     primals_out, series_out = tree_unflatten(out_tree_def(), result)
     return [JetTracer(self, p, ts) for p, ts in zip(primals_out, series_out)]
 
-  def post_process_call(self, call_primitive, out_tracers, params):
-    primals, series = unzip2((t.primal, t.terms) for t in out_tracers)
-    out, treedef = tree_flatten((primals, series))
-    del primals, series
-    main = self.main
-    def todo(x):
-      primals, series = tree_unflatten(treedef, x)
-      trace = JetTrace(main, core.cur_sublevel())
-      return map(partial(JetTracer, trace), primals, series)
-    return out, todo
-
   def process_custom_jvp_call(self, primitive, fun, jvp, tracers, *,
                               symbolic_zeros):
     # TODO(mattjj): don't just ignore custom jvp rules?
@@ -306,6 +309,8 @@ def deflinear(prim):
 def linear_prop(prim, primals_in, series_in, **params):
   primal_out = prim.bind(*primals_in, **params)
   series_out = [prim.bind(*terms_in, **params) for terms_in in zip(*series_in)]
+  if prim.multiple_results:
+    series_out = safe_zip(*series_out)
   return primal_out, series_out
 
 deflinear(lax.neg_p)
@@ -319,6 +324,7 @@ deflinear(lax.sub_p)
 deflinear(lax.convert_element_type_p)
 deflinear(lax.broadcast_in_dim_p)
 deflinear(lax.concatenate_p)
+deflinear(lax.split_p)
 deflinear(lax.pad_p)
 deflinear(lax.reshape_p)
 deflinear(lax.squeeze_p)
@@ -328,6 +334,7 @@ deflinear(lax.slice_p)
 deflinear(lax.reduce_sum_p)
 deflinear(lax.reduce_window_sum_p)
 deflinear(lax.fft_p)
+deflinear(lax.copy_p)
 deflinear(dispatch.device_put_p)
 
 def _dynamic_slice_jet_rule(primals_in, series_in, **params):
@@ -403,6 +410,7 @@ def def_comp(prim, comp):
 def_comp(lax.expm1_p, lambda x: lax.exp(x) - 1)
 def_comp(lax.log1p_p, lambda x: lax.log(1 + x))
 def_comp(lax.sqrt_p, lambda x: x ** 0.5)
+def_comp(lax.square_p, lambda x: x * x)
 def_comp(lax.rsqrt_p, lambda x: x ** -0.5)
 def_comp(lax.asinh_p, lambda x: lax.log(x + lax.sqrt(lax.square(x) + 1)))
 def_comp(lax.acosh_p, lambda x: lax.log(x + lax.sqrt(lax.square(x) - 1)))
@@ -724,7 +732,7 @@ def _jet_jaxpr(
 def _pjit_jet_rule(primals_in, series_in, **params):
   primals_and_series, in_tree_def = tree_flatten((primals_in, series_in))
   order = len(series_in[0])
-  primals_and_series_avals = tuple(shaped_abstractify(x) for x in primals_and_series)
+  primals_and_series_avals = tuple(core.shaped_abstractify(x) for x in primals_and_series)
   jaxpr_jet, out_tree_def = _jet_jaxpr(params['jaxpr'], order,
                                        primals_and_series_avals, in_tree_def)
   num_series_in = len(primals_in) * order
