@@ -16,8 +16,9 @@
 from __future__ import annotations
 
 import collections
+from collections.abc import Callable
 import functools
-from typing import Callable, Dict, List, NamedTuple, Set, Tuple
+from typing import NamedTuple
 from jax import util as jax_util
 from jax.experimental.pallas.ops.tpu.splash_attention import splash_attention_mask as mask_lib
 import numpy as np
@@ -55,7 +56,7 @@ class MaskInfo(NamedTuple):
       indicates that the corresponding block in the full mask contained both
       zeros and ones. An entry of 2 indicates the corresponding block was
       entirely ones.
-    partial_mask_blocks: A i32[num_partial_blocks, block_q, block_kv] NumPy
+    partial_mask_blocks: A bool[num_partial_blocks, block_q, block_kv] NumPy
       array that contains the blocks of the original mask that contained both
       zeros and ones. The entries in `mask_next` point to indices in the first
       axis of this array.
@@ -161,11 +162,11 @@ class _HashableNDArray:
 
 
 def _get_mask_info_for_shard(
-    output_shape: Tuple[int, int, int],
+    output_shape: tuple[int, int, int],
     has_mask_next: bool,
     mask: mask_lib.MultiHeadMask,
-    block_shape: Tuple[int, int],
-    coords_to_partial_mask_block_index: Dict[Tuple[int, int, int], int],
+    block_shape: tuple[int, int],
+    coords_to_partial_mask_block_index: dict[tuple[int, int, int], int],
     masks_per_head_shard: int,
     head_start: int,
     num_heads: int,
@@ -173,7 +174,7 @@ def _get_mask_info_for_shard(
     q_seq_shard_size: int,
     blocked_q_seq_start: int,
     is_dkv: bool,
-) -> Tuple[np.ndarray, np.ndarray | None]:
+) -> tuple[np.ndarray, np.ndarray | None]:
   """Process a slice of the mask to compute data_next and mask_next.
 
   Args:
@@ -304,13 +305,13 @@ def _get_mask_info_for_shard(
 
 
 # When used in a transformer network with multiple layers, the SplashAttention
-# kernel is created serveral times with the same mask. Cache MaskInfo to avoid
+# kernel is created several times with the same mask. Cache MaskInfo to avoid
 # blowing up compile times. Ideally the size of the cache should be determined
 # by the client.
 @functools.lru_cache(maxsize=12)
 def _process_mask(
     mask: mask_lib.MultiHeadMask,  # [num_heads, q_seq_len, kv_seq_len]
-    block_shape: Tuple[int, int],
+    block_shape: tuple[int, int],
     is_dkv: bool,
     *,
     downcast_smem_data: bool = True,
@@ -375,14 +376,6 @@ def _process_mask(
   if mod != 0:
     raise ValueError(f'{head_shards=} should divide {head_count=}.')
 
-  first_mask_size = mask.masks[0].shape
-  for h in range(head_count):
-    if mask.masks[h].shape != first_mask_size:
-      raise ValueError(
-          f'First head mask has shape {first_mask_size}, but head mask {h} has'
-          f' shape {mask.masks[h].shape}. All head masks must have the same'
-          ' shape.'
-      )
 
   # Uniquify the masks.
   # Create a collection of the unique head masks in the input multi-head mask.
@@ -394,18 +387,18 @@ def _process_mask(
     id_map = collections.defaultdict(lambda: len(id_map))
     return {obj: id_map[obj] for obj in objects}
 
-  unique_masks_dict: Dict[mask_lib.Mask, int] = assign_unique_ids(
+  unique_masks_dict: dict[mask_lib.Mask, int] = assign_unique_ids(
       head_mask for head_mask in mask.masks
   )
 
   # Build a mapping of heads to unique masks and masks to unique masks.
 
-  head_to_mask_id: List[int] = [0] * head_count
-  head_shard_to_mask_ids: List[Set[int]] = [set() for _ in range(head_shards)]
-  mask_id_to_heads: List[List[int]] = [
+  head_to_mask_id: list[int] = [0] * head_count
+  head_shard_to_mask_ids: list[set[int]] = [set() for _ in range(head_shards)]
+  mask_id_to_heads: list[list[int]] = [
       [] for _ in range(len(unique_masks_dict))
   ]
-  mask_id_to_head_shards: List[Set[int]] = [
+  mask_id_to_head_shards: list[set[int]] = [
       set() for _ in range(len(unique_masks_dict))
   ]
 
@@ -436,10 +429,10 @@ def _process_mask(
   # TODO(amagni): checking the validity of the masks is slow for large masks.
   # Disable it for now, reevalute in the future.
 
-  partial_mask_block_ids: Dict[_HashableNDArray, int] = collections.defaultdict(
+  partial_mask_block_ids: dict[_HashableNDArray, int] = collections.defaultdict(
       lambda: len(partial_mask_block_ids)
   )
-  block_id_to_block_coords: Dict[int, List[Tuple[int, ...]]] = (
+  block_id_to_block_coords: dict[int, list[tuple[int, ...]]] = (
       collections.defaultdict(list)
   )
 
@@ -525,13 +518,9 @@ def _process_mask(
 
   partial_mask_blocks = None
   has_mask_next = False
-  if len(unique_partial_mask_blocks) == 1:
+  if len(unique_partial_mask_blocks) >= 1:
     partial_mask_blocks = [x.array for x in unique_partial_mask_blocks]
-    partial_mask_blocks = partial_mask_blocks[0][None].astype(np.int32)
-    has_mask_next = True
-  elif len(unique_partial_mask_blocks) > 1:
-    partial_mask_blocks = [x.array for x in unique_partial_mask_blocks]
-    partial_mask_blocks = np.stack(partial_mask_blocks, axis=0).astype(np.int32)
+    partial_mask_blocks = np.stack(partial_mask_blocks, axis=0).astype(np.bool_)
     has_mask_next = True
   if is_dkv and partial_mask_blocks is not None:
     partial_mask_blocks = np.swapaxes(partial_mask_blocks, -1, -2)
