@@ -47,9 +47,9 @@ Array([-1.2655463 , -0.52060574, -0.14522289, -0.10817424,
        -0.15574613], dtype=float32)
 """
 
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import functools
-from typing import Any, Callable, NamedTuple, Optional
+from typing import Any, NamedTuple
 
 import numpy as np
 
@@ -97,6 +97,7 @@ _zero_preserving_unary_primitives = [
   lax.sin_p,
   lax.sinh_p,
   lax.sqrt_p,
+  lax.square_p,
   lax.tan_p,
   lax.tanh_p,
   lax.convert_element_type_p,
@@ -276,16 +277,6 @@ def spvalues_to_avals(
 # ------------------------------------------------------------------------------
 # Implementation of sparsify() using tracers.
 
-def popattr(obj: Any, name: str) -> Any:
-  assert hasattr(obj, name)
-  val = getattr(obj, name)
-  delattr(obj, name)
-  return val
-
-def setnewattr(obj: Any, name: str, val: Any):
-  assert not hasattr(obj, name)
-  setattr(obj, name, val)
-
 class SparseTracer(core.Tracer):
   def __init__(self, trace: core.Trace, *, spvalue):
     self._spvalue = spvalue
@@ -293,9 +284,9 @@ class SparseTracer(core.Tracer):
 
   @property
   def spenv(self):
-    if not hasattr(self._trace.main, 'spenv'):
-      raise RuntimeError("Internal: main does not have spenv defined.")
-    return self._trace.main.spenv
+    if not hasattr(self._trace, 'spenv'):
+      raise RuntimeError("Internal: trace does not have spenv defined.")
+    return self._trace.spenv
 
   @property
   def aval(self):
@@ -305,71 +296,71 @@ class SparseTracer(core.Tracer):
     return self
 
 class SparseTrace(core.Trace):
-  def pure(self, val: Any):
-    if not hasattr(self.main, 'spenv'):
-      raise RuntimeError("Internal: main does not have spenv defined.")
-    spvalue, = arrays_to_spvalues(self.main.spenv, [val])
-    return SparseTracer(self, spvalue=spvalue)
 
-  def lift(self, val: core.Tracer):
-    if not hasattr(self.main, 'spenv'):
-      raise RuntimeError("Internal: main does not have spenv defined.")
-    spvalue, = arrays_to_spvalues(self.main.spenv, [val])
-    return SparseTracer(self, spvalue=spvalue)
+  def __init__(self, parent_trace, tag, spenv):
+    self.parent_trace = parent_trace
+    self.tag = tag
+    self.spenv = spenv
 
-  def sublift(self, val: SparseTracer):
-    return SparseTracer(val._trace, spvalue=val._spvalue)
+  def to_sparse_tracer(self, val):
+    if isinstance(val, SparseTracer) and self.tag is val._trace.tag:
+      return val
+    else:
+      with core.set_current_trace(self.parent_trace):
+        spvalue, = arrays_to_spvalues(self.spenv, [val])
+      return SparseTracer(self, spvalue=spvalue)
 
   def process_primitive(self, primitive, tracers, params):
-    spenv = popattr(self.main, 'spenv')
+    tracers = [self.to_sparse_tracer(t) for t in tracers]
     spvalues = [t._spvalue for t in tracers]
     if any(spvalue.is_sparse() for spvalue in spvalues):
       if primitive not in sparse_rules_bcoo:
         _raise_unimplemented_primitive(primitive)
-      out_spvalues = sparse_rules_bcoo[primitive](spenv, *(t._spvalue for t in tracers), **params)
+      with core.set_current_trace(self.parent_trace):
+        out_spvalues = sparse_rules_bcoo[primitive](self.spenv, *(t._spvalue for t in tracers), **params)
     else:
-      out_bufs = primitive.bind(*(spenv.data(spvalue) for spvalue in spvalues), **params)
-      out_spvalues = arrays_to_spvalues(spenv, out_bufs if primitive.multiple_results else [out_bufs])
-    setnewattr(self.main, 'spenv', spenv)
+      out_bufs = primitive.bind_with_trace(self.parent_trace, tuple(self.spenv.data(spvalue) for spvalue in spvalues), params)
+      out_spvalues = arrays_to_spvalues(self.spenv, out_bufs if primitive.multiple_results else [out_bufs])
     out_tracers = tuple(SparseTracer(self, spvalue=spvalue) for spvalue in out_spvalues)
     return out_tracers if primitive.multiple_results else out_tracers[0]
 
   def process_call(self, call_primitive, f: lu.WrappedFun, tracers, params):
-    spenv = popattr(self.main, 'spenv')
+    assert False
     spvalues = tuple(t._spvalue for t in tracers)
-    in_bufs = spenv._buffers
+    in_bufs = self.spenv._buffers
     fun, out_spvalues = sparsify_subtrace(f, self.main, spvalues)
     if any(params['donated_invars']):
       raise NotImplementedError("sparsify does not support donated_invars")
     params = dict(params, donated_invars=tuple(False for buf in in_bufs))
     bufs_out = call_primitive.bind(fun, *in_bufs, **params)
-    setnewattr(self.main, 'spenv', SparsifyEnv(bufs_out))
     return [SparseTracer(self, spvalue=spvalue) for spvalue in out_spvalues()]
 
   def process_custom_jvp_call(self, primitive, fun, jvp, tracers, *, symbolic_zeros):
     # TODO(jakevdp): handle the jvp here
     del primitive, jvp, symbolic_zeros
-    return fun.call_wrapped(*tracers)
+    with core.set_current_trace(self):
+      return fun.call_wrapped(*tracers)
 
-@lu.transformation_with_aux
-def sparsify_subtrace(main, spvalues, *bufs):
-  setnewattr(main, 'spenv', SparsifyEnv(bufs))
-  trace = main.with_cur_sublevel()
-  in_tracers = [SparseTracer(trace, spvalue=spvalue) for spvalue in spvalues]
-  outs = yield in_tracers, {}
-  out_traces = [trace.full_raise(out) for out in outs]
-  buffers = popattr(main, 'spenv')._buffers
-  yield buffers, [out._spvalue for out in out_traces]
+@lu.transformation_with_aux2
+def sparsify_subtrace(f, store, tag, spenv, spvalues, *bufs):
+  with core.take_current_trace() as parent:
+    trace = SparseTrace(parent, tag, spenv)
+    with core.set_current_trace(trace):
+      in_tracers = [SparseTracer(trace, spvalue=spvalue) for spvalue in spvalues]
+      outs = f(*in_tracers)
+      out_traces = [trace.to_sparse_tracer(out) for out in outs]
+      buffers = spenv._buffers
+  store.store([out._spvalue for out in out_traces])
+  return buffers
 
 def sparsify_fun(wrapped_fun, args: list[ArrayOrSparse]):
-  with core.new_main(SparseTrace) as main:
-    spenv = SparsifyEnv()
-    spvalues = arrays_to_spvalues(spenv, args)
-    in_bufs = spenv._buffers
-    fun, out_spvalues = sparsify_subtrace(wrapped_fun, main, spvalues)
-    out_bufs = fun.call_wrapped(*in_bufs)
-    spenv = SparsifyEnv(out_bufs)
-    del main
+  tag = core.TraceTag()
+  spenv = SparsifyEnv()
+  spvalues = arrays_to_spvalues(spenv, args)
+  in_bufs = spenv._buffers
+  fun, out_spvalues = sparsify_subtrace(wrapped_fun, tag, spenv, spvalues)
+  out_bufs = fun.call_wrapped(*in_bufs)
+  spenv = SparsifyEnv(out_bufs)
   return spvalues_to_arrays(spenv, out_spvalues())
 
 def _sparsify_with_tracer(fun):
@@ -746,7 +737,7 @@ def _sparsify_jaxpr(spenv, jaxpr, *spvalues):
 
   args = spvalues_to_arrays(spenv, spvalues)
   args_flat, in_tree = tree_flatten(args)
-  avals_flat = [core.raise_to_shaped(core.get_aval(arg)) for arg in args_flat]
+  avals_flat = [core.get_aval(arg) for arg in args_flat]
   sp_jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(wrapped, avals_flat)
   sp_jaxpr = pe.ClosedJaxpr(sp_jaxpr, consts)
   assert out_tree is not None
@@ -773,7 +764,7 @@ sparse_rules_bcoo[lax.while_p] = _while_sparse
 
 def _pjit_sparse(spenv, *spvalues, jaxpr, in_shardings, out_shardings,
                  in_layouts, out_layouts, resource_env, donated_invars, name,
-                 keep_unused, inline):
+                 keep_unused, inline, compiler_options_kvs):
   if any(donated_invars):
     raise NotImplementedError("sparse xla_call with donated_invars")
 
@@ -809,7 +800,8 @@ def _pjit_sparse(spenv, *spvalues, jaxpr, in_shardings, out_shardings,
       donated_invars=donated_invars,
       name=name,
       keep_unused=keep_unused,
-      inline=inline)
+      inline=inline,
+      compiler_options_kvs=compiler_options_kvs)
   return arrays_to_spvalues(spenv, tree_unflatten(out_tree, out_flat))
 
 sparse_rules_bcoo[pjit.pjit_p] = _pjit_sparse
@@ -848,15 +840,15 @@ def _scan_sparse(spenv, *spvalues, jaxpr, num_consts, num_carry, **params):
 
 sparse_rules_bcoo[lax.scan_p] = _scan_sparse
 
-def _cond_sparse(spenv, pred, *operands, branches, linear, **params):
+def _cond_sparse(spenv, pred, *operands, branches, **params):
   sp_branches, treedefs = zip(*(_sparsify_jaxpr(spenv, jaxpr, *operands)
                                 for jaxpr in branches))
-  _check_tree_and_avals("sparsified true_fun and false_fun output",
+  _check_tree_and_avals("sparsified true_fun output",
                         treedefs[0], sp_branches[0].out_avals,
+                        "sparsified false_fun output",
                         treedefs[1], sp_branches[1].out_avals)
-  sp_linear = tuple(_duplicate_for_sparse_spvalues(operands, linear))
   args, _ = tree_flatten(spvalues_to_arrays(spenv, (pred, *operands)))
-  out_flat = lax.cond_p.bind(*args, branches=sp_branches, linear=sp_linear, **params)
+  out_flat = lax.cond_p.bind(*args, branches=sp_branches, **params)
   out = tree_unflatten(treedefs[0], out_flat)
   return arrays_to_spvalues(spenv, out)
 
