@@ -13,12 +13,12 @@
 # limitations under the License.
 
 # Note: import <name> as <name> is required for names to be exported.
-# See PEP 484 & https://github.com/google/jax/issues/7570
+# See PEP 484 & https://github.com/jax-ml/jax/issues/7570
 
 from jax.numpy import fft as fft
 from jax.numpy import linalg as linalg
 
-from jax._src.basearray import Array as ndarray
+from jax._src.basearray import Array as ndarray  # noqa: F401
 
 from jax._src.dtypes import (
     isdtype as isdtype,
@@ -53,7 +53,7 @@ from jax._src.numpy.lax_numpy import (
     bincount as bincount,
     blackman as blackman,
     block as block,
-    bool_ as bool,  # Array API alias for bool_
+    bool_ as bool,  # Array API alias for bool_  # noqa: F401
     bool_ as bool_,
     broadcast_arrays as broadcast_arrays,
     broadcast_shapes as broadcast_shapes,
@@ -176,6 +176,7 @@ from jax._src.numpy.lax_numpy import (
     logspace as logspace,
     mask_indices as mask_indices,
     matmul as matmul,
+    matvec as matvec,
     matrix_transpose as matrix_transpose,
     meshgrid as meshgrid,
     moveaxis as moveaxis,
@@ -202,6 +203,7 @@ from jax._src.numpy.lax_numpy import (
     printoptions as printoptions,
     promote_types as promote_types,
     put as put,
+    put_along_axis as put_along_axis,
     ravel as ravel,
     ravel_multi_index as ravel_multi_index,
     repeat as repeat,
@@ -212,7 +214,6 @@ from jax._src.numpy.lax_numpy import (
     rollaxis as rollaxis,
     rot90 as rot90,
     round as round,
-    round_ as round_,
     save as save,
     savez as savez,
     searchsorted as searchsorted,
@@ -258,6 +259,7 @@ from jax._src.numpy.lax_numpy import (
     vander as vander,
     vdot as vdot,
     vecdot as vecdot,
+    vecmat as vecmat,
     vsplit as vsplit,
     vstack as vstack,
     where as where,
@@ -265,6 +267,29 @@ from jax._src.numpy.lax_numpy import (
     zeros_like as zeros_like,
 )
 
+# TODO(slebedev): Remove the try-except once we upgrade to ml_dtypes 0.4.1.
+try:
+  from jax._src.numpy.lax_numpy import (
+    int2 as int2,
+    uint2 as uint2,
+  )
+except ImportError:
+  pass
+
+# TODO: Remove the try-except once we upgrade to ml_dtypes 0.5.0
+try:
+  from jax._src.numpy.lax_numpy import (
+    float8_e3m4 as float8_e3m4,
+    float8_e4m3 as float8_e4m3,
+  )
+except ImportError:
+  pass
+
+from jax._src.numpy.array_api_metadata import (
+  __array_api_version__ as __array_api_version__,
+  __array_namespace_info__ as __array_namespace_info__,
+)
+
 from jax._src.numpy.index_tricks import (
   c_ as c_,
   index_exp as index_exp,
@@ -294,8 +319,9 @@ from jax._src.numpy.reductions import (
     all as all,
     average as average,
     count_nonzero as count_nonzero,
-    cumsum as cumsum,
     cumprod as cumprod,
+    cumsum as cumsum,
+    cumulative_prod as cumulative_prod,
     cumulative_sum as cumulative_sum,
     max as max,
     mean as mean,
@@ -430,6 +456,7 @@ from jax._src.numpy.ufuncs import (
     sin as sin,
     sinc as sinc,
     sinh as sinh,
+    spacing as spacing,
     sqrt as sqrt,
     square as square,
     subtract as subtract,
@@ -452,13 +479,16 @@ del register_jax_array_methods
 
 
 _deprecations = {
-  # Deprecated 18 Sept 2023 and removed 06 Feb 2024
-  "trapz": (
-    "jnp.trapz is deprecated; use jnp.trapezoid instead.",
+  # Finalized 2024-12-13; remove after 2024-3-13
+  "round_": (
+    "jnp.round_ was deprecated in JAX 0.4.38; use jnp.round instead.",
     None
   ),
 }
 
-from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
-__getattr__ = _deprecation_getattr(__name__, _deprecations)
-del _deprecation_getattr
+import typing
+if not typing.TYPE_CHECKING:
+  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+  __getattr__ = _deprecation_getattr(__name__, _deprecations)
+  del _deprecation_getattr
+del typing
